@@ -26,14 +26,14 @@
 #endif
 
 static gpointer init_hs_helper (struct rspamd_config *cfg);
-static void start_hs_helper (struct rspamd_worker *worker);
+__attribute__((noreturn)) static void start_hs_helper (struct rspamd_worker *worker);
 
 worker_t hs_helper_worker = {
 		"hs_helper",                /* Name */
 		init_hs_helper,             /* Init function */
 		start_hs_helper,            /* Start function */
-		RSPAMD_WORKER_UNIQUE|RSPAMD_WORKER_KILLABLE|RSPAMD_WORKER_ALWAYS_START,
-		RSPAMD_WORKER_SOCKET_NONE,  /* No socket */
+		RSPAMD_WORKER_UNIQUE|RSPAMD_WORKER_KILLABLE|RSPAMD_WORKER_ALWAYS_START|RSPAMD_WORKER_NO_TERMINATE_DELAY,
+		RSPAMD_WORKER_SOCKET_NONE,
 		RSPAMD_WORKER_VER           /* Version info */
 };
 
@@ -47,7 +47,7 @@ static const guint64 rspamd_hs_helper_magic = 0x22d310157a2288a0ULL;
 struct hs_helper_ctx {
 	guint64 magic;
 	/* Events base */
-	struct event_base *ev_base;
+	struct ev_loop *event_loop;
 	/* DNS resolver */
 	struct rspamd_dns_resolver *resolver;
 	/* Config */
@@ -57,7 +57,7 @@ struct hs_helper_ctx {
 	gboolean loaded;
 	gdouble max_time;
 	gdouble recompile_time;
-	struct event recompile_timer;
+	ev_timer recompile_timer;
 };
 
 static gpointer
@@ -123,6 +123,12 @@ rspamd_hs_helper_cleanup_dir (struct hs_helper_ctx *ctx, gboolean forced)
 	gint rc;
 	gchar *pattern;
 	gboolean ret = TRUE;
+	pid_t our_pid = getpid ();
+
+	if (getenv("RSPAMD_NO_CLEANUP")) {
+		/* Skip all cleanup */
+		return TRUE;
+	}
 
 	if (stat (ctx->hs_dir, &st) == -1) {
 		msg_err ("cannot stat path %s, %s",
@@ -132,20 +138,39 @@ rspamd_hs_helper_cleanup_dir (struct hs_helper_ctx *ctx, gboolean forced)
 	}
 
 	globbuf.gl_offs = 0;
-	len = strlen (ctx->hs_dir) + 1 + sizeof ("*.hs.new") + 2;
+	/*
+	 * We reuse this buffer for .new patterns as well, so allocate with some
+	 * margin
+	 */
+	len = strlen (ctx->hs_dir) + 1 + sizeof ("*.hs") + sizeof (G_DIR_SEPARATOR);
 	pattern = g_malloc (len);
 	rspamd_snprintf (pattern, len, "%s%c%s", ctx->hs_dir, G_DIR_SEPARATOR, "*.hs");
 
 	if ((rc = glob (pattern, 0, NULL, &globbuf)) == 0) {
 		for (i = 0; i < globbuf.gl_pathc; i++) {
+			GError *err = NULL;
+
+			if (forced) {
+				g_set_error(&err, g_quark_from_static_string ("re_cache"),
+						0, "forced removal");
+			}
+
 			if (forced ||
 					!rspamd_re_cache_is_valid_hyperscan_file (ctx->cfg->re_cache,
-						globbuf.gl_pathv[i], TRUE, TRUE)) {
+						globbuf.gl_pathv[i], TRUE, TRUE, &err)) {
 				if (unlink (globbuf.gl_pathv[i]) == -1) {
-					msg_err ("cannot unlink %s: %s", globbuf.gl_pathv[i],
-							strerror (errno));
+					msg_err ("cannot unlink %s: %s; reason for expiration: %e", globbuf.gl_pathv[i],
+							strerror(errno), err);
 					ret = FALSE;
 				}
+				else {
+					msg_notice ("successfully removed outdated hyperscan file: %s; reason for expiration: %e",
+							globbuf.gl_pathv[i], err);
+				}
+			}
+
+			if (err) {
+				g_error_free(err);
 			}
 		}
 	}
@@ -160,10 +185,50 @@ rspamd_hs_helper_cleanup_dir (struct hs_helper_ctx *ctx, gboolean forced)
 	rspamd_snprintf (pattern, len, "%s%c%s", ctx->hs_dir, G_DIR_SEPARATOR, "*.hs.new");
 	if ((rc = glob (pattern, 0, NULL, &globbuf)) == 0) {
 		for (i = 0; i < globbuf.gl_pathc; i++) {
-			if (unlink (globbuf.gl_pathv[i]) == -1) {
-				msg_err ("cannot unlink %s: %s", globbuf.gl_pathv[i],
-						strerror (errno));
-				ret = FALSE;
+			/* Check if we have a pid in the filename */
+			const gchar *end_num = globbuf.gl_pathv[i] +
+					strlen (globbuf.gl_pathv[i]) - (sizeof (".hs.new") - 1);
+			const gchar *p = end_num - 1;
+			pid_t foreign_pid = -1;
+
+			while (p > globbuf.gl_pathv[i]) {
+				if (g_ascii_isdigit (*p)) {
+					p --;
+				}
+				else {
+					p ++;
+					break;
+				}
+			}
+
+			gulong ul;
+			if (p < end_num && rspamd_strtoul (p, end_num - p, &ul)) {
+				foreign_pid = ul;
+			}
+
+			/*
+			 * Remove only files that was left by us or some non-existing process
+			 * There could be another race condition but it would just leave
+			 * extra files which is relatively innocent?
+			 */
+			if (foreign_pid == -1 || foreign_pid == our_pid || kill (foreign_pid, 0) == -1) {
+				if (unlink(globbuf.gl_pathv[i]) == -1) {
+					msg_err ("cannot unlink %s: %s", globbuf.gl_pathv[i],
+							strerror(errno));
+					ret = FALSE;
+				}
+				else {
+					msg_notice ("successfully removed outdated hyperscan temporary file: %s; "
+								"pid of the file creator process: %P",
+							globbuf.gl_pathv[i],
+							foreign_pid);
+				}
+			}
+			else {
+				msg_notice ("skip removal of the hyperscan temporary file: %s; "
+							"pid of the file creator process: %P",
+						globbuf.gl_pathv[i],
+						foreign_pid);
 			}
 		}
 	}
@@ -178,55 +243,100 @@ rspamd_hs_helper_cleanup_dir (struct hs_helper_ctx *ctx, gboolean forced)
 	return ret;
 }
 
+/* Bad hack, but who cares */
+static gboolean hack_global_forced;
+
+static void
+rspamd_rs_delayed_cb (EV_P_ ev_timer *w, int revents)
+{
+	struct rspamd_worker *worker = (struct rspamd_worker *)w->data;
+	static struct rspamd_srv_command srv_cmd;
+	struct hs_helper_ctx *ctx;
+
+	ctx = (struct hs_helper_ctx *)worker->ctx;
+	memset (&srv_cmd, 0, sizeof (srv_cmd));
+	srv_cmd.type = RSPAMD_SRV_HYPERSCAN_LOADED;
+	rspamd_strlcpy (srv_cmd.cmd.hs_loaded.cache_dir, ctx->hs_dir,
+			sizeof (srv_cmd.cmd.hs_loaded.cache_dir));
+	srv_cmd.cmd.hs_loaded.forced = hack_global_forced;
+	hack_global_forced = FALSE;
+
+	rspamd_srv_send_command (worker,
+			ctx->event_loop, &srv_cmd, -1, NULL, NULL);
+	ev_timer_stop (EV_A_ w);
+	g_free (w);
+
+	ev_timer_again (EV_A_ &ctx->recompile_timer);
+}
+
+static void
+rspamd_rs_compile_cb (guint ncompiled, GError *err, void *cbd)
+{
+	struct rspamd_worker *worker = (struct rspamd_worker *)cbd;
+	ev_timer *tm;
+	ev_tstamp when = 0.0;
+	struct hs_helper_ctx *ctx;
+
+	ctx = (struct hs_helper_ctx *)worker->ctx;
+
+	if (err != NULL) {
+		/* Failed to compile: log and go out */
+		msg_err ("cannot compile Hyperscan database: %e", err);
+
+		return;
+	}
+
+	if (ncompiled > 0) {
+		/* Enforce update for other workers */
+		hack_global_forced = TRUE;
+	}
+
+	/*
+	 * Do not send notification unless all other workers are started
+	 * XXX: now we just sleep for 1 seconds to ensure that
+	 */
+	if (!ctx->loaded) {
+		when = 1.0; /* Postpone */
+		ctx->loaded = TRUE;
+		msg_info ("compiled %d regular expressions to the hyperscan tree, "
+				  "postpone loaded notification for %.0f seconds to avoid races",
+				ncompiled,
+				when);
+	}
+	else {
+		msg_info ("compiled %d regular expressions to the hyperscan tree, "
+				  "send loaded notification",
+				ncompiled);
+	}
+
+	tm = g_malloc0 (sizeof (*tm));
+	tm->data = (void *)worker;
+	ev_timer_init (tm, rspamd_rs_delayed_cb, when, 0);
+	ev_timer_start (ctx->event_loop, tm);
+}
+
 static gboolean
 rspamd_rs_compile (struct hs_helper_ctx *ctx, struct rspamd_worker *worker,
 		gboolean forced)
 {
-	GError *err = NULL;
-	static struct rspamd_srv_command srv_cmd;
-	gint ncompiled;
-
+#if !defined(__aarch64__) && !defined(__powerpc64__)
 	if (!(ctx->cfg->libs_ctx->crypto_ctx->cpu_config & CPUID_SSSE3)) {
 		msg_warn ("CPU doesn't have SSSE3 instructions set "
 				"required for hyperscan, disable hyperscan compilation");
 		return FALSE;
 	}
+#endif
 
 	if (!rspamd_hs_helper_cleanup_dir (ctx, forced)) {
 		msg_warn ("cannot cleanup cache dir '%s'", ctx->hs_dir);
 	}
 
-	if ((ncompiled = rspamd_re_cache_compile_hyperscan (ctx->cfg->re_cache,
+	hack_global_forced = forced; /* killmeplease */
+	rspamd_re_cache_compile_hyperscan (ctx->cfg->re_cache,
 			ctx->hs_dir, ctx->max_time, !forced,
-			&err)) == -1) {
-		msg_err ("failed to compile re cache: %e", err);
-		g_error_free (err);
-
-		return FALSE;
-	}
-
-	if (ncompiled > 0) {
-		msg_info ("compiled %d regular expressions to the hyperscan tree",
-				ncompiled);
-		forced = TRUE;
-	}
-
-	/*
-	 * Do not send notification unless all other workers are started
-	 * XXX: now we just sleep for 5 seconds to ensure that
-	 */
-	if (!ctx->loaded) {
-		sleep (5);
-		ctx->loaded = TRUE;
-	}
-
-	memset (&srv_cmd, 0, sizeof (srv_cmd));
-	srv_cmd.type = RSPAMD_SRV_HYPERSCAN_LOADED;
-	rspamd_strlcpy (srv_cmd.cmd.hs_loaded.cache_dir, ctx->hs_dir,
-			sizeof (srv_cmd.cmd.hs_loaded.cache_dir));
-	srv_cmd.cmd.hs_loaded.forced = forced;
-
-	rspamd_srv_send_command (worker, ctx->ev_base, &srv_cmd, -1, NULL, NULL);
+			ctx->event_loop,
+			rspamd_rs_compile_cb,
+			(void *)worker);
 
 	return TRUE;
 }
@@ -252,34 +362,33 @@ rspamd_hs_helper_reload (struct rspamd_main *rspamd_main,
 				strerror (errno));
 	}
 
+	/* Stop recompile */
+	ev_timer_stop (ctx->event_loop, &ctx->recompile_timer);
 	rspamd_rs_compile (ctx, worker, TRUE);
 
 	return TRUE;
 }
 
 static void
-rspamd_hs_helper_timer (gint fd, short what, gpointer ud)
+rspamd_hs_helper_timer (EV_P_ ev_timer *w, int revents)
 {
-	struct rspamd_worker *worker = ud;
+	struct rspamd_worker *worker = (struct rspamd_worker *)w->data;
 	struct hs_helper_ctx *ctx;
-	struct timeval tv;
 	double tim;
 
 	ctx = worker->ctx;
 	tim = rspamd_time_jitter (ctx->recompile_time, 0);
-	double_to_tv (tim, &tv);
-	event_del (&ctx->recompile_timer);
+	w->repeat = tim;
 	rspamd_rs_compile (ctx, worker, FALSE);
-	event_add (&ctx->recompile_timer, &tv);
 }
 
 static void
 start_hs_helper (struct rspamd_worker *worker)
 {
 	struct hs_helper_ctx *ctx = worker->ctx;
-	struct timeval tv;
 	double tim;
 
+	g_assert (rspamd_worker_check_context (worker->ctx, rspamd_hs_helper_magic));
 	ctx->cfg = worker->srv->cfg;
 
 	if (ctx->hs_dir == NULL) {
@@ -289,7 +398,7 @@ start_hs_helper (struct rspamd_worker *worker)
 		ctx->hs_dir = RSPAMD_DBDIR "/";
 	}
 
-	ctx->ev_base = rspamd_prepare_worker (worker,
+	ctx->event_loop = rspamd_prepare_worker (worker,
 			"hs_helper",
 			NULL);
 
@@ -301,17 +410,17 @@ start_hs_helper (struct rspamd_worker *worker)
 	rspamd_control_worker_add_cmd_handler (worker, RSPAMD_CONTROL_RECOMPILE,
 			rspamd_hs_helper_reload, ctx);
 
-	event_set (&ctx->recompile_timer, -1, EV_TIMEOUT, rspamd_hs_helper_timer,
-			worker);
-	event_base_set (ctx->ev_base, &ctx->recompile_timer);
+	ctx->recompile_timer.data = worker;
 	tim = rspamd_time_jitter (ctx->recompile_time, 0);
-	double_to_tv (tim, &tv);
-	event_add (&ctx->recompile_timer, &tv);
-	event_base_loop (ctx->ev_base, 0);
+	ev_timer_init (&ctx->recompile_timer, rspamd_hs_helper_timer, tim, 0.0);
+	ev_timer_start (ctx->event_loop, &ctx->recompile_timer);
+
+	ev_loop (ctx->event_loop, 0);
 	rspamd_worker_block_signals ();
 
 	rspamd_log_close (worker->srv->logger);
 	REF_RELEASE (ctx->cfg);
+	rspamd_unset_crash_handler (worker->srv);
 
 	exit (EXIT_SUCCESS);
 }
