@@ -19,6 +19,9 @@
 #include "libcryptobox/keypair_private.h"
 #include "libutil/str_util.h"
 #include "libutil/printf.h"
+#include "contrib/libottery/ottery.h"
+
+const guchar encrypted_magic[7] = {'r', 'u', 'c', 'l', 'e', 'v', '1'};
 
 static GQuark
 rspamd_keypair_quark (void)
@@ -209,6 +212,11 @@ rspamd_cryptobox_keypair_dtor (struct rspamd_cryptobox_keypair *kp)
 	sk = rspamd_cryptobox_keypair_sk (kp, &len);
 	g_assert (sk != NULL && len > 0);
 	rspamd_explicit_memzero (sk, len);
+
+	if (kp->extensions) {
+		ucl_object_unref (kp->extensions);
+	}
+
 	/* Not g_free as kp is aligned using posix_memalign */
 	free (kp);
 }
@@ -333,7 +341,7 @@ rspamd_pubkey_from_base32 (const gchar *b32,
 		len = strlen (b32);
 	}
 
-	decoded = rspamd_decode_base32 (b32, len, &dlen);
+	decoded = rspamd_decode_base32 (b32, len, &dlen, RSPAMD_BASE32_DEFAULT);
 
 	if (decoded == NULL) {
 		return NULL;
@@ -441,12 +449,19 @@ rspamd_pubkey_from_bin (const guchar *raw,
 
 
 const guchar *
-rspamd_pubkey_get_nm (struct rspamd_cryptobox_pubkey *p)
+rspamd_pubkey_get_nm (struct rspamd_cryptobox_pubkey *p,
+		struct rspamd_cryptobox_keypair *kp)
 {
 	g_assert (p != NULL);
 
 	if (p->nm) {
-		return p->nm->nm;
+		if (memcmp (kp->id, (const guchar *)&p->nm->sk_id, sizeof (guint64)) == 0) {
+			return p->nm->nm;
+		}
+
+		/* Wrong ID, need to recalculate */
+		REF_RELEASE (p->nm);
+		p->nm = NULL;
 	}
 
 	return NULL;
@@ -465,6 +480,7 @@ rspamd_pubkey_calculate_nm (struct rspamd_cryptobox_pubkey *p,
 			abort ();
 		}
 
+		memcpy (&p->nm->sk_id, kp->id, sizeof (guint64));
 		REF_INIT_RETAIN (p->nm, rspamd_cryptobox_nm_dtor);
 	}
 
@@ -496,6 +512,14 @@ rspamd_keypair_get_id (struct rspamd_cryptobox_keypair *kp)
 	return kp->id;
 }
 
+const ucl_object_t *
+rspamd_keypair_get_extensions (struct rspamd_cryptobox_keypair *kp)
+{
+	g_assert (kp != NULL);
+
+	return kp->extensions;
+}
+
 const guchar *
 rspamd_pubkey_get_id (struct rspamd_cryptobox_pubkey *pk)
 {
@@ -535,7 +559,7 @@ rspamd_keypair_print_component (guchar *data, gsize datalen,
 		g_string_set_size (res, res->len + b32_len);
 		res->len -= b32_len;
 		olen = rspamd_encode_base32_buf (data, datalen, res->str + res->len,
-				res->len + b32_len - 1);
+				res->len + b32_len - 1, RSPAMD_BASE32_DEFAULT);
 
 		if (olen > 0) {
 			res->len += olen;
@@ -725,7 +749,7 @@ rspamd_keypair_from_ucl (const ucl_object_t *obj)
 		dec_len = rspamd_decode_hex_buf (str, ucl_len, target, len);
 	}
 	else {
-		dec_len = rspamd_decode_base32_buf (str, ucl_len, target, len);
+		dec_len = rspamd_decode_base32_buf (str, ucl_len, target, len, RSPAMD_BASE32_DEFAULT);
 	}
 
 	if (dec_len != (gint)len) {
@@ -741,7 +765,7 @@ rspamd_keypair_from_ucl (const ucl_object_t *obj)
 		dec_len = rspamd_decode_hex_buf (str, ucl_len, target, len);
 	}
 	else {
-		dec_len = rspamd_decode_base32_buf (str, ucl_len, target, len);
+		dec_len = rspamd_decode_base32_buf (str, ucl_len, target, len, RSPAMD_BASE32_DEFAULT);
 	}
 
 	if (dec_len != (gint)len) {
@@ -752,12 +776,18 @@ rspamd_keypair_from_ucl (const ucl_object_t *obj)
 
 	rspamd_cryptobox_hash (kp->id, target, len, NULL, 0);
 
+	elt = ucl_object_lookup (obj, "extensions");
+	if (elt && ucl_object_type (elt) == UCL_OBJECT) {
+		/* Use copy to avoid issues with the refcounts */
+		kp->extensions = ucl_object_copy (elt);
+	}
+
 	return kp;
 }
 
 ucl_object_t *
 rspamd_keypair_to_ucl (struct rspamd_cryptobox_keypair *kp,
-		gboolean is_hex)
+					   enum rspamd_keypair_dump_flags flags)
 {
 	ucl_object_t *ucl_out, *elt;
 	gint how = 0;
@@ -766,7 +796,7 @@ rspamd_keypair_to_ucl (struct rspamd_cryptobox_keypair *kp,
 
 	g_assert (kp != NULL);
 
-	if (is_hex) {
+	if (flags & RSPAMD_KEYPAIR_DUMP_HEX) {
 		how |= RSPAMD_KEYPAIR_HEX;
 		encoding = "hex";
 	}
@@ -775,9 +805,16 @@ rspamd_keypair_to_ucl (struct rspamd_cryptobox_keypair *kp,
 		encoding = "base32";
 	}
 
-	ucl_out = ucl_object_typed_new (UCL_OBJECT);
-	elt = ucl_object_typed_new (UCL_OBJECT);
-	ucl_object_insert_key (ucl_out, elt, "keypair", 0, false);
+	if (flags & RSPAMD_KEYPAIR_DUMP_FLATTENED) {
+		ucl_out = ucl_object_typed_new (UCL_OBJECT);
+		elt = ucl_out;
+	}
+	else {
+		ucl_out = ucl_object_typed_new (UCL_OBJECT);
+		elt = ucl_object_typed_new (UCL_OBJECT);
+		ucl_object_insert_key (ucl_out, elt, "keypair", 0, false);
+	}
+
 
 	/* pubkey part */
 	keypair_out = rspamd_keypair_print (kp,
@@ -787,13 +824,15 @@ rspamd_keypair_to_ucl (struct rspamd_cryptobox_keypair *kp,
 			"pubkey", 0, false);
 	g_string_free (keypair_out, TRUE);
 
-	/* privkey part */
-	keypair_out = rspamd_keypair_print (kp,
-			RSPAMD_KEYPAIR_PRIVKEY|how);
-	ucl_object_insert_key (elt,
-			ucl_object_fromlstring (keypair_out->str, keypair_out->len),
+	if (!(flags & RSPAMD_KEYPAIR_DUMP_NO_SECRET)) {
+		/* privkey part */
+		keypair_out = rspamd_keypair_print(kp,
+			RSPAMD_KEYPAIR_PRIVKEY | how);
+		ucl_object_insert_key(elt,
+			ucl_object_fromlstring(keypair_out->str, keypair_out->len),
 			"privkey", 0, false);
-	g_string_free (keypair_out, TRUE);
+		g_string_free(keypair_out, TRUE);
+	}
 
 	keypair_out = rspamd_keypair_print (kp,
 			RSPAMD_KEYPAIR_ID|how);
@@ -818,93 +857,176 @@ rspamd_keypair_to_ucl (struct rspamd_cryptobox_keypair *kp,
 							"kex" : "sign"),
 					"type", 0, false);
 
+	if (kp->extensions) {
+		ucl_object_insert_key (elt, ucl_object_copy (kp->extensions),
+			"extensions", 0, false);
+	}
+
 	return ucl_out;
 }
 
 gboolean
-rspamd_keypair_sign (struct rspamd_cryptobox_keypair *kp,
-		const void *data, gsize len, guchar **sig, gsize *outlen,
-		GError **err)
+rspamd_keypair_decrypt (struct rspamd_cryptobox_keypair *kp,
+						const guchar *in, gsize inlen,
+						guchar **out, gsize *outlen,
+						GError **err)
 {
-	gsize siglen;
-	guint sklen;
+	const guchar *nonce, *mac, *data, *pubkey;
 
 	g_assert (kp != NULL);
-	g_assert (data != NULL);
-	g_assert (sig != NULL);
+	g_assert (in != NULL);
 
-	if (kp->type != RSPAMD_KEYPAIR_SIGN) {
+	if (kp->type != RSPAMD_KEYPAIR_KEX) {
 		g_set_error (err, rspamd_keypair_quark (), EINVAL,
-				"invalid keypair: expected signature pair");
+				"invalid keypair type");
 
 		return FALSE;
 	}
 
-	siglen = rspamd_cryptobox_signature_bytes (kp->alg);
-	*sig = g_malloc (siglen);
-	rspamd_cryptobox_sign (*sig, &siglen, data, len,
-			rspamd_cryptobox_keypair_sk (kp, &sklen), kp->alg);
+	if (inlen < sizeof (encrypted_magic) + rspamd_cryptobox_pk_bytes (kp->alg) +
+				rspamd_cryptobox_mac_bytes (kp->alg) +
+				rspamd_cryptobox_nonce_bytes (kp->alg)) {
+		g_set_error (err, rspamd_keypair_quark (), E2BIG, "invalid size: too small");
 
-	if (outlen != NULL) {
-		*outlen = siglen;
+		return FALSE;
+	}
+
+	if (memcmp (in, encrypted_magic, sizeof (encrypted_magic)) != 0) {
+		g_set_error (err, rspamd_keypair_quark (), EINVAL,
+				"invalid magic");
+
+		return FALSE;
+	}
+
+	/* Set pointers */
+	pubkey = in + sizeof (encrypted_magic);
+	mac = pubkey + rspamd_cryptobox_pk_bytes (kp->alg);
+	nonce = mac + rspamd_cryptobox_mac_bytes (kp->alg);
+	data = nonce + rspamd_cryptobox_nonce_bytes (kp->alg);
+
+	if (data - in >= inlen) {
+		g_set_error (err, rspamd_keypair_quark (), E2BIG, "invalid size: too small");
+
+		return FALSE;
+	}
+
+	inlen -= data - in;
+
+	/* Allocate memory for output */
+	*out = g_malloc (inlen);
+	memcpy (*out, data, inlen);
+
+	if (!rspamd_cryptobox_decrypt_inplace (*out, inlen, nonce, pubkey,
+			rspamd_keypair_component (kp, RSPAMD_KEYPAIR_COMPONENT_SK, NULL),
+			mac, kp->alg)) {
+		g_set_error (err, rspamd_keypair_quark (), EPERM, "verification failed");
+		g_free (*out);
+
+		return FALSE;
+	}
+
+	if (outlen) {
+		*outlen = inlen;
 	}
 
 	return TRUE;
 }
 
 gboolean
-rspamd_keypair_verify (struct rspamd_cryptobox_pubkey *pk,
-		const void *data, gsize len, const guchar *sig, gsize siglen,
-		GError **err)
+rspamd_keypair_encrypt (struct rspamd_cryptobox_keypair *kp,
+						const guchar *in, gsize inlen,
+						guchar **out, gsize *outlen,
+						GError **err)
 {
-	guint pklen;
+	guchar *nonce, *mac, *data, *pubkey;
+	struct rspamd_cryptobox_keypair *local;
+	gsize olen;
+
+	g_assert (kp != NULL);
+	g_assert (in != NULL);
+
+	if (kp->type != RSPAMD_KEYPAIR_KEX) {
+		g_set_error (err, rspamd_keypair_quark (), EINVAL,
+				"invalid keypair type");
+
+		return FALSE;
+	}
+
+	local = rspamd_keypair_new (kp->type, kp->alg);
+
+	olen = inlen + sizeof (encrypted_magic) +
+			rspamd_cryptobox_pk_bytes (kp->alg) +
+			rspamd_cryptobox_mac_bytes (kp->alg) +
+			rspamd_cryptobox_nonce_bytes (kp->alg);
+	*out = g_malloc (olen);
+	memcpy (*out, encrypted_magic, sizeof (encrypted_magic));
+	pubkey = *out + sizeof (encrypted_magic);
+	mac = pubkey + rspamd_cryptobox_pk_bytes (kp->alg);
+	nonce = mac + rspamd_cryptobox_mac_bytes (kp->alg);
+	data = nonce + rspamd_cryptobox_nonce_bytes (kp->alg);
+
+	ottery_rand_bytes (nonce, rspamd_cryptobox_nonce_bytes (kp->alg));
+	memcpy (data, in, inlen);
+	memcpy (pubkey, rspamd_keypair_component (kp,
+			RSPAMD_KEYPAIR_COMPONENT_PK, NULL),
+			rspamd_cryptobox_pk_bytes (kp->alg));
+	rspamd_cryptobox_encrypt_inplace (data, inlen, nonce, pubkey,
+			rspamd_keypair_component (local, RSPAMD_KEYPAIR_COMPONENT_SK, NULL),
+			mac, kp->alg);
+	rspamd_keypair_unref (local);
+
+	if (outlen) {
+		*outlen = olen;
+	}
+
+	return TRUE;
+}
+
+gboolean
+rspamd_pubkey_encrypt (struct rspamd_cryptobox_pubkey *pk,
+						const guchar *in, gsize inlen,
+						guchar **out, gsize *outlen,
+						GError **err)
+{
+	guchar *nonce, *mac, *data, *pubkey;
+	struct rspamd_cryptobox_keypair *local;
+	gsize olen;
 
 	g_assert (pk != NULL);
-	g_assert (data != NULL);
-	g_assert (sig != NULL);
+	g_assert (in != NULL);
 
-	if (pk->type != RSPAMD_KEYPAIR_SIGN) {
+	if (pk->type != RSPAMD_KEYPAIR_KEX) {
 		g_set_error (err, rspamd_keypair_quark (), EINVAL,
-				"invalid keypair: expected signature pair");
+				"invalid pubkey type");
 
 		return FALSE;
 	}
 
-	if (siglen != rspamd_cryptobox_signature_bytes (pk->alg)) {
-		g_set_error (err, rspamd_keypair_quark (), E2BIG,
-				"invalid signature length: %d; expected %d",
-				(int)siglen, (int) rspamd_cryptobox_signature_bytes (pk->alg));
+	local = rspamd_keypair_new (pk->type, pk->alg);
 
-		return FALSE;
-	}
+	olen = inlen + sizeof (encrypted_magic) +
+		   rspamd_cryptobox_pk_bytes (pk->alg) +
+		   rspamd_cryptobox_mac_bytes (pk->alg) +
+		   rspamd_cryptobox_nonce_bytes (pk->alg);
+	*out = g_malloc (olen);
+	memcpy (*out, encrypted_magic, sizeof (encrypted_magic));
+	pubkey = *out + sizeof (encrypted_magic);
+	mac = pubkey + rspamd_cryptobox_pk_bytes (pk->alg);
+	nonce = mac + rspamd_cryptobox_mac_bytes (pk->alg);
+	data = nonce + rspamd_cryptobox_nonce_bytes (pk->alg);
 
-	if (!rspamd_cryptobox_verify (sig, data, len,
-			rspamd_cryptobox_pubkey_pk (pk, &pklen), pk->alg)) {
-		g_set_error (err, rspamd_keypair_quark (), EPERM,
-				"signature verification failed");
+	ottery_rand_bytes (nonce, rspamd_cryptobox_nonce_bytes (pk->alg));
+	memcpy (data, in, inlen);
+	memcpy (pubkey, rspamd_pubkey_get_pk (pk, NULL),
+			rspamd_cryptobox_pk_bytes (pk->alg));
+	rspamd_cryptobox_encrypt_inplace (data, inlen, nonce, pubkey,
+			rspamd_keypair_component (local, RSPAMD_KEYPAIR_COMPONENT_SK, NULL),
+			mac, pk->alg);
+	rspamd_keypair_unref (local);
 
-		return FALSE;
+	if (outlen) {
+		*outlen = olen;
 	}
 
 	return TRUE;
-}
-
-gboolean
-rspamd_pubkey_equal (const struct rspamd_cryptobox_pubkey *k1,
-		const struct rspamd_cryptobox_pubkey *k2)
-{
-	guchar *p1 = NULL, *p2 = NULL;
-	guint len1, len2;
-
-
-	if (k1->alg == k2->alg && k1->type == k2->type) {
-		p1 = rspamd_cryptobox_pubkey_pk (k1, &len1);
-		p2 = rspamd_cryptobox_pubkey_pk (k2, &len2);
-
-		if (len1 == len2) {
-			return (memcmp (p1, p2, len1) == 0);
-		}
-	}
-
-	return FALSE;
-}
+}
\ No newline at end of file
