@@ -40,7 +40,7 @@ static const guint max_key_usages = 10000;
 
 #define msg_debug_mime(...)  rspamd_conditional_debug_fast (NULL, task->from_addr, \
         rspamd_mime_log_id, "mime", task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 INIT_LOG_MODULE(mime)
@@ -480,7 +480,7 @@ rspamd_mime_part_get_cd (struct rspamd_task *task, struct rspamd_mime_part *part
 		cd = rspamd_mempool_alloc0 (task->task_pool, sizeof (*cd));
 		cd->type = RSPAMD_CT_INLINE;
 
-		/* We can also have content dispositon definitions in Content-Type */
+		/* We can also have content disposition definitions in Content-Type */
 		if (part->ct && part->ct->attrs) {
 			RSPAMD_FTOK_ASSIGN (&srch, "name");
 			found = g_hash_table_lookup (part->ct->attrs, &srch);
@@ -539,7 +539,7 @@ rspamd_mime_part_get_cd (struct rspamd_task *task, struct rspamd_mime_part *part
 				cd = rspamd_mempool_alloc0 (task->task_pool, sizeof (*cd));
 				cd->type = RSPAMD_CT_INLINE;
 
-				/* We can also have content dispositon definitions in Content-Type */
+				/* We can also have content disposition definitions in Content-Type */
 				if (part->ct->attrs) {
 					RSPAMD_FTOK_ASSIGN (&srch, "name");
 					found = g_hash_table_lookup (part->ct->attrs, &srch);
@@ -1153,7 +1153,6 @@ rspamd_mime_preprocess_cb (struct rspamd_multipattern *mp,
 		void *context)
 {
 	const gchar *end = text + len, *p = text + match_pos, *bend;
-	gchar *lc_copy;
 	gsize blen;
 	gboolean closing = FALSE;
 	struct rspamd_mime_boundary b;
@@ -1163,7 +1162,6 @@ rspamd_mime_preprocess_cb (struct rspamd_multipattern *mp,
 	task = st->task;
 
 	if (G_LIKELY (p < end)) {
-		gboolean seen_non_dash = FALSE;
 
 		blen = 0;
 
@@ -1171,15 +1169,12 @@ rspamd_mime_preprocess_cb (struct rspamd_multipattern *mp,
 			if (*p == '\r' || *p == '\n') {
 				break;
 			}
-			else if (*p != '-') {
-				seen_non_dash = TRUE;
-			}
 
 			blen ++;
 			p ++;
 		}
 
-		if (blen > 0 && seen_non_dash) {
+		if (blen > 0) {
 			/* We have found something like boundary */
 			p = text + match_pos;
 			bend = p + blen - 1;
@@ -1192,7 +1187,7 @@ rspamd_mime_preprocess_cb (struct rspamd_multipattern *mp,
 					blen -= 2;
 				}
 				else {
-					/* Not a closing boundary somehow */
+					/* Not a closing boundary somehow, e.g. if a boundary=='-' */
 					bend ++;
 				}
 			}
@@ -1225,13 +1220,21 @@ rspamd_mime_preprocess_cb (struct rspamd_multipattern *mp,
 			b.boundary = p - st->start - 2;
 			b.start = bend - st->start;
 
-			if (closing) {
+			/* Small optimisation as boundaries are usually short strings */
+			gchar *lc_copy, lc_copy_buf[128];
+
+			if (blen + 2 < sizeof(lc_copy_buf)) {
+				lc_copy = lc_copy_buf;
+			}
+			else {
 				lc_copy = g_malloc (blen + 2);
+			}
+
+			if (closing) {
 				memcpy (lc_copy, p, blen + 2);
 				rspamd_str_lc (lc_copy, blen + 2);
 			}
 			else {
-				lc_copy = g_malloc (blen);
 				memcpy (lc_copy, p, blen);
 				rspamd_str_lc (lc_copy, blen);
 			}
@@ -1256,7 +1259,10 @@ rspamd_mime_preprocess_cb (struct rspamd_multipattern *mp,
 				b.closed_hash = 0;
 			}
 
-			g_free (lc_copy);
+			/* Check if a string has been allocated on the heap */
+			if (blen + 2 >= sizeof(lc_copy_buf)) {
+				g_free(lc_copy);
+			}
 			g_array_append_val (st->boundaries, b);
 		}
 	}
