@@ -27,7 +27,7 @@
 
 #define msg_debug_metric(...)  rspamd_conditional_debug_fast (NULL, NULL, \
         rspamd_metric_log_id, "metric", task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 INIT_LOG_MODULE(metric)
@@ -91,15 +91,19 @@ rspamd_create_metric_result (struct rspamd_task *task,
 	if (task->cfg) {
 		struct rspamd_action *act, *tmp;
 
-		metric_res->actions_limits = rspamd_mempool_alloc0 (task->task_pool,
-			sizeof (struct rspamd_action_result) * HASH_COUNT (task->cfg->actions));
+		metric_res->actions_config = rspamd_mempool_alloc0 (task->task_pool,
+				sizeof (struct rspamd_action_config) * HASH_COUNT (task->cfg->actions));
 		i = 0;
 
 		HASH_ITER (hh, task->cfg->actions, act, tmp) {
+			metric_res->actions_config[i].flags = RSPAMD_ACTION_RESULT_DEFAULT;
 			if (!(act->flags & RSPAMD_ACTION_NO_THRESHOLD)) {
-				metric_res->actions_limits[i].cur_limit = act->threshold;
+				metric_res->actions_config[i].cur_limit = act->threshold;
 			}
-			metric_res->actions_limits[i].action = act;
+			else {
+				metric_res->actions_config[i].flags |= RSPAMD_ACTION_RESULT_NO_THRESHOLD;
+			}
+			metric_res->actions_config[i].action = act;
 
 			i ++;
 		}
@@ -122,9 +126,14 @@ rspamd_pr_sort (const struct rspamd_passthrough_result *pra,
 	return prb->priority - pra->priority;
 }
 
-void
-rspamd_add_passthrough_result (struct rspamd_task *task, struct rspamd_action *action, guint priority,
-							   double target_score, const gchar *message, const gchar *module, guint flags,
+bool
+rspamd_add_passthrough_result (struct rspamd_task *task,
+							   struct rspamd_action *action,
+							   guint priority,
+							   double target_score,
+							   const gchar *message,
+							   const gchar *module,
+							   uint flags,
 							   struct rspamd_scan_result *scan_result)
 {
 	struct rspamd_passthrough_result *pr;
@@ -133,6 +142,29 @@ rspamd_add_passthrough_result (struct rspamd_task *task, struct rspamd_action *a
 		scan_result = task->result;
 	}
 
+	/* Find the specific action config */
+	struct rspamd_action_config *action_config = NULL;
+
+	for (unsigned int i = 0; i < scan_result->nactions; i ++) {
+		struct rspamd_action_config *cur = &scan_result->actions_config[i];
+
+		/* We assume that all action pointers are static */
+		if (cur->action == action) {
+			action_config = cur;
+			break;
+		}
+	}
+
+	if (action_config && (action_config->flags & RSPAMD_ACTION_RESULT_DISABLED)) {
+		msg_info_task ("<%s>: NOT set pre-result to '%s' %s(%.2f): '%s' from %s(%d); action is disabled",
+				MESSAGE_FIELD_CHECK (task, message_id), action->name,
+				flags & RSPAMD_PASSTHROUGH_LEAST ? "*least " : "",
+				target_score,
+				message, module, priority);
+
+		return false;
+	}
+
 	pr = rspamd_mempool_alloc (task->task_pool, sizeof (*pr));
 	pr->action = action;
 	pr->priority = priority;
@@ -160,6 +192,8 @@ rspamd_add_passthrough_result (struct rspamd_task *task, struct rspamd_action *a
 	}
 
 	scan_result->nresults ++;
+
+	return true;
 }
 
 static inline gdouble
@@ -197,12 +231,12 @@ insert_metric_result (struct rspamd_task *task,
 					  enum rspamd_symbol_insert_flags flags,
 					  bool *new_sym)
 {
-	struct rspamd_symbol_result *s = NULL;
+	struct rspamd_symbol_result *symbol_result = NULL;
 	gdouble final_score, *gr_score = NULL, next_gf = 1.0, diff;
 	struct rspamd_symbol *sdef;
 	struct rspamd_symbols_group *gr = NULL;
 	const ucl_object_t *mobj, *sobj;
-	gint max_shots, ret;
+	gint max_shots = G_MAXINT, ret;
 	guint i;
 	khiter_t k;
 	gboolean single = !!(flags & RSPAMD_SYMBOL_INSERT_SINGLE);
@@ -279,13 +313,21 @@ insert_metric_result (struct rspamd_task *task,
 	k = kh_get (rspamd_symbols_hash, metric_res->symbols, symbol);
 	if (k != kh_end (metric_res->symbols)) {
 		/* Existing metric score */
-		s = kh_value (metric_res->symbols, k);
+		symbol_result = kh_value (metric_res->symbols, k);
 		if (single) {
 			max_shots = 1;
 		}
 		else {
 			if (sdef) {
-				max_shots = sdef->nshots;
+				if (sdef->groups) {
+					PTR_ARRAY_FOREACH(sdef->groups, i, gr) {
+						if (gr->flags & RSPAMD_SYMBOL_GROUP_ONE_SHOT) {
+							max_shots = 1;
+						}
+					}
+				}
+
+				max_shots = MIN(max_shots, sdef->nshots);
 			}
 			else {
 				max_shots = task->cfg->default_max_shots;
@@ -294,14 +336,14 @@ insert_metric_result (struct rspamd_task *task,
 
 		msg_debug_metric ("nshots: %d for symbol %s", max_shots, symbol);
 
-		if (!single && (max_shots > 0 && (s->nshots >= max_shots))) {
+		if (!single && (max_shots > 0 && (symbol_result->nshots >= max_shots))) {
 			single = TRUE;
 		}
 
-		s->nshots ++;
+		symbol_result->nshots ++;
 
 		if (opt) {
-			rspamd_task_add_result_option (task, s, opt, strlen (opt));
+			rspamd_task_add_result_option (task, symbol_result, opt, strlen (opt));
 		}
 
 		/* Adjust diff */
@@ -311,13 +353,13 @@ insert_metric_result (struct rspamd_task *task,
 					symbol, diff);
 		}
 		else {
-			if (fabs (s->score) < fabs (final_score) &&
-				signbit (s->score) == signbit (final_score)) {
+			if (fabs (symbol_result->score) < fabs (final_score) &&
+				signbit (symbol_result->score) == signbit (final_score)) {
 				/* Replace less significant weight with a more significant one */
-				diff = final_score - s->score;
+				diff = final_score - symbol_result->score;
 				msg_debug_metric ("symbol %s can be inserted single time;"
 					  " weight adjusted %.2f + %.2f",
-						symbol, s->score, diff);
+						symbol, symbol_result->score, diff);
 			}
 			else {
 				diff = 0;
@@ -380,11 +422,11 @@ insert_metric_result (struct rspamd_task *task,
 				if (single) {
 					msg_debug_metric ("final score for single symbol %s = %.2f; %.2f diff",
 							symbol, final_score, diff);
-					s->score = final_score;
+					symbol_result->score = final_score;
 				} else {
 					msg_debug_metric ("increase final score for multiple symbol %s += %.2f = %.2f",
-							symbol, s->score, diff);
-					s->score += diff;
+							symbol, symbol_result->score, diff);
+					symbol_result->score += diff;
 				}
 			}
 		}
@@ -399,8 +441,8 @@ insert_metric_result (struct rspamd_task *task,
 		k = kh_put (rspamd_symbols_hash, metric_res->symbols,
 				sym_cpy, &ret);
 		g_assert (ret > 0);
-		s = rspamd_mempool_alloc0 (task->task_pool, sizeof (*s));
-		kh_value (metric_res->symbols, k) = s;
+		symbol_result = rspamd_mempool_alloc0 (task->task_pool, sizeof (*symbol_result));
+		kh_value (metric_res->symbols, k) = symbol_result;
 
 		/* Handle grow factor */
 		if (metric_res->grow_factor && final_score > 0) {
@@ -414,9 +456,9 @@ insert_metric_result (struct rspamd_task *task,
 		msg_debug_metric ("adjust grow factor to %.2f for symbol %s (%.2f final)",
 				next_gf, symbol, final_score);
 
-		s->name = sym_cpy;
-		s->sym = sdef;
-		s->nshots = 1;
+		symbol_result->name = sym_cpy;
+		symbol_result->sym = sdef;
+		symbol_result->nshots = 1;
 
 		if (sdef) {
 			/* Check group limits */
@@ -457,7 +499,7 @@ insert_metric_result (struct rspamd_task *task,
 
 			metric_res->score += final_score;
 			metric_res->grow_factor = next_gf;
-			s->score = final_score;
+			symbol_result->score = final_score;
 
 			if (final_score > epsilon) {
 				metric_res->npositive ++;
@@ -469,21 +511,21 @@ insert_metric_result (struct rspamd_task *task,
 			}
 		}
 		else {
-			s->score = 0;
+			symbol_result->score = 0;
 		}
 
 		if (opt) {
-			rspamd_task_add_result_option (task, s, opt, strlen (opt));
+			rspamd_task_add_result_option (task, symbol_result, opt, strlen (opt));
 		}
 	}
 
 	msg_debug_metric ("final insertion for symbol %s, score %.2f, factor: %f",
 			symbol,
-			s->score,
+			symbol_result->score,
 			final_score);
 	metric_res->nresults ++;
 
-	return s;
+	return symbol_result;
 }
 
 struct rspamd_symbol_result *
@@ -494,7 +536,7 @@ rspamd_task_insert_result_full (struct rspamd_task *task,
 								enum rspamd_symbol_insert_flags flags,
 								struct rspamd_scan_result *result)
 {
-	struct rspamd_symbol_result *s = NULL, *ret = NULL;
+	struct rspamd_symbol_result *symbol_result = NULL, *ret = NULL;
 	struct rspamd_scan_result *mres;
 
 	if (task->processed_stages & (RSPAMD_TASK_STAGE_IDEMPOTENT >> 1)) {
@@ -537,7 +579,7 @@ rspamd_task_insert_result_full (struct rspamd_task *task,
 
 			bool new_symbol = false;
 
-			s = insert_metric_result (task,
+			symbol_result = insert_metric_result (task,
 					symbol,
 					weight,
 					opt,
@@ -547,36 +589,38 @@ rspamd_task_insert_result_full (struct rspamd_task *task,
 
 			if (mres->name == NULL) {
 				/* Default result */
-				ret = s;
+				ret = symbol_result;
 
 				/* Process cache item */
-				if (s && task->cfg->cache && s->sym) {
+				if (symbol_result && task->cfg->cache && symbol_result->sym && symbol_result->nshots == 1) {
 					rspamd_symcache_inc_frequency (task->cfg->cache,
-							s->sym->cache_item);
+							symbol_result->sym->cache_item,
+							symbol_result->sym->name);
 				}
 			}
 			else if (new_symbol) {
 				/* O(N) but we normally don't have any shadow results */
-				LL_APPEND (ret, s);
+				LL_APPEND (ret, symbol_result);
 			}
 		}
 	}
 	else {
 		/* Specific insertion */
-		s = insert_metric_result (task,
+		symbol_result = insert_metric_result (task,
 				symbol,
 				weight,
 				opt,
 				result,
 				flags,
 				NULL);
-		ret = s;
+		ret = symbol_result;
 
 		if (result->name == NULL) {
 			/* Process cache item */
-			if (s && task->cfg->cache && s->sym) {
+			if (symbol_result && task->cfg->cache && symbol_result->sym && symbol_result->nshots == 1) {
 				rspamd_symcache_inc_frequency (task->cfg->cache,
-						s->sym->cache_item);
+						symbol_result->sym->cache_item,
+						symbol_result->sym->name);
 			}
 		}
 	}
@@ -772,17 +816,31 @@ rspamd_task_add_result_option (struct rspamd_task *task,
 	return ret;
 }
 
+struct rspamd_action_config *
+rspamd_find_action_config_for_action (struct rspamd_scan_result *scan_result,
+									  struct rspamd_action *act)
+{
+	for (unsigned int i = 0; i < scan_result->nactions; i ++) {
+		struct rspamd_action_config *cur = &scan_result->actions_config[i];
+
+		if (act == cur->action) {
+			return cur;
+		}
+	}
+
+	return NULL;
+}
+
 struct rspamd_action *
 rspamd_check_action_metric (struct rspamd_task *task,
 							struct rspamd_passthrough_result **ppr,
 							struct rspamd_scan_result *scan_result)
 {
-	struct rspamd_action_result *action_lim,
+	struct rspamd_action_config *action_lim,
 			*noaction = NULL;
 	struct rspamd_action *selected_action = NULL, *least_action = NULL;
 	struct rspamd_passthrough_result *pr, *sel_pr = NULL;
 	double max_score = -(G_MAXDOUBLE), sc;
-	int i;
 	gboolean seen_least = FALSE;
 
 	if (scan_result == NULL) {
@@ -791,6 +849,14 @@ rspamd_check_action_metric (struct rspamd_task *task,
 
 	if (scan_result->passthrough_result != NULL)  {
 		DL_FOREACH (scan_result->passthrough_result, pr) {
+			struct rspamd_action_config *act_config =
+					rspamd_find_action_config_for_action (scan_result, pr->action);
+
+			/* Skip disabled actions */
+			if (act_config && (act_config->flags & RSPAMD_ACTION_RESULT_DISABLED)) {
+				continue;
+			}
+
 			if (!seen_least || !(pr->flags & RSPAMD_PASSTHROUGH_LEAST)) {
 				sc = pr->target_score;
 				selected_action = pr->action;
@@ -849,14 +915,18 @@ rspamd_check_action_metric (struct rspamd_task *task,
 	/*
 	 * Select result by score
 	 */
-	for (i = scan_result->nactions - 1; i >= 0; i--) {
-		action_lim = &scan_result->actions_limits[i];
+	for (size_t i = scan_result->nactions - 1; i != (size_t)-1; i--) {
+		action_lim = &scan_result->actions_config[i];
 		sc = action_lim->cur_limit;
 
 		if (action_lim->action->action_type == METRIC_ACTION_NOACTION) {
 			noaction = action_lim;
 		}
 
+		if ((action_lim->flags & (RSPAMD_ACTION_RESULT_DISABLED|RSPAMD_ACTION_RESULT_NO_THRESHOLD))) {
+			continue;
+		}
+
 		if (isnan (sc) ||
 			(action_lim->action->flags & (RSPAMD_ACTION_NO_THRESHOLD|RSPAMD_ACTION_HAM))) {
 			continue;
@@ -956,15 +1026,16 @@ struct rspamd_symbol_result* rspamd_task_remove_symbol_result (
 			if (result->sym_groups && res->sym) {
 				struct rspamd_symbol_group *gr;
 				gint i;
+				khiter_t k_groups;
 
 				PTR_ARRAY_FOREACH (res->sym->groups, i, gr) {
 					gdouble *gr_score;
 
-					k = kh_get (rspamd_symbols_group_hash,
+					k_groups = kh_get (rspamd_symbols_group_hash,
 							result->sym_groups, gr);
 
-					if (k != kh_end (result->sym_groups)) {
-						gr_score = &kh_value (result->sym_groups, k);
+					if (k_groups != kh_end (result->sym_groups)) {
+						gr_score = &kh_value (result->sym_groups, k_groups);
 
 						if (gr_score) {
 							*gr_score -= res->score;
@@ -1023,4 +1094,4 @@ rspamd_find_metric_result (struct rspamd_task *task,
 	}
 
 	return NULL;
-}
\ No newline at end of file
+}
