@@ -164,13 +164,14 @@ rspamd_config_new (enum rspamd_config_init_flags flags)
 	rspamd_mempool_t *pool;
 
 	pool = rspamd_mempool_new (8 * 1024 * 1024, "cfg", 0);
-	cfg = rspamd_mempool_alloc0 (pool, sizeof (*cfg));
+	cfg = rspamd_mempool_alloc0_type(pool, struct rspamd_config);
 	/* Allocate larger pool for cfg */
 	cfg->cfg_pool = pool;
 	cfg->dns_timeout = 1.0;
 	cfg->dns_retransmits = 5;
 	/* 16 sockets per DNS server */
 	cfg->dns_io_per_server = 16;
+	cfg->unknown_weight = NAN;
 
 	/* Add all internal actions to keep compatibility */
 	for (int i = METRIC_ACTION_REJECT; i < METRIC_ACTION_MAX; i ++) {
@@ -248,6 +249,9 @@ rspamd_config_new (enum rspamd_config_init_flags flags)
 	cfg->lua_gc_step = DEFAULT_LUA_GC_STEP;
 	cfg->full_gc_iters = DEFAULT_GC_MAXITERS;
 
+	/* Default hyperscan cache */
+	cfg->hs_cache_dir = RSPAMD_DBDIR "/";
+
 	if (!(flags & RSPAMD_CONFIG_INIT_SKIP_LUA)) {
 		cfg->lua_state = rspamd_lua_init (flags & RSPAMD_CONFIG_INIT_WIPE_LUA_MEM);
 		cfg->own_lua_state = TRUE;
@@ -269,9 +273,7 @@ rspamd_config_new (enum rspamd_config_init_flags flags)
 	cfg->images_cache_size = 256;
 	cfg->monitored_ctx = rspamd_monitored_ctx_init ();
 	cfg->neighbours = ucl_object_typed_new (UCL_OBJECT);
-#ifdef WITH_HIREDIS
 	cfg->redis_pool = rspamd_redis_pool_init ();
-#endif
 	cfg->default_max_shots = DEFAULT_MAX_SHOTS;
 	cfg->max_sessions_cache = DEFAULT_MAX_SESSIONS;
 	cfg->maps_cache_dir = rspamd_mempool_strdup (cfg->cfg_pool, RSPAMD_DBDIR);
@@ -342,14 +344,12 @@ rspamd_config_free (struct rspamd_config *cfg)
 
 	if (cfg->lua_state && cfg->own_lua_state) {
 		lua_thread_pool_free (cfg->lua_thread_pool);
-		lua_close (cfg->lua_state);
+		rspamd_lua_close (cfg->lua_state);
 	}
 
-#ifdef WITH_HIREDIS
 	if (cfg->redis_pool) {
 		rspamd_redis_pool_destroy (cfg->redis_pool);
 	}
-#endif
 
 	rspamd_upstreams_library_unref (cfg->ups_ctx);
 	HASH_CLEAR (hh, cfg->actions);
@@ -385,10 +385,10 @@ rspamd_config_get_module_opt (struct rspamd_config *cfg,
 	return res;
 }
 
-gchar
+gint
 rspamd_config_parse_flag (const gchar *str, guint len)
 {
-	gchar c;
+	gint c;
 
 	if (!str || !*str) {
 		return -1;
@@ -820,7 +820,7 @@ rspamd_config_post_load (struct rspamd_config *cfg,
 	rspamd_regexp_library_init (cfg);
 	rspamd_multipattern_library_init (cfg->hs_cache_dir);
 
-#if defined(WITH_HYPERSCAN) && !defined(__aarch64__)
+#if defined(WITH_HYPERSCAN) && !defined(__aarch64__) && !defined(__powerpc64__)
 	if (!cfg->disable_hyperscan) {
 		if (!(cfg->libs_ctx->crypto_ctx->cpu_config & CPUID_SSSE3)) {
 			msg_warn_config ("CPU doesn't have SSSE3 instructions set "
@@ -1109,7 +1109,7 @@ rspamd_include_map_handler (const guchar *data, gsize len,
  * $RUNDIR - local states directory
  * $DBDIR - databases dir
  * $LOGDIR - logs dir
- * $PLUGINSDIR - pluggins dir
+ * $PLUGINSDIR - plugins dir
  * $PREFIX - installation prefix
  * $VERSION - rspamd version
  */
@@ -1277,7 +1277,7 @@ rspamd_config_check_statfiles (struct rspamd_classifier_config *cf)
 		/* We have only one statfile */
 		return FALSE;
 	}
-	/* We have not detected any statfile that has different class, so turn on euristic based on symbol's name */
+	/* We have not detected any statfile that has different class, so turn on heuristic based on symbol's name */
 	has_other = FALSE;
 	cur = cf->statfiles;
 	while (cur) {
@@ -1570,11 +1570,11 @@ rspamd_config_new_symbol (struct rspamd_config *cfg, const gchar *symbol,
 	/* Search for symbol group */
 	if (group == NULL) {
 		group = "ungrouped";
-		sym_def->flags |= RSPAMD_SYMBOL_FLAG_UNGROUPPED;
+		sym_def->flags |= RSPAMD_SYMBOL_FLAG_UNGROUPED;
 	}
 	else {
 		if (strcmp (group, "ungrouped") == 0) {
-			sym_def->flags |= RSPAMD_SYMBOL_FLAG_UNGROUPPED;
+			sym_def->flags |= RSPAMD_SYMBOL_FLAG_UNGROUPED;
 		}
 	}
 
@@ -1587,7 +1587,7 @@ rspamd_config_new_symbol (struct rspamd_config *cfg, const gchar *symbol,
 	sym_def->gr = sym_group;
 	g_hash_table_insert (sym_group->symbols, sym_def->name, sym_def);
 
-	if (!(sym_def->flags & RSPAMD_SYMBOL_FLAG_UNGROUPPED)) {
+	if (!(sym_def->flags & RSPAMD_SYMBOL_FLAG_UNGROUPED)) {
 		g_ptr_array_add (sym_def->groups, sym_group);
 	}
 }
@@ -1625,7 +1625,7 @@ rspamd_config_add_symbol (struct rspamd_config *cfg,
 			}
 
 			if (!has_group) {
-				/* Non-empty group has a priority over non-groupped one */
+				/* Non-empty group has a priority over non-grouped one */
 				sym_group = g_hash_table_lookup (cfg->groups, group);
 
 				if (sym_group == NULL) {
@@ -1633,12 +1633,13 @@ rspamd_config_add_symbol (struct rspamd_config *cfg,
 					sym_group = rspamd_config_new_group (cfg, group);
 				}
 
-				if (!sym_def->gr) {
+				if ((!sym_def->gr) || (sym_def->flags & RSPAMD_SYMBOL_FLAG_UNGROUPED)) {
 					sym_def->gr = sym_group;
+					sym_def->flags &= ~RSPAMD_SYMBOL_FLAG_UNGROUPED;
 				}
 
 				g_hash_table_insert (sym_group->symbols, sym_def->name, sym_def);
-				sym_def->flags &= ~(RSPAMD_SYMBOL_FLAG_UNGROUPPED);
+				sym_def->flags &= ~(RSPAMD_SYMBOL_FLAG_UNGROUPED);
 				g_ptr_array_add (sym_def->groups, sym_group);
 			}
 		}
@@ -1757,7 +1758,7 @@ rspamd_config_add_symbol_group (struct rspamd_config *cfg,
 		}
 
 		if (!has_group) {
-			/* Non-empty group has a priority over non-groupped one */
+			/* Non-empty group has a priority over non-grouped one */
 			sym_group = g_hash_table_lookup (cfg->groups, group);
 
 			if (sym_group == NULL) {
@@ -1770,7 +1771,7 @@ rspamd_config_add_symbol_group (struct rspamd_config *cfg,
 			}
 
 			g_hash_table_insert (sym_group->symbols, sym_def->name, sym_def);
-			sym_def->flags &= ~(RSPAMD_SYMBOL_FLAG_UNGROUPPED);
+			sym_def->flags &= ~(RSPAMD_SYMBOL_FLAG_UNGROUPED);
 			g_ptr_array_add (sym_def->groups, sym_group);
 
 			return TRUE;
@@ -1780,13 +1781,70 @@ rspamd_config_add_symbol_group (struct rspamd_config *cfg,
 	return FALSE;
 }
 
+gboolean
+rspamd_config_is_enabled_from_ucl (rspamd_mempool_t *pool,
+		const ucl_object_t *obj)
+{
+	{
+		const ucl_object_t *enabled;
+
+		enabled = ucl_object_lookup(obj, "enabled");
+
+		if (enabled) {
+			if (ucl_object_type(enabled) == UCL_BOOLEAN) {
+				return ucl_object_toboolean(enabled);
+			}
+			else if (ucl_object_type(enabled) == UCL_STRING) {
+				gint ret = rspamd_config_parse_flag(ucl_object_tostring(enabled), 0);
+
+				if (ret == 0) {
+					return FALSE;
+				}
+				else if (ret == -1) {
+
+					msg_info_pool_check ("wrong value for the `enabled` key");
+					return FALSE;
+				}
+				/* Default return is TRUE here */
+			}
+		}
+	}
+
+	{
+		const ucl_object_t *disabled;
+
+		disabled = ucl_object_lookup(obj, "disabled");
+
+		if (disabled) {
+			if (ucl_object_type(disabled) == UCL_BOOLEAN) {
+				return !ucl_object_toboolean(disabled);
+			}
+			else if (ucl_object_type(disabled) == UCL_STRING) {
+				gint ret = rspamd_config_parse_flag(ucl_object_tostring(disabled), 0);
+
+				if (ret == 0) {
+					return TRUE;
+				}
+				else if (ret == -1) {
+
+					msg_info_pool_check ("wrong value for the `disabled` key");
+					return FALSE;
+				}
+
+				return FALSE;
+			}
+		}
+	}
+
+	return TRUE;
+}
 
 gboolean
 rspamd_config_is_module_enabled (struct rspamd_config *cfg,
 		const gchar *module_name)
 {
-	gboolean is_c = FALSE;
-	const ucl_object_t *conf, *enabled;
+	gboolean is_c = FALSE, enabled;
+	const ucl_object_t *conf;
 	GList *cur;
 	struct rspamd_symbols_group *gr;
 	lua_State *L = cfg->lua_state;
@@ -1845,45 +1903,16 @@ rspamd_config_is_module_enabled (struct rspamd_config *cfg,
 		}
 	}
 	else {
-		enabled = ucl_object_lookup (conf, "enabled");
+		enabled = rspamd_config_is_enabled_from_ucl (cfg->cfg_pool, conf);
 
-		if (enabled) {
-			if (ucl_object_type (enabled) == UCL_BOOLEAN) {
-				if (!ucl_object_toboolean (enabled)) {
-					rspamd_plugins_table_push_elt (L,
-							"disabled_explicitly", module_name);
+		if (!enabled) {
+			rspamd_plugins_table_push_elt (L,
+					"disabled_explicitly", module_name);
 
-					msg_info_config (
-							"%s module %s is disabled in the configuration",
-							is_c ? "internal" : "lua", module_name);
-					return FALSE;
-				}
-			}
-			else if (ucl_object_type (enabled) == UCL_STRING) {
-				gint ret;
-
-				ret = rspamd_config_parse_flag (ucl_object_tostring (enabled), 0);
-
-				if (ret == 0) {
-					rspamd_plugins_table_push_elt (L,
-							"disabled_explicitly", module_name);
-
-					msg_info_config (
-							"%s module %s is disabled in the configuration",
-							is_c ? "internal" : "lua", module_name);
-					return FALSE;
-				}
-				else if (ret == -1) {
-					rspamd_plugins_table_push_elt (L,
-							"disabled_failed", module_name);
-
-					msg_info_config (
-							"%s module %s has wrong enabled flag (%s) in the configuration",
-							is_c ? "internal" : "lua", module_name,
-							ucl_object_tostring (enabled));
-					return FALSE;
-				}
-			}
+			msg_info_config (
+					"%s module %s is disabled in the configuration",
+					is_c ? "internal" : "lua", module_name);
+			return FALSE;
 		}
 	}
 
@@ -2608,7 +2637,7 @@ rspamd_init_libs (void)
 	if ((ctx->crypto_ctx->cpu_config & CPUID_RDRAND) == 0) {
 		ottery_config_disable_entropy_sources (ottery_cfg,
 				OTTERY_ENTROPY_SRC_RDRAND);
-#if OPENSSL_VERSION_NUMBER >= 0x1000104fL && !defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER >= 0x1000104fL && OPENSSL_VERSION_NUMBER < 0x30000000L && !defined(LIBRESSL_VERSION_NUMBER)
 		RAND_set_rand_engine (NULL);
 #endif
 	}
@@ -2830,7 +2859,7 @@ rspamd_libs_reset_decompression (struct rspamd_external_libs_ctx *ctx)
 		return FALSE;
 	}
 	else {
-		r = ZSTD_resetDStream (ctx->in_zstream);
+		r = ZSTD_DCtx_reset (ctx->in_zstream, ZSTD_reset_session_only);
 
 		if (ZSTD_isError (r)) {
 			msg_err ("cannot init decompression stream: %s",
@@ -2855,7 +2884,10 @@ rspamd_libs_reset_compression (struct rspamd_external_libs_ctx *ctx)
 	}
 	else {
 		/* Dictionary will be reused automatically if specified */
-		r = ZSTD_resetCStream (ctx->out_zstream, 0);
+		r = ZSTD_CCtx_reset (ctx->out_zstream, ZSTD_reset_session_only);
+		if (!ZSTD_isError (r)) {
+			r = ZSTD_CCtx_setPledgedSrcSize (ctx->out_zstream, ZSTD_CONTENTSIZE_UNKNOWN);
+		}
 
 		if (ZSTD_isError (r)) {
 			msg_err ("cannot init compression stream: %s",
