@@ -91,6 +91,8 @@ static const rspamd_ftok_t last_modified_header = {
 		.len = 13
 };
 
+static void rspamd_http_event_handler (int fd, short what, gpointer ud);
+static void rspamd_http_ssl_err_handler (gpointer ud, GError *err);
 
 
 #define HTTP_ERROR http_error_quark ()
@@ -1030,52 +1032,48 @@ rspamd_http_event_handler (int fd, short what, gpointer ud)
 		}
 	}
 	else if (what == EV_TIMEOUT) {
-		/* Let's try to read from the socket first */
-		r = rspamd_http_try_read (fd, conn, priv, pbuf, &d);
+		if (!priv->ssl) {
+			/* Let's try to read from the socket first */
+			r = rspamd_http_try_read(fd, conn, priv, pbuf, &d);
 
-		if (r > 0) {
-			if (http_parser_execute (&priv->parser, &priv->parser_cb,
-					d, r) != (size_t)r || priv->parser.http_errno != 0) {
-				err = g_error_new (HTTP_ERROR, 400,
-						"HTTP parser error: %s",
-						http_errno_description (priv->parser.http_errno));
+			if (r > 0) {
+				if (http_parser_execute(&priv->parser, &priv->parser_cb,
+						d, r) != (size_t) r || priv->parser.http_errno != 0) {
+					err = g_error_new(HTTP_ERROR, 400,
+							"HTTP parser error: %s",
+							http_errno_description(priv->parser.http_errno));
 
-				if (!conn->finished) {
-					conn->error_handler (conn, err);
+					if (!conn->finished) {
+						conn->error_handler(conn, err);
+					}
+					else {
+						msg_err ("got error after HTTP request is finished: %e", err);
+					}
+
+					g_error_free(err);
+
+					REF_RELEASE (pbuf);
+					rspamd_http_connection_unref(conn);
+
+					return;
 				}
-				else {
-					msg_err ("got error after HTTP request is finished: %e", err);
-				}
-
-				g_error_free (err);
+			}
+			else {
+				err = g_error_new(HTTP_ERROR, 408,
+						"IO timeout");
+				conn->error_handler(conn, err);
+				g_error_free(err);
 
 				REF_RELEASE (pbuf);
-				rspamd_http_connection_unref (conn);
+				rspamd_http_connection_unref(conn);
 
 				return;
 			}
 		}
-		else if (r == 0) {
-			if (!conn->finished) {
-				err = g_error_new (HTTP_ERROR, 408,
-						"IO timeout");
-				conn->error_handler (conn, err);
-				g_error_free (err);
-
-			}
-			REF_RELEASE (pbuf);
-			rspamd_http_connection_unref (conn);
-
-			return;
-		}
 		else {
-			err = g_error_new (HTTP_ERROR, 408,
-					"IO timeout");
-			conn->error_handler (conn, err);
-			g_error_free (err);
-
+			/* In case of SSL we disable this logic as we already came from SSL handler */
 			REF_RELEASE (pbuf);
-			rspamd_http_connection_unref (conn);
+			rspamd_http_connection_unref(conn);
 
 			return;
 		}
@@ -1240,12 +1238,13 @@ rspamd_http_connection_new_client (struct rspamd_http_context *ctx,
 }
 
 struct rspamd_http_connection *
-rspamd_http_connection_new_keepalive (struct rspamd_http_context *ctx,
-									  rspamd_http_body_handler_t body_handler,
-									  rspamd_http_error_handler_t error_handler,
-									  rspamd_http_finish_handler_t finish_handler,
-									  rspamd_inet_addr_t *addr,
-									  const gchar *host)
+rspamd_http_connection_new_client_keepalive (struct rspamd_http_context *ctx,
+											 rspamd_http_body_handler_t body_handler,
+											 rspamd_http_error_handler_t error_handler,
+											 rspamd_http_finish_handler_t finish_handler,
+											 unsigned opts,
+											 rspamd_inet_addr_t *addr,
+											 const gchar *host)
 {
 	struct rspamd_http_connection *conn;
 
@@ -1253,7 +1252,8 @@ rspamd_http_connection_new_keepalive (struct rspamd_http_context *ctx,
 		ctx = rspamd_http_context_default ();
 	}
 
-	conn = rspamd_http_context_check_keepalive (ctx, addr, host);
+	conn = rspamd_http_context_check_keepalive(ctx, addr, host,
+			opts & RSPAMD_HTTP_CLIENT_SSL);
 
 	if (conn) {
 		return conn;
@@ -1261,11 +1261,12 @@ rspamd_http_connection_new_keepalive (struct rspamd_http_context *ctx,
 
 	conn = rspamd_http_connection_new_client (ctx,
 			body_handler, error_handler, finish_handler,
-			RSPAMD_HTTP_CLIENT_SIMPLE|RSPAMD_HTTP_CLIENT_KEEP_ALIVE,
+			opts|RSPAMD_HTTP_CLIENT_SIMPLE|RSPAMD_HTTP_CLIENT_KEEP_ALIVE,
 			addr);
 
 	if (conn) {
-		rspamd_http_context_prepare_keepalive (ctx, conn, addr, host);
+		rspamd_http_context_prepare_keepalive(ctx, conn, addr, host,
+				opts & RSPAMD_HTTP_CLIENT_SSL);
 	}
 
 	return conn;
@@ -1348,7 +1349,7 @@ rspamd_http_connection_copy_msg (struct rspamd_http_message *msg, GError **err)
 	if (msg->body_buf.len > 0) {
 
 		if (msg->flags & RSPAMD_HTTP_FLAG_SHMEM) {
-			/* Avoid copying by just maping a shared segment */
+			/* Avoid copying by just mapping a shared segment */
 			new_msg->flags |= RSPAMD_HTTP_FLAG_SHMEM_IMMUTABLE;
 
 			storage = &new_msg->body_buf.c;
@@ -1527,9 +1528,18 @@ rspamd_http_connection_read_message_common (struct rspamd_http_connection *conn,
 	priv->buf->data = rspamd_fstring_sized_new (8192);
 	priv->flags |= RSPAMD_HTTP_CONN_FLAG_NEW_HEADER;
 
-	rspamd_ev_watcher_init (&priv->ev, conn->fd, EV_READ,
-			rspamd_http_event_handler, conn);
-	rspamd_ev_watcher_start (priv->ctx->event_loop, &priv->ev, priv->timeout);
+	if (!priv->ssl) {
+		rspamd_ev_watcher_init(&priv->ev, conn->fd, EV_READ,
+				rspamd_http_event_handler, conn);
+		rspamd_ev_watcher_start(priv->ctx->event_loop, &priv->ev, priv->timeout);
+	}
+	else {
+		rspamd_ssl_connection_restore_handlers (priv->ssl,
+				rspamd_http_event_handler,
+				rspamd_http_ssl_err_handler,
+				conn,
+				EV_READ);
+	}
 
 	priv->flags &= ~RSPAMD_HTTP_CONN_FLAG_RESETED;
 }
@@ -1857,7 +1867,8 @@ rspamd_http_message_write_header (const gchar* mime_type, gboolean encrypted,
 
 			if (encrypted) {
 				/* TODO: Add proxy support to HTTPCrypt */
-				rspamd_printf_fstring (buf,
+				if (rspamd_http_message_is_standard_port(msg)) {
+					rspamd_printf_fstring(buf,
 						"%s %s HTTP/1.1\r\n"
 						"Connection: %s\r\n"
 						"Host: %s\r\n"
@@ -1868,16 +1879,32 @@ rspamd_http_message_write_header (const gchar* mime_type, gboolean encrypted,
 						conn_type,
 						host,
 						enclen);
+				}
+				else {
+					rspamd_printf_fstring(buf,
+						"%s %s HTTP/1.1\r\n"
+						"Connection: %s\r\n"
+						"Host: %s:%d\r\n"
+						"Content-Length: %z\r\n"
+						"Content-Type: application/octet-stream\r\n",
+						"POST",
+						"/post",
+						conn_type,
+						host,
+						msg->port,
+						enclen);
+				}
 			}
 			else {
 				if (conn->priv->flags & RSPAMD_HTTP_CONN_FLAG_PROXY) {
+					/* Write proxied request */
 					if ((msg->flags & RSPAMD_HTTP_FLAG_HAS_HOST_HEADER)) {
 						rspamd_printf_fstring(buf,
 								"%s %s://%s:%d/%V HTTP/1.1\r\n"
 								"Connection: %s\r\n"
 								"Content-Length: %z\r\n",
 								http_method_str(msg->method),
-								(msg->flags & RSPAMD_HTTP_FLAG_SSL) ? "https" : "http",
+								(conn->opts & RSPAMD_HTTP_CLIENT_SSL) ? "https" : "http",
 								host,
 								msg->port,
 								msg->url,
@@ -1885,22 +1912,41 @@ rspamd_http_message_write_header (const gchar* mime_type, gboolean encrypted,
 								bodylen);
 					}
 					else {
-						rspamd_printf_fstring(buf,
+						if (rspamd_http_message_is_standard_port(msg)) {
+							rspamd_printf_fstring(buf,
 								"%s %s://%s:%d/%V HTTP/1.1\r\n"
 								"Connection: %s\r\n"
 								"Host: %s\r\n"
 								"Content-Length: %z\r\n",
 								http_method_str(msg->method),
-								(msg->flags & RSPAMD_HTTP_FLAG_SSL) ? "https" : "http",
+								(conn->opts & RSPAMD_HTTP_CLIENT_SSL) ? "https" : "http",
 								host,
 								msg->port,
 								msg->url,
 								conn_type,
 								host,
 								bodylen);
+						}
+						else {
+							rspamd_printf_fstring(buf,
+								"%s %s://%s:%d/%V HTTP/1.1\r\n"
+								"Connection: %s\r\n"
+								"Host: %s:%d\r\n"
+								"Content-Length: %z\r\n",
+								http_method_str(msg->method),
+								(conn->opts & RSPAMD_HTTP_CLIENT_SSL) ? "https" : "http",
+								host,
+								msg->port,
+								msg->url,
+								conn_type,
+								host,
+								msg->port,
+								bodylen);
+						}
 					}
 				}
 				else {
+					/* Unproxied version */
 					if ((msg->flags & RSPAMD_HTTP_FLAG_HAS_HOST_HEADER)) {
 						rspamd_printf_fstring(buf,
 								"%s %V HTTP/1.1\r\n"
@@ -1912,7 +1958,8 @@ rspamd_http_message_write_header (const gchar* mime_type, gboolean encrypted,
 								bodylen);
 					}
 					else {
-						rspamd_printf_fstring(buf,
+						if (rspamd_http_message_is_standard_port(msg)) {
+							rspamd_printf_fstring(buf,
 								"%s %V HTTP/1.1\r\n"
 								"Connection: %s\r\n"
 								"Host: %s\r\n"
@@ -1922,6 +1969,20 @@ rspamd_http_message_write_header (const gchar* mime_type, gboolean encrypted,
 								conn_type,
 								host,
 								bodylen);
+						}
+						else {
+							rspamd_printf_fstring(buf,
+								"%s %V HTTP/1.1\r\n"
+								"Connection: %s\r\n"
+								"Host: %s:%d\r\n"
+								"Content-Length: %z\r\n",
+								http_method_str(msg->method),
+								msg->url,
+								conn_type,
+								host,
+								msg->port,
+								bodylen);
+						}
 					}
 				}
 
@@ -1984,6 +2045,16 @@ rspamd_http_connection_write_message_common (struct rspamd_http_connection *conn
 	priv->buf->data = rspamd_fstring_sized_new (512);
 	buf = priv->buf->data;
 
+	if ((msg->flags & RSPAMD_HTTP_FLAG_WANT_SSL) && !(conn->opts & RSPAMD_HTTP_CLIENT_SSL)) {
+		err = g_error_new (HTTP_ERROR, 400,
+				"SSL connection requested but not created properly, internal error");
+		rspamd_http_connection_ref (conn);
+		conn->error_handler (conn, err);
+		rspamd_http_connection_unref (conn);
+		g_error_free (err);
+		return FALSE;
+	}
+
 	if (priv->peer_key && priv->local_key) {
 		priv->msg->peer_key = priv->peer_key;
 		priv->peer_key = NULL;
@@ -2280,14 +2351,19 @@ rspamd_http_connection_write_message_common (struct rspamd_http_connection *conn
 
 	priv->flags &= ~RSPAMD_HTTP_CONN_FLAG_RESETED;
 
-	if (priv->flags & RSPAMD_HTTP_CONN_FLAG_PROXY) {
+	if ((priv->flags & RSPAMD_HTTP_CONN_FLAG_PROXY) && (conn->opts & RSPAMD_HTTP_CLIENT_SSL)) {
 		/* We need to disable SSL flag! */
-		msg->flags &=~ RSPAMD_HTTP_FLAG_SSL;
+		err = g_error_new (HTTP_ERROR, 400, "cannot use proxy for SSL connections");
+		rspamd_http_connection_ref (conn);
+		conn->error_handler (conn, err);
+		rspamd_http_connection_unref (conn);
+		g_error_free (err);
+		return FALSE;
 	}
 
 	rspamd_ev_watcher_stop (priv->ctx->event_loop, &priv->ev);
 
-	if (msg->flags & RSPAMD_HTTP_FLAG_SSL) {
+	if (conn->opts & RSPAMD_HTTP_CLIENT_SSL) {
 		gpointer ssl_ctx = (msg->flags & RSPAMD_HTTP_FLAG_SSL_NOVERIFY) ?
 				priv->ctx->ssl_ctx_noverify : priv->ctx->ssl_ctx;
 
@@ -2301,29 +2377,34 @@ rspamd_http_connection_write_message_common (struct rspamd_http_connection *conn
 			return FALSE;
 		}
 		else {
-			if (priv->ssl) {
-				/* Cleanup the existing connection */
-				rspamd_ssl_connection_free (priv->ssl);
+			if (!priv->ssl) {
+				priv->ssl = rspamd_ssl_connection_new (ssl_ctx, priv->ctx->event_loop,
+						!(msg->flags & RSPAMD_HTTP_FLAG_SSL_NOVERIFY),
+						conn->log_tag);
+				g_assert (priv->ssl != NULL);
+
+				if (!rspamd_ssl_connect_fd (priv->ssl, conn->fd, host, &priv->ev,
+						priv->timeout, rspamd_http_event_handler,
+						rspamd_http_ssl_err_handler, conn)) {
+
+					err = g_error_new (HTTP_ERROR, 400,
+							"ssl connection error: ssl error=%s, errno=%s",
+							ERR_error_string (ERR_get_error (), NULL),
+							strerror (errno));
+					rspamd_http_connection_ref (conn);
+					conn->error_handler (conn, err);
+					rspamd_http_connection_unref (conn);
+					g_error_free (err);
+					return FALSE;
+				}
 			}
-
-			priv->ssl = rspamd_ssl_connection_new (ssl_ctx, priv->ctx->event_loop,
-					!(msg->flags & RSPAMD_HTTP_FLAG_SSL_NOVERIFY),
-					conn->log_tag);
-			g_assert (priv->ssl != NULL);
-
-			if (!rspamd_ssl_connect_fd (priv->ssl, conn->fd, host, &priv->ev,
-					priv->timeout, rspamd_http_event_handler,
-					rspamd_http_ssl_err_handler, conn)) {
-
-				err = g_error_new (HTTP_ERROR, 400,
-						"ssl connection error: ssl error=%s, errno=%s",
-						ERR_error_string (ERR_get_error (), NULL),
-						strerror (errno));
-				rspamd_http_connection_ref (conn);
-				conn->error_handler (conn, err);
-				rspamd_http_connection_unref (conn);
-				g_error_free (err);
-				return FALSE;
+			else {
+				/* Just restore SSL handlers */
+				rspamd_ssl_connection_restore_handlers (priv->ssl,
+						rspamd_http_event_handler,
+						rspamd_http_ssl_err_handler,
+						conn,
+						EV_WRITE);
 			}
 		}
 	}
