@@ -499,6 +499,10 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 			case 'f':
 			case 'F':
 				IF_HEADER (FROM_HEADER) {
+					if (hv_tok->len == 0) {
+						/* Replace '' with '<>' to fix parsing issue */
+						RSPAMD_FTOK_ASSIGN(hv_tok, "<>");
+					}
 					task->from_envelope = rspamd_email_address_from_smtp (
 							hv_tok->begin,
 							hv_tok->len);
@@ -625,7 +629,7 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 					 */
 					msg_debug_protocol ("read user header, value: %T", hv_tok);
 					if (!RSPAMD_TASK_IS_SPAMC (task)) {
-						task->user = rspamd_mempool_ftokdup (task->task_pool,
+						task->auth_user = rspamd_mempool_ftokdup (task->task_pool,
 								hv_tok);
 					}
 					else {
@@ -801,7 +805,7 @@ rspamd_protocol_control_parser_init (void)
 		rspamd_rcl_add_default_handler (sub,
 				"user",
 				rspamd_rcl_parse_struct_string,
-				G_STRUCT_OFFSET (struct rspamd_task, user),
+				G_STRUCT_OFFSET (struct rspamd_task, auth_user),
 				0,
 				NULL);
 		rspamd_rcl_add_default_handler (sub,
@@ -960,9 +964,9 @@ urls_protocol_cb (struct rspamd_url *url, struct tree_cb_data *cb)
 	ucl_array_append (cb->top, obj);
 
 	if (cb->task->cfg->log_urls) {
-		if (task->user) {
-			user_field = task->user;
-			len = strlen (task->user);
+		if (task->auth_user) {
+			user_field = task->auth_user;
+			len = strlen (task->auth_user);
 			has_user = TRUE;
 		}
 		else if (task->from_envelope) {
@@ -1207,14 +1211,20 @@ rspamd_scan_result_ucl (struct rspamd_task *task,
 		obj = top;
 	}
 
-	if (pr && pr->message && !(pr->flags & RSPAMD_PASSTHROUGH_NO_SMTP_MESSAGE)) {
-		/* Add smtp message if it does not exists: see #3269 for details */
-		if (ucl_object_lookup (task->messages, "smtp_message") == NULL) {
-			ucl_object_insert_key (task->messages,
-					ucl_object_fromstring_common (pr->message, 0, UCL_STRING_RAW),
+	if (pr) {
+		if (pr->message && !(pr->flags & RSPAMD_PASSTHROUGH_NO_SMTP_MESSAGE)) {
+			/* Add smtp message if it does not exist: see #3269 for details */
+			if (ucl_object_lookup(task->messages, "smtp_message") == NULL) {
+				ucl_object_insert_key(task->messages,
+					ucl_object_fromstring_common(pr->message, 0, UCL_STRING_RAW),
 					"smtp_message", 0,
 					false);
+			}
 		}
+
+		ucl_object_insert_key (obj,
+			ucl_object_fromstring (pr->module),
+			"passthrough_module", 0, false);
 	}
 
 	ucl_object_insert_key (obj,
@@ -1258,6 +1268,21 @@ rspamd_scan_result_ucl (struct rspamd_task *task,
 
 	/* Now handle symbols */
 	if (task->cmd != CMD_CHECK) {
+		/* Insert actions thresholds */
+		ucl_object_t *actions_obj = ucl_object_typed_new (UCL_OBJECT);
+
+		for (int i = task->result->nactions - 1; i >= 0; i --) {
+			struct rspamd_action_config *action_lim = &task->result->actions_config[i];
+
+			if (!isnan (action_lim->cur_limit) &&
+				!(action_lim->action->flags & (RSPAMD_ACTION_NO_THRESHOLD|RSPAMD_ACTION_HAM))) {
+				ucl_object_insert_key(actions_obj, ucl_object_fromdouble(action_lim->cur_limit),
+					action_lim->action->name, 0, true);
+			}
+		}
+
+		ucl_object_insert_key(obj, actions_obj, "thresholds", 0, false);
+
 		/* For checkv2 we insert symbols as a separate object */
 		obj = ucl_object_typed_new (UCL_OBJECT);
 	}
@@ -1508,12 +1533,16 @@ rspamd_protocol_write_ucl (struct rspamd_task *task,
 					if (task->protocol_flags & RSPAMD_TASK_PROTOCOL_FLAG_MILTER ||
 						!task->message) {
 
-						folded_header = rspamd_header_value_fold ("DKIM-Signature",
-								dkim_sig->str, 80, RSPAMD_TASK_NEWLINES_LF, NULL);
+						folded_header = rspamd_header_value_fold (
+								"DKIM-Signature", strlen ("DKIM-Signature"),
+								dkim_sig->str, dkim_sig->len,
+								80, RSPAMD_TASK_NEWLINES_LF, NULL);
 					}
 					else {
-						folded_header = rspamd_header_value_fold ("DKIM-Signature",
-								dkim_sig->str, 80,
+						folded_header = rspamd_header_value_fold (
+								"DKIM-Signature", strlen ("DKIM-Signature"),
+								dkim_sig->str, dkim_sig->len,
+								80,
 								MESSAGE_FIELD (task, nlines_type),
 								NULL);
 					}
@@ -1535,12 +1564,16 @@ rspamd_protocol_write_ucl (struct rspamd_task *task,
 				dkim_sig = (GString *) dkim_sigs->data;
 
 				if (task->protocol_flags & RSPAMD_TASK_PROTOCOL_FLAG_MILTER) {
-					folded_header = rspamd_header_value_fold ("DKIM-Signature",
-							dkim_sig->str, 80, RSPAMD_TASK_NEWLINES_LF, NULL);
+					folded_header = rspamd_header_value_fold (
+							"DKIM-Signature", strlen ("DKIM-Signature"),
+							dkim_sig->str, dkim_sig->len,
+							80, RSPAMD_TASK_NEWLINES_LF, NULL);
 				}
 				else {
-					folded_header = rspamd_header_value_fold ("DKIM-Signature",
-							dkim_sig->str, 80, MESSAGE_FIELD (task, nlines_type),
+					folded_header = rspamd_header_value_fold (
+							"DKIM-Signature", strlen ("DKIM-Signature"),
+							dkim_sig->str, dkim_sig->len,
+							80, MESSAGE_FIELD (task, nlines_type),
 							NULL);
 				}
 
@@ -1802,6 +1835,21 @@ end:
 		__atomic_add_fetch (&task->worker->srv->stat->messages_scanned,
 				1, __ATOMIC_RELEASE);
 #endif
+
+		/* Set average processing time */
+		guint32 slot;
+		float processing_time = task->time_real_finish - task->task_timestamp;
+
+#ifndef HAVE_ATOMIC_BUILTINS
+		slot = task->worker->srv->stat->avg_time.cur_slot++;
+#else
+		slot = __atomic_fetch_add (&task->worker->srv->stat->avg_time.cur_slot,
+				1, __ATOMIC_RELEASE);
+#endif
+		slot = slot % MAX_AVG_TIME_SLOTS;
+		/* TODO: this should be atomic but it is not supported in C */
+		task->worker->srv->stat->avg_time.avg_time[slot] = processing_time;
+
 	}
 }
 
