@@ -171,15 +171,6 @@ rspamd_task_fin (void *arg)
 	return FALSE;
 }
 
-/*
- * Called if session was restored inside fin callback
- */
-void
-rspamd_task_restore (void *arg)
-{
-	/* XXX: not needed now ? */
-}
-
 /*
  * Free all structures of worker_task
  */
@@ -309,8 +300,17 @@ rspamd_task_free (struct rspamd_task *task)
 		rspamd_message_unref (task->message);
 
 		if (task->flags & RSPAMD_TASK_FLAG_OWN_POOL) {
+			rspamd_mempool_destructors_enforce (task->task_pool);
+
+			if (task->symcache_runtime) {
+				rspamd_symcache_runtime_destroy (task);
+			}
+
 			rspamd_mempool_delete (task->task_pool);
 		}
+		else if (task->symcache_runtime) {
+			rspamd_symcache_runtime_destroy (task);
+		}
 	}
 }
 
@@ -991,7 +991,7 @@ rspamd_task_log_check_condition (struct rspamd_task *task,
 		}
 		break;
 	case RSPAMD_LOG_USER:
-		if (task->user) {
+		if (task->auth_user) {
 			ret = TRUE;
 		}
 		break;
@@ -1424,8 +1424,8 @@ rspamd_task_log_variable (struct rspamd_task *task,
 		}
 		break;
 	case RSPAMD_LOG_USER:
-		if (task->user) {
-			var.begin = task->user;
+		if (task->auth_user) {
+			var.begin = task->auth_user;
 			var.len = strlen (var.begin);
 		}
 		else {
@@ -1673,12 +1673,12 @@ rspamd_task_get_required_score (struct rspamd_task *task, struct rspamd_scan_res
 	}
 
 	for (i = m->nactions - 1; i >= 0; i --) {
-		struct rspamd_action_result *action_lim = &m->actions_limits[i];
+		struct rspamd_action_config *action_lim = &m->actions_config[i];
 
 
 		if (!isnan (action_lim->cur_limit) &&
 				!(action_lim->action->flags & (RSPAMD_ACTION_NO_THRESHOLD|RSPAMD_ACTION_HAM))) {
-			return m->actions_limits[i].cur_limit;
+			return m->actions_config[i].cur_limit;
 		}
 	}
 
@@ -1911,7 +1911,7 @@ rspamd_task_timeout (EV_P_ ev_timer *w, int revents)
 
 		ev_timer_again (EV_A_ w);
 		task->processed_stages |= RSPAMD_TASK_STAGE_FILTERS;
-		rspamd_session_cleanup (task->s);
+		rspamd_session_cleanup (task->s, true);
 		rspamd_task_process (task, RSPAMD_TASK_PROCESS_ALL);
 		rspamd_session_pending (task->s);
 	}
@@ -1940,7 +1940,7 @@ rspamd_task_timeout (EV_P_ ev_timer *w, int revents)
 
 		ev_timer_stop (EV_A_ w);
 		task->processed_stages |= RSPAMD_TASK_STAGE_DONE;
-		rspamd_session_cleanup (task->s);
+		rspamd_session_cleanup (task->s, true);
 		rspamd_task_process (task, RSPAMD_TASK_PROCESS_ALL);
 		rspamd_session_pending (task->s);
 	}
