@@ -57,11 +57,15 @@
 
 #ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
+#include <math.h>
+
 #endif
 
 #include "contrib/libev/ev.h"
 #include "libstat/stat_api.h"
 
+struct rspamd_worker *rspamd_current_worker = NULL;
+
 /* Forward declaration */
 static void rspamd_worker_heartbeat_start (struct rspamd_worker *,
 		struct ev_loop *);
@@ -486,10 +490,8 @@ rspamd_prepare_worker (struct rspamd_worker *worker, const char *name,
 	rspamd_worker_init_signals (worker, event_loop);
 	rspamd_control_worker_add_default_cmd_handlers (worker, event_loop);
 	rspamd_worker_heartbeat_start (worker, event_loop);
-#ifdef WITH_HIREDIS
 	rspamd_redis_pool_config (worker->srv->cfg->redis_pool,
 			worker->srv->cfg, event_loop);
-#endif
 
 	/* Accept all sockets */
 	if (hdl) {
@@ -684,7 +686,7 @@ rspamd_controller_send_ucl (struct rspamd_http_connection_entry *entry,
 static void
 rspamd_worker_drop_priv (struct rspamd_main *rspamd_main)
 {
-	if (rspamd_main->is_privilleged) {
+	if (rspamd_main->is_privileged) {
 		if (setgid (rspamd_main->workers_gid) == -1) {
 			msg_err_main ("cannot setgid to %d (%s), aborting",
 					(gint) rspamd_main->workers_gid,
@@ -960,8 +962,6 @@ rspamd_main_heartbeat_start (struct rspamd_worker *wrk, struct ev_loop *event_lo
 static bool
 rspamd_maybe_reuseport_socket (struct rspamd_worker_listen_socket *ls)
 {
-	gint nfd = -1;
-
 	if (ls->is_systemd) {
 		/* No need to reuseport */
 		return true;
@@ -978,6 +978,7 @@ rspamd_maybe_reuseport_socket (struct rspamd_worker_listen_socket *ls)
 	}
 
 #if defined(SO_REUSEPORT) && defined(SO_REUSEADDR) && defined(LINUX)
+	gint nfd = -1;
 
 	if (ls->type == RSPAMD_WORKER_SOCKET_UDP) {
 		nfd = rspamd_inet_address_listen (ls->addr,
@@ -1005,8 +1006,6 @@ rspamd_maybe_reuseport_socket (struct rspamd_worker_listen_socket *ls)
 		 */
 		nfd = ls->fd;
 	}
-#else
-	nfd = ls->fd;
 #endif
 
 #if 0
@@ -1176,8 +1175,14 @@ rspamd_handle_child_fork (struct rspamd_worker *wrk,
 	/* Close parent part of socketpair */
 	close (wrk->control_pipe[0]);
 	close (wrk->srv_pipe[0]);
+	/*
+	 * Read comments in `rspamd_handle_main_fork` for details why these channel
+	 * is blocking.
+	 */
 	rspamd_socket_nonblocking (wrk->control_pipe[1]);
+#if 0
 	rspamd_socket_nonblocking (wrk->srv_pipe[1]);
+#endif
 	rspamd_main->cfg->cur_worker = wrk;
 	/* Execute worker (this function should not return normally!) */
 	cf->worker->worker_start_func (wrk);
@@ -1195,8 +1200,22 @@ rspamd_handle_main_fork (struct rspamd_worker *wrk,
 	close (wrk->control_pipe[1]);
 	close (wrk->srv_pipe[1]);
 
-	rspamd_socket_nonblocking (wrk->control_pipe[0]);
+	/*
+	 * There are no reasons why control pipes are blocking: the messages
+	 * there are rare and are strictly bounded by command sizes, so if we block
+	 * on some pipe, it is ok, as we still poll that for all operations.
+	 * It is also impossible to block on writing in normal conditions.
+	 * And if the conditions are not normal, e.g. a worker is unresponsive, then
+	 * we can safely think that the non-blocking behaviour as it is implemented
+	 * currently will not make things better, as it would lead to incomplete
+	 * reads/writes that are not handled anyhow and are totally broken from the
+	 * beginning.
+	 */
+#if 0
 	rspamd_socket_nonblocking (wrk->srv_pipe[0]);
+#endif
+	rspamd_socket_nonblocking (wrk->control_pipe[0]);
+
 	rspamd_srv_start_watching (rspamd_main, wrk, ev_base);
 	/* Child event */
 	wrk->cld_ev.data = wrk;
@@ -1229,6 +1248,9 @@ rspamd_handle_main_fork (struct rspamd_worker *wrk,
 #endif
 }
 
+#ifndef SOCK_SEQPACKET
+#define SOCK_SEQPACKET SOCK_DGRAM
+#endif
 struct rspamd_worker *
 rspamd_fork_worker (struct rspamd_main *rspamd_main,
 					struct rspamd_worker_conf *cf,
@@ -1242,12 +1264,12 @@ rspamd_fork_worker (struct rspamd_main *rspamd_main,
 	/* Starting worker process */
 	wrk = (struct rspamd_worker *) g_malloc0 (sizeof (struct rspamd_worker));
 
-	if (!rspamd_socketpair (wrk->control_pipe, SOCK_DGRAM)) {
+	if (!rspamd_socketpair (wrk->control_pipe, SOCK_SEQPACKET)) {
 		msg_err ("socketpair failure: %s", strerror (errno));
 		rspamd_hard_terminate (rspamd_main);
 	}
 
-	if (!rspamd_socketpair (wrk->srv_pipe, SOCK_DGRAM)) {
+	if (!rspamd_socketpair (wrk->srv_pipe, SOCK_SEQPACKET)) {
 		msg_err ("socketpair failure: %s", strerror (errno));
 		rspamd_hard_terminate (rspamd_main);
 	}
@@ -1279,6 +1301,7 @@ rspamd_fork_worker (struct rspamd_main *rspamd_main,
 
 	switch (wrk->pid) {
 	case 0:
+		rspamd_current_worker = wrk;
 		rspamd_handle_child_fork (wrk, rspamd_main, cf, listen_sockets);
 		break;
 	case -1:
@@ -1383,6 +1406,70 @@ rspamd_worker_is_primary_controller (struct rspamd_worker *w)
 	return FALSE;
 }
 
+gboolean
+rspamd_worker_check_controller_presence (struct rspamd_worker *w)
+{
+	if (w->index == 0) {
+		GQuark our_type = w->type;
+		gboolean controller_seen = FALSE;
+		GList *cur;
+
+		enum {
+			low_priority_worker,
+			high_priority_worker
+		} our_priority;
+
+		if (our_type == g_quark_from_static_string("rspamd_proxy")) {
+			our_priority = low_priority_worker;
+		}
+		else if (our_type == g_quark_from_static_string("normal")) {
+			our_priority = high_priority_worker;
+		}
+		else {
+			msg_err ("function is called for a wrong worker type: %s", g_quark_to_string(our_type));
+			return FALSE;
+		}
+
+		cur = w->srv->cfg->workers;
+
+		while (cur) {
+			struct rspamd_worker_conf *cf;
+
+			cf = (struct rspamd_worker_conf *)cur->data;
+
+			if (our_priority == low_priority_worker) {
+				if ((cf->type == g_quark_from_static_string("controller")) ||
+					(cf->type == g_quark_from_static_string("normal"))) {
+
+					if (cf->enabled && cf->count >= 0) {
+						controller_seen = TRUE;
+						break;
+					}
+				}
+			}
+			else {
+				if (cf->type == g_quark_from_static_string("controller")) {
+					if (cf->enabled && cf->count >= 0) {
+						controller_seen = TRUE;
+						break;
+					}
+				}
+			}
+
+			cur = g_list_next (cur);
+		}
+
+		if (!controller_seen) {
+			msg_info ("no controller or normal workers defined, execute "
+					  "controller periodics in this worker");
+			w->flags |= RSPAMD_WORKER_CONTROLLER;
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
 struct rspamd_worker_session_elt {
 	void *ptr;
 	guint *pref;
@@ -1642,12 +1729,6 @@ rspamd_set_crash_handler (struct rspamd_main *rspamd_main)
 	stack_t ss;
 	memset (&ss, 0, sizeof ss);
 
-	/*
-	 * Allocate special stack, NOT freed at the end so far
-	 * It also cannot be on stack as this memory is used when
-	 * stack corruption is detected. Leak sanitizer blames about it but
-	 * I don't know any good ways to stop this behaviour.
-	 */
 	ss.ss_size = MAX (SIGSTKSZ, 8192 * 4);
 	stack_mem = g_malloc0 (ss.ss_size);
 	ss.ss_sp = stack_mem;
@@ -1665,6 +1746,28 @@ rspamd_set_crash_handler (struct rspamd_main *rspamd_main)
 #endif
 }
 
+RSPAMD_NO_SANITIZE void rspamd_unset_crash_handler (struct rspamd_main *unused_)
+{
+#ifdef HAVE_SIGALTSTACK
+	int ret;
+	stack_t ss;
+	ret = sigaltstack (NULL, &ss);
+
+	if (ret != -1) {
+		if (ss.ss_size > 0 && ss.ss_sp) {
+			g_free(ss.ss_sp);
+		}
+
+		ss.ss_size = 0;
+		ss.ss_sp = NULL;
+#ifdef SS_DISABLE
+		ss.ss_flags |= SS_DISABLE;
+#endif
+		sigaltstack(&ss, NULL);
+	}
+#endif
+}
+
 static void
 rspamd_enable_accept_event (EV_P_ ev_timer *w, int revents)
 {
@@ -2223,3 +2326,49 @@ rspamd_worker_init_controller (struct rspamd_worker *worker,
 				ctx->resolver, worker, RSPAMD_MAP_WATCH_SCANNER);
 	}
 }
+
+gdouble
+rspamd_worker_check_and_adjust_timeout (struct rspamd_config *cfg, gdouble timeout)
+{
+	if (isnan (timeout)) {
+		/* Use implicit timeout from cfg->task_timeout */
+		timeout = cfg->task_timeout;
+	}
+
+	if (isnan (timeout)) {
+		return timeout;
+	}
+
+	struct rspamd_symcache_timeout_result *tres = rspamd_symcache_get_max_timeout (cfg->cache);
+	g_assert (tres != 0);
+
+	if (tres->max_timeout > timeout) {
+		msg_info_config("configured task_timeout %.2f is less than maximum symbols cache timeout %.2f; "
+						"some symbols can be terminated before checks", timeout, tres->max_timeout);
+		GString *buf = g_string_sized_new(512);
+		static const int max_displayed_items = 12;
+
+		for (int i = 0; i < MIN(tres->nitems, max_displayed_items); i++) {
+			if (i == 0) {
+				rspamd_printf_gstring(buf, "%s(%.2f)",
+						rspamd_symcache_item_name((struct rspamd_symcache_item *)tres->items[i].item),
+						tres->items[i].timeout);
+			}
+			else {
+				rspamd_printf_gstring(buf, "; %s(%.2f)",
+						rspamd_symcache_item_name((struct rspamd_symcache_item *)tres->items[i].item),
+						tres->items[i].timeout);
+			}
+		}
+		msg_info_config("list of top %d symbols by execution time: %v",
+				(int)MIN(tres->nitems, max_displayed_items),
+				buf);
+
+		g_string_free(buf, TRUE);
+	}
+
+	rspamd_symcache_timeout_result_free (tres);
+
+	/* TODO: maybe adjust timeout */
+	return timeout;
+}
