@@ -36,6 +36,8 @@ typedef void (*rspamd_sig_handler_t) (gint, siginfo_t *, void *);
 struct rspamd_worker;
 struct rspamd_worker_signal_handler;
 
+extern struct rspamd_worker *rspamd_current_worker;
+
 /**
  * Init basic signals for a worker
  * @param worker
@@ -88,7 +90,7 @@ typedef gint (*rspamd_controller_func_t) (
 struct rspamd_custom_controller_command {
 	const gchar *command;
 	struct module_ctx *ctx;
-	gboolean privilleged;
+	gboolean privileged;
 	gboolean require_message;
 	rspamd_controller_func_t handler;
 };
@@ -106,7 +108,7 @@ struct rspamd_controller_session {
 	struct rspamd_config *cfg;
 	struct rspamd_lang_detector *lang_det;
 	gboolean is_spam;
-	gboolean is_enable;
+	gboolean is_read_only;
 };
 
 /**
@@ -173,6 +175,15 @@ void rspamd_hard_terminate (struct rspamd_main *rspamd_main) G_GNUC_NORETURN;
  */
 gboolean rspamd_worker_is_scanner (struct rspamd_worker *w);
 
+/**
+ * Checks
+ * @param cfg
+ * @param timeout
+ * @return
+ */
+gdouble rspamd_worker_check_and_adjust_timeout (struct rspamd_config *cfg,
+		gdouble timeout);
+
 /**
  * Returns TRUE if a specific worker is a primary controller
  * @param w
@@ -180,6 +191,11 @@ gboolean rspamd_worker_is_scanner (struct rspamd_worker *w);
  */
 gboolean rspamd_worker_is_primary_controller (struct rspamd_worker *w);
 
+/**
+ * Returns TRUE if a specific worker should take a role of a controller
+ */
+gboolean rspamd_worker_check_controller_presence (struct rspamd_worker *w);
+
 /**
  * Creates new session cache
  * @param w
@@ -219,6 +235,11 @@ struct rspamd_worker *rspamd_fork_worker (struct rspamd_main *,
  */
 RSPAMD_NO_SANITIZE void rspamd_set_crash_handler (struct rspamd_main *);
 
+/**
+ * Restore memory for crash signals
+ */
+RSPAMD_NO_SANITIZE void rspamd_unset_crash_handler (struct rspamd_main *);
+
 /**
  * Initialise the main monitoring worker
  * @param worker
