@@ -23,8 +23,11 @@
 #include "logger.h"
 #include "unix-std.h"
 #include "hs.h"
+#include "libserver/hyperscan_tools.h"
 #endif
 #include "acism.h"
+#include "libutil/regexp.h"
+#include <stdalign.h>
 
 #define MAX_SCRATCH 4
 
@@ -38,18 +41,19 @@ static const char *hs_cache_dir = NULL;
 static enum rspamd_hs_check_state hs_suitable_cpu = RSPAMD_HS_UNCHECKED;
 
 
-struct rspamd_multipattern {
+struct RSPAMD_ALIGNED(64) rspamd_multipattern {
 #ifdef WITH_HYPERSCAN
-	hs_database_t *db;
+	rspamd_cryptobox_hash_state_t hash_state;
+	rspamd_hyperscan_t *hs_db;
 	hs_scratch_t *scratch[MAX_SCRATCH];
 	GArray *hs_pats;
 	GArray *hs_ids;
 	GArray *hs_flags;
-	rspamd_cryptobox_hash_state_t hash_state;
 	guint scratch_used;
 #endif
 	ac_trie_t *t;
 	GArray *pats;
+	GArray *res;
 
 	gboolean compiled;
 	guint cnt;
@@ -94,18 +98,17 @@ rspamd_multipattern_escape_tld_hyperscan (const gchar *pattern, gsize slen,
 		gsize *dst_len)
 {
 	gsize len;
-	const gchar *p, *prefix;
+	const gchar *p, *prefix, *suffix;
 	gchar *res;
 
 	/*
 	 * We understand the following cases
-	 * 1) blah -> \\.blah
-	 * 2) *.blah -> \\..*\\.blah
+	 * 1) blah -> .blah\b
+	 * 2) *.blah -> ..*\\.blah\b|$
 	 * 3) ???
 	 */
 
 	if (pattern[0] == '*') {
-		len = slen + 4;
 		p = strchr (pattern, '.');
 
 		if (p == NULL) {
@@ -116,17 +119,22 @@ rspamd_multipattern_escape_tld_hyperscan (const gchar *pattern, gsize slen,
 			p ++;
 		}
 
-		prefix = "\\..*\\.";
+		prefix = "\\.";
+		len = slen + strlen (prefix);
 	}
 	else {
-		len = slen + 2;
 		prefix = "\\.";
 		p = pattern;
+		len = slen + strlen (prefix);
 	}
 
+	suffix = "(:?\\b|$)";
+	len += strlen (suffix);
+
 	res = g_malloc (len + 1);
 	slen = rspamd_strlcpy (res, prefix, len + 1);
 	slen += rspamd_strlcpy (res + slen, p, len + 1 - slen);
+	slen += rspamd_strlcpy (res + slen, suffix, len + 1 - slen);
 
 	*dst_len = slen;
 
@@ -191,20 +199,33 @@ rspamd_multipattern_pattern_filter (const gchar *pattern, gsize len,
 		gsize *dst_len)
 {
 	gchar *ret = NULL;
+	gint gl_flags = RSPAMD_REGEXP_ESCAPE_ASCII;
+
+	if (flags & RSPAMD_MULTIPATTERN_UTF8) {
+		gl_flags |= RSPAMD_REGEXP_ESCAPE_UTF;
+	}
+
 #ifdef WITH_HYPERSCAN
 	if (rspamd_hs_check ()) {
 		if (flags & RSPAMD_MULTIPATTERN_TLD) {
-			ret = rspamd_multipattern_escape_tld_hyperscan (pattern, len, dst_len);
+			gchar *tmp;
+			gsize tlen;
+			tmp = rspamd_multipattern_escape_tld_hyperscan (pattern, len, &tlen);
+
+			ret = rspamd_str_regexp_escape (tmp, tlen, dst_len,
+					gl_flags|RSPAMD_REGEXP_ESCAPE_RE);
+			g_free (tmp);
 		}
 		else if (flags & RSPAMD_MULTIPATTERN_RE) {
-			ret = malloc (len + 1);
-			*dst_len = rspamd_strlcpy (ret, pattern, len + 1);
+			ret = rspamd_str_regexp_escape (pattern, len, dst_len, gl_flags |
+					RSPAMD_REGEXP_ESCAPE_RE);
 		}
 		else if (flags & RSPAMD_MULTIPATTERN_GLOB) {
-			ret = rspamd_str_regexp_escape (pattern, len, dst_len, TRUE);
+			ret = rspamd_str_regexp_escape (pattern, len, dst_len,
+					gl_flags | RSPAMD_REGEXP_ESCAPE_GLOB);
 		}
 		else {
-			ret = rspamd_str_regexp_escape (pattern, len, dst_len, FALSE);
+			ret = rspamd_str_regexp_escape (pattern, len, dst_len, gl_flags);
 		}
 
 		return ret;
@@ -214,6 +235,14 @@ rspamd_multipattern_pattern_filter (const gchar *pattern, gsize len,
 	if (flags & RSPAMD_MULTIPATTERN_TLD) {
 		ret = rspamd_multipattern_escape_tld_acism (pattern, len, dst_len);
 	}
+	else if (flags & RSPAMD_MULTIPATTERN_RE) {
+		ret = rspamd_str_regexp_escape (pattern, len, dst_len, gl_flags |
+															   RSPAMD_REGEXP_ESCAPE_RE);
+	}
+	else if (flags & RSPAMD_MULTIPATTERN_GLOB) {
+		ret = rspamd_str_regexp_escape (pattern, len, dst_len,
+				gl_flags | RSPAMD_REGEXP_ESCAPE_GLOB);
+	}
 	else {
 		ret = malloc (len + 1);
 		*dst_len = rspamd_strlcpy (ret, pattern, len + 1);
@@ -227,7 +256,11 @@ rspamd_multipattern_create (enum rspamd_multipattern_flags flags)
 {
 	struct rspamd_multipattern *mp;
 
-	mp = g_malloc0 (sizeof (*mp));
+	/* Align due to blake2b state */
+	(void) !posix_memalign((void **)&mp, RSPAMD_ALIGNOF(struct rspamd_multipattern),
+			sizeof (*mp));
+	g_assert (mp != NULL);
+	memset (mp, 0, sizeof (*mp));
 	mp->flags = flags;
 
 #ifdef WITH_HYPERSCAN
@@ -252,7 +285,10 @@ rspamd_multipattern_create_sized (guint npatterns,
 {
 	struct rspamd_multipattern *mp;
 
-	mp = g_malloc0 (sizeof (*mp));
+	/* Align due to blake2b state */
+	(void) !posix_memalign((void **)&mp, RSPAMD_ALIGNOF(struct rspamd_multipattern), sizeof (*mp));
+	g_assert (mp != NULL);
+	memset (mp, 0, sizeof (*mp));
 	mp->flags = flags;
 
 #ifdef WITH_HYPERSCAN
@@ -294,12 +330,28 @@ rspamd_multipattern_add_pattern_len (struct rspamd_multipattern *mp,
 	if (rspamd_hs_check ()) {
 		gchar *np;
 		gint fl = HS_FLAG_SOM_LEFTMOST;
+		gint adjusted_flags = mp->flags | flags;
 
-		if (mp->flags & RSPAMD_MULTIPATTERN_ICASE) {
+		if (adjusted_flags & RSPAMD_MULTIPATTERN_ICASE) {
 			fl |= HS_FLAG_CASELESS;
 		}
-		if (mp->flags & RSPAMD_MULTIPATTERN_UTF8) {
-			fl |= HS_FLAG_UTF8|HS_FLAG_UCP;
+		if (adjusted_flags & RSPAMD_MULTIPATTERN_UTF8) {
+			if (adjusted_flags & RSPAMD_MULTIPATTERN_TLD) {
+				fl |= HS_FLAG_UTF8;
+			}
+			else {
+				fl |= HS_FLAG_UTF8 | HS_FLAG_UCP;
+			}
+		}
+		if (adjusted_flags & RSPAMD_MULTIPATTERN_DOTALL) {
+			fl |= HS_FLAG_DOTALL;
+		}
+		if (adjusted_flags & RSPAMD_MULTIPATTERN_SINGLEMATCH) {
+			fl |= HS_FLAG_SINGLEMATCH;
+			fl &= ~HS_FLAG_SOM_LEFTMOST; /* According to hyperscan docs */
+		}
+		if (adjusted_flags & RSPAMD_MULTIPATTERN_NO_START) {
+			fl &= ~HS_FLAG_SOM_LEFTMOST;
 		}
 
 		g_array_append_val (mp->hs_flags, fl);
@@ -349,8 +401,6 @@ rspamd_multipattern_try_load_hs (struct rspamd_multipattern *mp,
 		const guchar *hash)
 {
 	gchar fp[PATH_MAX];
-	gpointer map;
-	gsize len;
 
 	if (hs_cache_dir == NULL) {
 		return FALSE;
@@ -358,19 +408,9 @@ rspamd_multipattern_try_load_hs (struct rspamd_multipattern *mp,
 
 	rspamd_snprintf (fp, sizeof (fp), "%s/%*xs.hsmp", hs_cache_dir,
 			(gint)rspamd_cryptobox_HASHBYTES / 2, hash);
+	mp->hs_db = rspamd_hyperscan_maybe_load(fp, 0);
 
-	if ((map = rspamd_file_xmap (fp, PROT_READ, &len, TRUE)) != NULL) {
-		if (hs_deserialize_database (map, len, &mp->db) == HS_SUCCESS) {
-			munmap (map, len);
-			return TRUE;
-		}
-
-		munmap (map, len);
-		/* Remove stale file */
-		(void)unlink (fp);
-	}
-
-	return FALSE;
+	return mp->hs_db != NULL;
 }
 
 static void
@@ -386,11 +426,12 @@ rspamd_multipattern_try_save_hs (struct rspamd_multipattern *mp,
 		return;
 	}
 
-	rspamd_snprintf (fp, sizeof (fp), "%s/%*xs.hsmp.tmp", hs_cache_dir,
-			(gint)rspamd_cryptobox_HASHBYTES / 2, hash);
+	rspamd_snprintf (fp, sizeof (fp), "%s%shsmp-XXXXXXXXXXXXX", G_DIR_SEPARATOR_S,
+		hs_cache_dir);
 
-	if ((fd = rspamd_file_xopen (fp, O_WRONLY | O_CREAT | O_EXCL, 00644, 0)) != -1) {
-		if (hs_serialize_database (mp->db, &bytes, &len) == HS_SUCCESS) {
+	if ((fd = g_mkstemp_full(fp, O_CREAT|O_EXCL|O_WRONLY, 00644)) != -1) {
+		int ret;
+		if ((ret = hs_serialize_database (rspamd_hyperscan_get_database(mp->hs_db), &bytes, &len)) == HS_SUCCESS) {
 			if (write (fd, bytes, len) == -1) {
 				msg_warn ("cannot write hyperscan cache to %s: %s",
 						fp, strerror (errno));
@@ -409,17 +450,23 @@ rspamd_multipattern_try_save_hs (struct rspamd_multipattern *mp,
 							fp, np, strerror (errno));
 					unlink (fp);
 				}
+				else {
+					rspamd_hyperscan_notice_known(np);
+				}
 			}
 		}
 		else {
-			msg_warn ("cannot serialize hyperscan cache to %s: %s",
-					fp, strerror (errno));
+			msg_warn ("cannot serialize hyperscan cache to %s: error code %d",
+					fp, ret);
 			unlink (fp);
 		}
 
 
 		close (fd);
 	}
+	else {
+		msg_warn ("cannot open a temp file %s to write hyperscan cache: %s", fp, strerror(errno));
+	}
 }
 #endif
 
@@ -442,13 +489,15 @@ rspamd_multipattern_compile (struct rspamd_multipattern *mp, GError **err)
 			rspamd_cryptobox_hash_final (&mp->hash_state, hash);
 
 			if (!rspamd_multipattern_try_load_hs (mp, hash)) {
+				hs_database_t *db = NULL;
+
 				if (hs_compile_multi ((const char *const *)mp->hs_pats->data,
 						(const unsigned int *)mp->hs_flags->data,
 						(const unsigned int *)mp->hs_ids->data,
 						mp->cnt,
 						HS_MODE_BLOCK,
 						&plt,
-						&mp->db,
+						&db,
 						&hs_errors) != HS_SUCCESS) {
 
 					g_set_error (err, rspamd_multipattern_quark (), EINVAL,
@@ -459,12 +508,45 @@ rspamd_multipattern_compile (struct rspamd_multipattern *mp, GError **err)
 
 					return FALSE;
 				}
+
+				if (hs_cache_dir != NULL) {
+					char fpath[PATH_MAX];
+					rspamd_snprintf (fpath, sizeof (fpath), "%s/%*xs.hsmp", hs_cache_dir,
+						(gint)rspamd_cryptobox_HASHBYTES / 2, hash);
+					mp->hs_db = rspamd_hyperscan_from_raw_db(db, fpath);
+				}
+				else {
+					/* Should not happen in the real life */
+					mp->hs_db = rspamd_hyperscan_from_raw_db(db, NULL);
+				}
+
+				rspamd_multipattern_try_save_hs (mp, hash);
 			}
 
-			rspamd_multipattern_try_save_hs (mp, hash);
+			for (i = 0; i < MAX_SCRATCH; i ++) {
+				mp->scratch[i] = NULL;
+			}
 
 			for (i = 0; i < MAX_SCRATCH; i ++) {
-				g_assert (hs_alloc_scratch (mp->db, &mp->scratch[i]) == HS_SUCCESS);
+				int ret;
+
+				if ((ret = hs_alloc_scratch (rspamd_hyperscan_get_database(mp->hs_db), &mp->scratch[i])) != HS_SUCCESS) {
+					msg_err("cannot allocate scratch space for hyperscan: error code %d", ret);
+
+					/* Clean all scratches that are non-NULL */
+					for (int ii = 0; ii < MAX_SCRATCH; ii ++) {
+						if (mp->scratch[ii] != NULL) {
+							hs_free_scratch(mp->scratch[ii]);
+						}
+					}
+					g_set_error (err, rspamd_multipattern_quark (), EINVAL,
+						"cannot allocate scratch space for hyperscan: error code %d", ret);
+
+					rspamd_hyperscan_free(mp->hs_db, true);
+					mp->hs_db = NULL;
+
+					return FALSE;
+				}
 			}
 		}
 
@@ -475,7 +557,34 @@ rspamd_multipattern_compile (struct rspamd_multipattern *mp, GError **err)
 #endif
 
 	if (mp->cnt > 0) {
-		mp->t = acism_create ((const ac_trie_pat_t *)mp->pats->data, mp->cnt);
+
+		if (mp->flags & (RSPAMD_MULTIPATTERN_GLOB|RSPAMD_MULTIPATTERN_RE)) {
+			/* Fallback to pcre... */
+			rspamd_regexp_t *re;
+			mp->res = g_array_sized_new (FALSE, TRUE,
+					sizeof (rspamd_regexp_t *), mp->cnt);
+
+			for (guint i = 0; i < mp->cnt; i ++) {
+				const ac_trie_pat_t *pat;
+				const gchar *pat_flags = NULL;
+
+				if (mp->flags & RSPAMD_MULTIPATTERN_UTF8) {
+					pat_flags = "u";
+				}
+
+				pat = &g_array_index (mp->pats, ac_trie_pat_t, i);
+				re = rspamd_regexp_new (pat->ptr, pat_flags, err);
+
+				if (re == NULL) {
+					return FALSE;
+				}
+
+				g_array_append_val (mp->res, re);
+			}
+		}
+		else {
+			mp->t = acism_create ((const ac_trie_pat_t *) mp->pats->data, mp->cnt);
+		}
 	}
 
 	mp->compiled = TRUE;
@@ -547,7 +656,7 @@ rspamd_multipattern_lookup (struct rspamd_multipattern *mp,
 
 	g_assert (mp != NULL);
 
-	if (mp->cnt == 0 || !mp->compiled) {
+	if (mp->cnt == 0 || !mp->compiled || len == 0) {
 		return 0;
 	}
 
@@ -574,7 +683,7 @@ rspamd_multipattern_lookup (struct rspamd_multipattern *mp,
 
 		g_assert (scr != NULL);
 
-		ret = hs_scan (mp->db, in, len, 0, scr,
+		ret = hs_scan (rspamd_hyperscan_get_database(mp->hs_db), in, len, 0, scr,
 				rspamd_multipattern_hs_cb, &cbd);
 
 		mp->scratch_used &= ~(1 << i);
@@ -596,11 +705,39 @@ rspamd_multipattern_lookup (struct rspamd_multipattern *mp,
 
 	gint state = 0;
 
-	ret = acism_lookup (mp->t, in, len, rspamd_multipattern_acism_cb, &cbd,
-			&state, mp->flags & RSPAMD_MULTIPATTERN_ICASE);
+	if (mp->flags & (RSPAMD_MULTIPATTERN_GLOB|RSPAMD_MULTIPATTERN_RE)) {
+		/* Terribly inefficient, but who cares - just use hyperscan */
+		for (guint i = 0; i < mp->cnt; i ++) {
+			rspamd_regexp_t *re = g_array_index (mp->res, rspamd_regexp_t *, i);
+			const gchar *start = NULL, *end = NULL;
 
-	if (pnfound) {
-		*pnfound = cbd.nfound;
+			while (rspamd_regexp_search (re,
+					in,
+					len,
+					&start,
+					&end,
+					TRUE,
+					NULL)) {
+				if (rspamd_multipattern_acism_cb (i, end - in, &cbd)) {
+					goto out;
+				}
+			}
+		}
+out:
+		ret = cbd.ret;
+
+		if (pnfound) {
+			*pnfound = cbd.nfound;
+		}
+	}
+	else {
+		/* Plain trie */
+		ret = acism_lookup (mp->t, in, len, rspamd_multipattern_acism_cb, &cbd,
+				&state, mp->flags & RSPAMD_MULTIPATTERN_ICASE);
+
+		if (pnfound) {
+			*pnfound = cbd.nfound;
+		}
 	}
 
 	return ret;
@@ -622,7 +759,9 @@ rspamd_multipattern_destroy (struct rspamd_multipattern *mp)
 					hs_free_scratch (mp->scratch[i]);
 				}
 
-				hs_free_database (mp->db);
+				if (mp->hs_db) {
+					rspamd_hyperscan_free(mp->hs_db, false);
+				}
 			}
 
 			for (i = 0; i < mp->cnt; i ++) {
@@ -633,7 +772,7 @@ rspamd_multipattern_destroy (struct rspamd_multipattern *mp)
 			g_array_free (mp->hs_pats, TRUE);
 			g_array_free (mp->hs_ids, TRUE);
 			g_array_free (mp->hs_flags, TRUE);
-			g_free (mp);
+			free (mp); /* Due to posix_memalign */
 
 			return;
 		}
@@ -687,4 +826,4 @@ gboolean
 rspamd_multipattern_has_hyperscan (void)
 {
 	return rspamd_hs_check ();
-}
\ No newline at end of file
+}
