@@ -19,6 +19,7 @@
 #include "ref.h"
 #include "util.h"
 #include "rspamd.h"
+#include "contrib/fastutf8/fastutf8.h"
 
 #ifndef WITH_PCRE2
 /* Normal pcre path */
@@ -65,11 +66,11 @@ struct rspamd_regexp_s {
 	gpointer ud;
 	gpointer re_class;
 	guint64 cache_id;
+	gsize match_limit;
 	guint max_hits;
 	gint flags;
 	gint pcre_flags;
 	gint ncaptures;
-	gint nbackref;
 };
 
 struct rspamd_regexp_cache {
@@ -82,6 +83,7 @@ struct rspamd_regexp_cache {
 static struct rspamd_regexp_cache *global_re_cache = NULL;
 static gboolean can_jit = FALSE;
 static gboolean check_jit = TRUE;
+static const int max_re_cache_size = 8192;
 
 #ifdef WITH_PCRE2
 static pcre2_compile_context *pcre2_ctx = NULL;
@@ -115,28 +117,33 @@ rspamd_regexp_dtor (rspamd_regexp_t *re)
 	if (re) {
 		if (re->raw_re && re->raw_re != re->re) {
 #ifndef WITH_PCRE2
-#ifdef HAVE_PCRE_JIT
+			/* PCRE1 version */
+# ifdef HAVE_PCRE_JIT
 			if (re->raw_extra) {
 				pcre_free_study (re->raw_extra);
 			}
-#endif
+# endif
 #else
-			if (re->mcontext) {
-				pcre2_match_context_free (re->mcontext);
+			/* PCRE 2 version */
+			if (re->raw_mcontext) {
+				pcre2_match_context_free (re->raw_mcontext);
 			}
 #endif
 			PCRE_FREE (re->raw_re);
 		}
+
 		if (re->re) {
 #ifndef WITH_PCRE2
-#ifdef HAVE_PCRE_JIT
+			/* PCRE1 version */
+# ifdef HAVE_PCRE_JIT
 			if (re->extra) {
 				pcre_free_study (re->extra);
 			}
-#endif
+# endif
 #else
-			if (re->raw_mcontext) {
-				pcre2_match_context_free (re->raw_mcontext);
+			/* PCRE 2 version */
+			if (re->mcontext) {
+				pcre2_match_context_free (re->mcontext);
 			}
 #endif
 			PCRE_FREE (re->re);
@@ -158,35 +165,51 @@ rspamd_regexp_post_process (rspamd_regexp_t *r)
 	}
 #if defined(WITH_PCRE2)
 	gsize jsz;
-	guint jit_flags = PCRE2_JIT_COMPLETE;
+	static const guint max_recursion_depth = 100000, max_backtrack = 1000000;
+
+	guint jit_flags = can_jit ? PCRE2_JIT_COMPLETE : 0;
+
 	/* Create match context */
-
 	r->mcontext = pcre2_match_context_create (NULL);
+	g_assert (r->mcontext != NULL);
+	pcre2_set_recursion_limit (r->mcontext, max_recursion_depth);
+	pcre2_set_match_limit (r->mcontext, max_backtrack);
 
-	if (r->re != r->raw_re) {
+	if (r->raw_re && r->re != r->raw_re) {
 		r->raw_mcontext = pcre2_match_context_create (NULL);
+		g_assert (r->raw_mcontext != NULL);
+		pcre2_set_recursion_limit (r->raw_mcontext, max_recursion_depth);
+		pcre2_set_match_limit (r->raw_mcontext, max_backtrack);
+	}
+	else if (r->raw_re) {
+		r->raw_mcontext = r->mcontext;
 	}
 	else {
-		r->raw_mcontext = r->mcontext;
+		r->raw_mcontext = NULL;
 	}
 
 #ifdef HAVE_PCRE_JIT
-	if (pcre2_jit_compile (r->re, jit_flags) < 0) {
-		msg_err ("jit compilation of %s is not supported: %d", r->pattern, jit_flags);
-		r->flags |= RSPAMD_REGEXP_FLAG_DISABLE_JIT;
-	}
-	else {
-		if (!(pcre2_pattern_info (r->re, PCRE2_INFO_JITSIZE, &jsz) >= 0 && jsz > 0)) {
-			msg_err ("jit compilation of %s is not supported", r->pattern);
+	if (can_jit) {
+		if (pcre2_jit_compile (r->re, jit_flags) < 0) {
+			msg_err ("jit compilation of %s is not supported: %d", r->pattern, jit_flags);
 			r->flags |= RSPAMD_REGEXP_FLAG_DISABLE_JIT;
 		}
+		else {
+			if (!(pcre2_pattern_info (r->re, PCRE2_INFO_JITSIZE, &jsz) >= 0 && jsz > 0)) {
+				msg_err ("jit compilation of %s is not supported", r->pattern);
+				r->flags |= RSPAMD_REGEXP_FLAG_DISABLE_JIT;
+			}
+		}
+	}
+	else {
+		r->flags |= RSPAMD_REGEXP_FLAG_DISABLE_JIT;
 	}
 
 	if (!(r->flags & RSPAMD_REGEXP_FLAG_DISABLE_JIT)) {
 		pcre2_jit_stack_assign (r->mcontext, NULL, global_re_cache->jstack);
 	}
 
-	if (r->re != r->raw_re) {
+	if (r->raw_re && r->re != r->raw_re && !(r->flags & RSPAMD_REGEXP_FLAG_DISABLE_JIT)) {
 		if (pcre2_jit_compile (r->raw_re, jit_flags) < 0) {
 			msg_debug ("jit compilation of %s is not supported", r->pattern);
 			r->flags |= RSPAMD_REGEXP_FLAG_DISABLE_JIT;
@@ -196,6 +219,7 @@ rspamd_regexp_post_process (rspamd_regexp_t *r)
 			msg_debug ("jit compilation of raw %s is not supported", r->pattern);
 		}
 		else if (!(r->flags & RSPAMD_REGEXP_FLAG_DISABLE_JIT)) {
+			g_assert (r->raw_mcontext != NULL);
 			pcre2_jit_stack_assign (r->raw_mcontext, NULL, global_re_cache->jstack);
 		}
 	}
@@ -291,12 +315,13 @@ rspamd_regexp_post_process (rspamd_regexp_t *r)
 }
 
 rspamd_regexp_t*
-rspamd_regexp_new (const gchar *pattern, const gchar *flags,
+rspamd_regexp_new_len (const gchar *pattern, gsize len, const gchar *flags,
 		GError **err)
 {
-	const gchar *start = pattern, *end, *flags_str = NULL;
+	const gchar *start = pattern, *end = start + len, *flags_str = NULL, *flags_end = NULL;
 	gchar *err_str;
 	rspamd_regexp_t *res;
+	gboolean explicit_utf = FALSE;
 	PCRE_T *r;
 	gchar sep = 0, *real_pattern;
 #ifndef WITH_PCRE2
@@ -309,12 +334,18 @@ rspamd_regexp_new (const gchar *pattern, const gchar *flags,
 
 	rspamd_regexp_library_init (NULL);
 
-	if (flags == NULL) {
+	if (pattern == NULL) {
+		g_set_error (err, rspamd_regexp_quark(), EINVAL,
+				"cannot create regexp from a NULL pattern");
+		return NULL;
+	}
+
+	if (flags == NULL && start + 1 < end) {
 		/* We need to parse pattern and detect flags set */
 		if (*start == '/') {
 			sep = '/';
 		}
-		else if (*start == 'm') {
+		else if (*start == 'm' && start[1] != '\\' && g_ascii_ispunct(start[1])) {
 			start ++;
 			sep = *start;
 
@@ -325,22 +356,23 @@ rspamd_regexp_new (const gchar *pattern, const gchar *flags,
 
 			rspamd_flags |= RSPAMD_REGEXP_FLAG_FULL_MATCH;
 		}
-		if (sep == '\0' || g_ascii_isalnum (sep)) {
+		if (sep == 0) {
 			/* We have no flags, no separators and just use all line as expr */
 			start = pattern;
-			end = start + strlen (pattern);
 			rspamd_flags &= ~RSPAMD_REGEXP_FLAG_FULL_MATCH;
 		}
 		else {
-			end = strrchr (pattern, sep);
+			gchar *last_sep = rspamd_memrchr(pattern, sep, len);
 
-			if (end == NULL || end <= start) {
+			if (last_sep == NULL || last_sep <= start) {
 				g_set_error (err, rspamd_regexp_quark(), EINVAL,
 						"pattern is not enclosed with %c: %s",
 						sep, pattern);
 				return NULL;
 			}
-			flags_str = end + 1;
+			flags_str = last_sep + 1;
+			flags_end = end;
+			end = last_sep;
 			start ++;
 		}
 	}
@@ -348,8 +380,10 @@ rspamd_regexp_new (const gchar *pattern, const gchar *flags,
 		/* Strictly check all flags */
 		strict_flags = TRUE;
 		start = pattern;
-		end = pattern + strlen (pattern);
 		flags_str = flags;
+		if (flags) {
+			flags_end = flags + strlen(flags);
+		}
 	}
 
 	rspamd_flags |= RSPAMD_REGEXP_FLAG_RAW;
@@ -362,7 +396,7 @@ rspamd_regexp_new (const gchar *pattern, const gchar *flags,
 #endif
 
 	if (flags_str != NULL) {
-		while (*flags_str) {
+		while (flags_str < flags_end) {
 			switch (*flags_str) {
 			case 'i':
 				regexp_flags |= PCRE_FLAG(CASELESS);
@@ -378,18 +412,25 @@ rspamd_regexp_new (const gchar *pattern, const gchar *flags,
 				break;
 			case 'u':
 				rspamd_flags &= ~RSPAMD_REGEXP_FLAG_RAW;
+				rspamd_flags |= RSPAMD_REGEXP_FLAG_UTF;
 #ifndef WITH_PCRE2
 				regexp_flags |= PCRE_FLAG(UTF8);
 #else
 				regexp_flags |= PCRE_FLAG(UTF);
 #endif
+				explicit_utf = TRUE;
 				break;
 			case 'O':
 				/* We optimize all regexps by default */
 				rspamd_flags |= RSPAMD_REGEXP_FLAG_NOOPT;
 				break;
+			case 'L':
+				/* SOM_LEFTMOST hyperscan flag */
+				rspamd_flags |= RSPAMD_REGEXP_FLAG_LEFTMOST;
+				break;
 			case 'r':
 				rspamd_flags |= RSPAMD_REGEXP_FLAG_RAW;
+				rspamd_flags &= ~RSPAMD_REGEXP_FLAG_UTF;
 #ifndef WITH_PCRE2
 				regexp_flags &= ~PCRE_FLAG(UTF8);
 #else
@@ -433,8 +474,8 @@ fin:
 
 	if (r == NULL) {
 		g_set_error (err, rspamd_regexp_quark(), EINVAL,
-			"regexp parsing error: '%s' at position %d",
-			err_str, (gint)err_off);
+			"regexp parsing error: '%s' at position %d; pattern: %s",
+			err_str, (gint)err_off, real_pattern);
 		g_free (real_pattern);
 
 		return NULL;
@@ -453,7 +494,7 @@ fin:
 	if (rspamd_flags & RSPAMD_REGEXP_FLAG_RAW) {
 		res->raw_re = r;
 	}
-	else {
+	else if (!explicit_utf) {
 #ifndef WITH_PCRE2
 		res->raw_re = pcre_compile (real_pattern, regexp_flags & ~PCRE_FLAG(UTF8),
 				(const char **)&err_str, &err_off, NULL);
@@ -483,32 +524,27 @@ fin:
 			&ncaptures) == 0) {
 		res->ncaptures = ncaptures;
 	}
-
-	/* Check number of backrefs */
-	if (pcre_fullinfo (res->raw_re, res->extra, PCRE_INFO_BACKREFMAX,
-			&ncaptures) == 0) {
-		res->nbackref = ncaptures;
-	}
 #else
 	/* Check number of captures */
 	if (pcre2_pattern_info (res->raw_re, PCRE2_INFO_CAPTURECOUNT,
 			&ncaptures) == 0) {
 		res->ncaptures = ncaptures;
 	}
-
-	/* Check number of backrefs */
-	if (pcre2_pattern_info (res->raw_re, PCRE2_INFO_BACKREFMAX,
-			&ncaptures) == 0) {
-		res->nbackref = ncaptures;
-	}
 #endif
 
 	return res;
 }
 
+rspamd_regexp_t *
+rspamd_regexp_new (const gchar *pattern, const gchar *flags,
+									GError **err)
+{
+	return rspamd_regexp_new_len(pattern, strlen(pattern), flags, err);
+}
+
 #ifndef WITH_PCRE2
 gboolean
-rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
+rspamd_regexp_search (const rspamd_regexp_t *re, const gchar *text, gsize len,
 		const gchar **start, const gchar **end, gboolean raw,
 		GArray *captures)
 {
@@ -520,6 +556,7 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 	const gchar *mt;
 	gsize remain = 0;
 	gint rc, match_flags = 0, *ovec, ncaptures, i;
+	const int junk = 0xdeadbabe;
 
 	g_assert (re != NULL);
 	g_assert (text != NULL);
@@ -528,6 +565,10 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 		len = strlen (text);
 	}
 
+	if (re->match_limit > 0 && len > re->match_limit) {
+		len = re->match_limit;
+	}
+
 	if (end != NULL && *end != NULL) {
 		/* Incremental search */
 		mt = (*end);
@@ -558,20 +599,29 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 		r = re->re;
 		ext = re->extra;
 #if defined(HAVE_PCRE_JIT) && defined(HAVE_PCRE_JIT_FAST) && !defined(DISABLE_JIT_FAST)
-		if (g_utf8_validate (mt, remain, NULL)) {
+		if (rspamd_fast_utf8_validate (mt, remain) == 0) {
 			st = global_re_cache->jstack;
 		}
 		else {
-			msg_err ("bad utf8 input for JIT re");
+			msg_err ("bad utf8 input for JIT re '%s'", re->pattern);
 			return FALSE;
 		}
 #endif
 	}
 
-	g_assert (r != NULL);
+	if (r == NULL) {
+		/* Invalid regexp type for the specified input */
+		return FALSE;
+	}
+
 	ncaptures = (re->ncaptures + 1) * 3;
 	ovec = g_alloca (sizeof (gint) * ncaptures);
 
+
+	for (i = 0; i < ncaptures; i ++) {
+		ovec[i] = junk;
+	}
+
 	if (!(re->flags & RSPAMD_REGEXP_FLAG_NOOPT)) {
 #ifdef HAVE_PCRE_JIT
 # if defined(HAVE_PCRE_JIT_FAST) && !defined(DISABLE_JIT_FAST)
@@ -602,14 +652,24 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 	}
 
 	if (rc >= 0) {
-		if (start) {
-			*start = mt + ovec[0];
+		if (rc > 0) {
+			if (start) {
+				*start = mt + ovec[0];
+			}
+			if (end) {
+				*end = mt + ovec[1];
+			}
 		}
-		if (end) {
-			*end = mt + ovec[1];
+		else {
+			if (start) {
+				*start = mt;
+			}
+			if (end) {
+				*end = mt + remain;
+			}
 		}
 
-		if (captures != NULL && rc > 1) {
+		if (captures != NULL && rc >= 1) {
 			struct rspamd_re_capture *elt;
 
 			g_assert (g_array_get_element_size (captures) ==
@@ -617,9 +677,16 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 			g_array_set_size (captures, rc);
 
 			for (i = 0; i < rc; i ++) {
-				elt = &g_array_index (captures, struct rspamd_re_capture, i);
-				elt->p = mt + ovec[i * 2];
-				elt->len = (mt + ovec[i * 2 + 1]) - elt->p;
+				if (ovec[i * 2] != junk && ovec[i * 2] >= 0) {
+					elt = &g_array_index (captures, struct rspamd_re_capture, i);
+					elt->p = mt + ovec[i * 2];
+					elt->len = (mt + ovec[i * 2 + 1]) - elt->p;
+				}
+				else {
+					/* Runtime match returned fewer captures than expected */
+					g_array_set_size (captures, i);
+					break;
+				}
 
 			}
 		}
@@ -639,7 +706,7 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 #else
 /* PCRE 2 version */
 gboolean
-rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
+rspamd_regexp_search (const rspamd_regexp_t *re, const gchar *text, gsize len,
 		const gchar **start, const gchar **end, gboolean raw,
 		GArray *captures)
 {
@@ -647,7 +714,8 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 	pcre2_match_context *mcontext;
 	PCRE_T *r;
 	const gchar *mt;
-	gsize remain = 0, *ovec;
+	PCRE2_SIZE remain = 0, *ovec;
+	const PCRE2_SIZE junk = 0xdeadbabeeeeeeeeULL;
 	gint rc, match_flags, novec, i;
 	gboolean ret = FALSE;
 
@@ -658,6 +726,10 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 		len = strlen (text);
 	}
 
+	if (re->match_limit > 0 && len > re->match_limit) {
+		len = re->match_limit;
+	}
+
 	if (end != NULL && *end != NULL) {
 		/* Incremental search */
 		mt = (*end);
@@ -686,12 +758,25 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 		mcontext = re->mcontext;
 	}
 
+	if (r == NULL) {
+		/* Invalid regexp type for the specified input */
+		return FALSE;
+	}
+
 	match_data = pcre2_match_data_create (re->ncaptures + 1, NULL);
+	novec = pcre2_get_ovector_count (match_data);
+	ovec = pcre2_get_ovector_pointer (match_data);
+
+	/* Fill ovec with crap, so we can stop if actual matches is less than announced */
+	for (i = 0; i < novec; i ++) {
+		ovec[i * 2] = junk;
+		ovec[i * 2 + 1] = junk;
+	}
 
 #ifdef HAVE_PCRE_JIT
 	if (!(re->flags & RSPAMD_REGEXP_FLAG_DISABLE_JIT) && can_jit) {
-		if (re->re != re->raw_re && !g_utf8_validate (mt, remain, NULL)) {
-			msg_err ("bad utf8 input for JIT re");
+		if (re->re != re->raw_re && rspamd_fast_utf8_validate (mt, remain) != 0) {
+			msg_err ("bad utf8 input for JIT re '%s'", re->pattern);
 			return FALSE;
 		}
 
@@ -708,17 +793,24 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 #endif
 
 	if (rc >= 0) {
-		novec = pcre2_get_ovector_count (match_data);
-		ovec = pcre2_get_ovector_pointer (match_data);
-
-		if (start) {
-			*start = mt + ovec[0];
+		if (novec > 0) {
+			if (start) {
+				*start = mt + ovec[0];
+			}
+			if (end) {
+				*end = mt + ovec[1];
+			}
 		}
-		if (end) {
-			*end = mt + ovec[1];
+		else {
+			if (start) {
+				*start = mt;
+			}
+			if (end) {
+				*end = mt + remain;
+			}
 		}
 
-		if (captures != NULL && novec > 1) {
+		if (captures != NULL && novec >= 1) {
 			struct rspamd_re_capture *elt;
 
 			g_assert (g_array_get_element_size (captures) ==
@@ -726,10 +818,15 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 			g_array_set_size (captures, novec);
 
 			for (i = 0; i < novec; i ++) {
-				elt = &g_array_index (captures, struct rspamd_re_capture, i);
-				elt->p = mt + ovec[i * 2];
-				elt->len = (mt + ovec[i * 2 + 1]) - elt->p;
-
+				if (ovec[i * 2] != junk && ovec[i * 2] != PCRE2_UNSET) {
+					elt = &g_array_index (captures, struct rspamd_re_capture, i);
+					elt->p = mt + ovec[i * 2];
+					elt->len = (mt + ovec[i * 2 + 1]) - elt->p;
+				}
+				else {
+					g_array_set_size (captures, i);
+					break;
+				}
 			}
 		}
 
@@ -750,7 +847,7 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 #endif
 
 const char*
-rspamd_regexp_get_pattern (rspamd_regexp_t *re)
+rspamd_regexp_get_pattern (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
@@ -770,7 +867,7 @@ rspamd_regexp_set_flags (rspamd_regexp_t *re, guint new_flags)
 }
 
 guint
-rspamd_regexp_get_flags (rspamd_regexp_t *re)
+rspamd_regexp_get_flags (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
@@ -778,31 +875,15 @@ rspamd_regexp_get_flags (rspamd_regexp_t *re)
 }
 
 guint
-rspamd_regexp_get_pcre_flags (rspamd_regexp_t *re)
+rspamd_regexp_get_pcre_flags (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
 	return re->pcre_flags;
 }
 
-gint
-rspamd_regexp_get_nbackrefs (rspamd_regexp_t *re)
-{
-	g_assert (re != NULL);
-
-	return re->nbackref;
-}
-
-gint
-rspamd_regexp_get_ncaptures (rspamd_regexp_t *re)
-{
-	g_assert (re != NULL);
-
-	return re->ncaptures;
-}
-
 guint
-rspamd_regexp_get_maxhits (rspamd_regexp_t *re)
+rspamd_regexp_get_maxhits (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
@@ -822,7 +903,7 @@ rspamd_regexp_set_maxhits (rspamd_regexp_t *re, guint new_maxhits)
 }
 
 guint64
-rspamd_regexp_get_cache_id (rspamd_regexp_t *re)
+rspamd_regexp_get_cache_id (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
@@ -841,8 +922,28 @@ rspamd_regexp_set_cache_id (rspamd_regexp_t *re, guint64 id)
 	return old;
 }
 
+gsize
+rspamd_regexp_get_match_limit (const rspamd_regexp_t *re)
+{
+	g_assert (re != NULL);
+
+	return re->match_limit;
+}
+
+gsize
+rspamd_regexp_set_match_limit (rspamd_regexp_t *re, gsize lim)
+{
+	gsize old;
+
+	g_assert (re != NULL);
+	old = re->match_limit;
+	re->match_limit = lim;
+
+	return old;
+}
+
 gboolean
-rspamd_regexp_match (rspamd_regexp_t *re, const gchar *text, gsize len,
+rspamd_regexp_match (const rspamd_regexp_t *re, const gchar *text, gsize len,
 		gboolean raw)
 {
 	const gchar *start = NULL, *end = NULL;
@@ -850,6 +951,10 @@ rspamd_regexp_match (rspamd_regexp_t *re, const gchar *text, gsize len,
 	g_assert (re != NULL);
 	g_assert (text != NULL);
 
+	if (len == 0) {
+		len = strlen (text);
+	}
+
 	if (rspamd_regexp_search (re, text, len, &start, &end, raw, NULL)) {
 		if (start == text && end == text + len) {
 			return TRUE;
@@ -884,7 +989,7 @@ rspamd_regexp_set_ud (rspamd_regexp_t *re, gpointer ud)
 }
 
 gpointer
-rspamd_regexp_get_ud (rspamd_regexp_t *re)
+rspamd_regexp_get_ud (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
@@ -981,33 +1086,20 @@ rspamd_regexp_cache_create (struct rspamd_regexp_cache *cache,
 	res = rspamd_regexp_new (pattern, flags, err);
 
 	if (res) {
-		REF_RETAIN (res);
-		g_hash_table_insert (cache->tbl, res->id, res);
+		/* REF_RETAIN (res); */
+		if (g_hash_table_size (cache->tbl) < max_re_cache_size) {
+			g_hash_table_insert(cache->tbl, res->id, res);
+		}
+		else {
+			msg_warn("cannot insert regexp to the cache: maximum size is reached (%d expressions); "
+					 "it might be cached regexp misuse; regexp pattern: %s",
+				max_re_cache_size, pattern);
+		}
 	}
 
 	return res;
 }
 
-void rspamd_regexp_cache_insert (struct rspamd_regexp_cache* cache,
-		const gchar *pattern,
-		const gchar *flags, rspamd_regexp_t *re)
-{
-	g_assert (re != NULL);
-	g_assert (pattern != NULL);
-
-	if (cache == NULL) {
-		rspamd_regexp_library_init (NULL);
-		cache = global_re_cache;
-	}
-
-	g_assert (cache != NULL);
-	/* Generate custom id */
-	rspamd_regexp_generate_id (pattern, flags, re->id);
-
-	REF_RETAIN (re);
-	g_hash_table_insert (cache->tbl, re->id, re);
-}
-
 gboolean
 rspamd_regexp_cache_remove (struct rspamd_regexp_cache *cache,
 		rspamd_regexp_t *re)
@@ -1038,9 +1130,28 @@ rspamd_regexp_cache_destroy (struct rspamd_regexp_cache *cache)
 		}
 #endif
 #endif
+		g_free (cache);
 	}
 }
 
+RSPAMD_CONSTRUCTOR (rspamd_re_static_pool_ctor)
+{
+	global_re_cache = rspamd_regexp_cache_new ();
+#ifdef WITH_PCRE2
+	pcre2_ctx = pcre2_compile_context_create (NULL);
+	pcre2_set_newline (pcre2_ctx, PCRE_FLAG(NEWLINE_ANY));
+#endif
+}
+
+RSPAMD_DESTRUCTOR (rspamd_re_static_pool_dtor)
+{
+	rspamd_regexp_cache_destroy (global_re_cache);
+#ifdef WITH_PCRE2
+	pcre2_compile_context_free (pcre2_ctx);
+#endif
+}
+
+
 void
 rspamd_regexp_library_init (struct rspamd_config *cfg)
 {
@@ -1049,19 +1160,16 @@ rspamd_regexp_library_init (struct rspamd_config *cfg)
 			can_jit = FALSE;
 			check_jit = FALSE;
 		}
+		else if (!can_jit) {
+			check_jit = TRUE;
+		}
 	}
 
-	if (global_re_cache == NULL) {
-		global_re_cache = rspamd_regexp_cache_new ();
+	if (check_jit) {
 #ifdef HAVE_PCRE_JIT
 		gint jit, rc;
 		gchar *str;
 
-		if (check_jit) {
-#ifdef WITH_PCRE2
-			pcre2_ctx = pcre2_compile_context_create (NULL);
-			pcre2_set_newline (pcre2_ctx, PCRE_FLAG(NEWLINE_ANY));
-#endif
 #ifndef WITH_PCRE2
 			rc = pcre_config (PCRE_CONFIG_JIT, &jit);
 #else
@@ -1099,36 +1207,27 @@ rspamd_regexp_library_init (struct rspamd_config *cfg)
 			} else {
 				msg_info ("pcre is compiled without JIT support, so many optimizations"
 						  " are impossible");
+				can_jit = FALSE;
 			}
-		}
 #else
 		msg_info ("pcre is too old and has no JIT support, so many optimizations"
-				" are impossible");
-#endif
-	}
-}
-
-void
-rspamd_regexp_library_finalize (void)
-{
-	if (global_re_cache != NULL) {
-		rspamd_regexp_cache_destroy (global_re_cache);
-#ifdef WITH_PCRE2
-		pcre2_compile_context_free (pcre2_ctx);
+				  " are impossible");
+		can_jit = FALSE;
 #endif
+		check_jit = FALSE;
 	}
 }
 
 gpointer
-rspamd_regexp_get_id (rspamd_regexp_t *re)
+rspamd_regexp_get_id (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
-	return re->id;
+	return (gpointer)re->id;
 }
 
 gpointer
-rspamd_regexp_get_class (rspamd_regexp_t *re)
+rspamd_regexp_get_class (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
