@@ -33,6 +33,13 @@
 #define RSPAMD_REGEXP_FLAG_FULL_MATCH (1 << 3)
 #define RSPAMD_REGEXP_FLAG_PCRE_ONLY (1 << 4)
 #define RSPAMD_REGEXP_FLAG_DISABLE_JIT (1 << 5)
+#define RSPAMD_REGEXP_FLAG_UTF (1 << 6)
+#define RSPAMD_REGEXP_FLAG_LEFTMOST (1 << 7)
+
+
+#ifdef  __cplusplus
+extern "C" {
+#endif
 
 struct rspamd_config;
 
@@ -50,8 +57,18 @@ struct rspamd_re_capture {
  * @param err error pointer set if compilation failed
  * @return new regexp object
  */
-rspamd_regexp_t* rspamd_regexp_new (const gchar *pattern, const gchar *flags,
-		GError **err);
+rspamd_regexp_t *rspamd_regexp_new (const gchar *pattern, const gchar *flags,
+									GError **err);
+
+/**
+ * Create new rspamd regexp
+ * @param pattern regexp pattern
+ * @param flags flags (may be enclosed inside pattern)
+ * @param err error pointer set if compilation failed
+ * @return new regexp object
+ */
+rspamd_regexp_t *rspamd_regexp_new_len (const gchar *pattern, gsize len, const gchar *flags,
+									GError **err);
 
 /**
  * Search the specified regexp in the text
@@ -64,10 +81,10 @@ rspamd_regexp_t* rspamd_regexp_new (const gchar *pattern, const gchar *flags,
  * @param captures array of captured strings of type rspamd_fstring_capture or NULL
  * @return
  */
-gboolean rspamd_regexp_search (rspamd_regexp_t *re,
-		const gchar *text, gsize len,
-		const gchar **start, const gchar **end, gboolean raw,
-		GArray *captures);
+gboolean rspamd_regexp_search (const rspamd_regexp_t *re,
+							   const gchar *text, gsize len,
+							   const gchar **start, const gchar **end, gboolean raw,
+							   GArray *captures);
 
 
 /**
@@ -77,13 +94,13 @@ gboolean rspamd_regexp_search (rspamd_regexp_t *re,
  * @param len
  * @return
  */
-gboolean rspamd_regexp_match (rspamd_regexp_t *re,
-		const gchar *text, gsize len, gboolean raw);
+gboolean rspamd_regexp_match (const rspamd_regexp_t *re,
+							  const gchar *text, gsize len, gboolean raw);
 
 /**
  * Increase refcount for a regexp object
  */
-rspamd_regexp_t* rspamd_regexp_ref (rspamd_regexp_t *re);
+rspamd_regexp_t *rspamd_regexp_ref (rspamd_regexp_t *re);
 
 /**
  * Unref regexp object
@@ -103,30 +120,31 @@ void rspamd_regexp_set_ud (rspamd_regexp_t *re, gpointer ud);
  * @param re regexp object
  * @return opaque pointer
  */
-gpointer rspamd_regexp_get_ud (rspamd_regexp_t *re);
+gpointer rspamd_regexp_get_ud (const rspamd_regexp_t *re);
 
 /**
  * Get regexp ID suitable for hashing
  * @param re
  * @return
  */
-gpointer rspamd_regexp_get_id (rspamd_regexp_t *re);
+gpointer rspamd_regexp_get_id (const rspamd_regexp_t *re);
 
 /**
  * Get pattern for the specified regexp object
  * @param re
  * @return
  */
-const char* rspamd_regexp_get_pattern (rspamd_regexp_t *re);
+const char *rspamd_regexp_get_pattern (const rspamd_regexp_t *re);
 
 /**
  * Get PCRE flags for the regexp
  */
-guint rspamd_regexp_get_pcre_flags (rspamd_regexp_t *re);
+guint rspamd_regexp_get_pcre_flags (const rspamd_regexp_t *re);
+
 /**
  * Get rspamd flags for the regexp
  */
-guint rspamd_regexp_get_flags (rspamd_regexp_t *re);
+guint rspamd_regexp_get_flags (const rspamd_regexp_t *re);
 
 /**
  * Set rspamd flags for the regexp
@@ -136,37 +154,37 @@ guint rspamd_regexp_set_flags (rspamd_regexp_t *re, guint new_flags);
 /**
  * Set regexp maximum hits
  */
-guint rspamd_regexp_get_maxhits (rspamd_regexp_t *re);
+guint rspamd_regexp_get_maxhits (const rspamd_regexp_t *re);
 
 /**
  * Get regexp maximum hits
  */
 guint rspamd_regexp_set_maxhits (rspamd_regexp_t *re, guint new_maxhits);
 
-/**
- * Returns number of backreferences in a regexp
- */
-gint rspamd_regexp_get_nbackrefs (rspamd_regexp_t *re);
-
-/**
- * Returns number of capture groups in a regexp
- */
-gint rspamd_regexp_get_ncaptures (rspamd_regexp_t *re);
-
 /**
  * Returns cache id for a regexp
  */
-guint64 rspamd_regexp_get_cache_id (rspamd_regexp_t *re);
+guint64 rspamd_regexp_get_cache_id (const rspamd_regexp_t *re);
 
 /**
  * Sets cache id for a regexp
  */
 guint64 rspamd_regexp_set_cache_id (rspamd_regexp_t *re, guint64 id);
 
+/**
+ * Returns match limit for a regexp
+ */
+gsize rspamd_regexp_get_match_limit (const rspamd_regexp_t *re);
+
+/**
+ * Sets cache id for a regexp
+ */
+gsize rspamd_regexp_set_match_limit (rspamd_regexp_t *re, gsize lim);
+
 /**
  * Get regexp class for the re object
  */
-gpointer rspamd_regexp_get_class (rspamd_regexp_t *re);
+gpointer rspamd_regexp_get_class (const rspamd_regexp_t *re);
 
 /**
  * Set regexp class for the re object
@@ -178,7 +196,7 @@ gpointer rspamd_regexp_set_class (rspamd_regexp_t *re, gpointer re_class);
  * Create new regexp cache
  * @return
  */
-struct rspamd_regexp_cache* rspamd_regexp_cache_new (void);
+struct rspamd_regexp_cache *rspamd_regexp_cache_new (void);
 
 /**
  * Query rspamd cache for a specified regexp
@@ -187,20 +205,9 @@ struct rspamd_regexp_cache* rspamd_regexp_cache_new (void);
  * @param flags
  * @return
  */
-rspamd_regexp_t* rspamd_regexp_cache_query (struct rspamd_regexp_cache* cache,
-		const gchar *pattern,
-		const gchar *flags);
-
-/**
- * Insert item to the cache using custom pattern and flags
- * @param cache
- * @param pattern
- * @param flags
- * @param re
- */
-void rspamd_regexp_cache_insert (struct rspamd_regexp_cache* cache,
-		const gchar *pattern,
-		const gchar *flags, rspamd_regexp_t *re);
+rspamd_regexp_t *rspamd_regexp_cache_query (struct rspamd_regexp_cache *cache,
+											const gchar *pattern,
+											const gchar *flags);
 
 /**
  * Create or get cached regexp from the specified cache
@@ -210,9 +217,9 @@ void rspamd_regexp_cache_insert (struct rspamd_regexp_cache* cache,
  * @param err error pointer set if compilation failed
  * @return new regexp object
  */
-rspamd_regexp_t* rspamd_regexp_cache_create (struct rspamd_regexp_cache *cache,
-		const gchar *pattern,
-		const gchar *flags, GError **err);
+rspamd_regexp_t *rspamd_regexp_cache_create (struct rspamd_regexp_cache *cache,
+											 const gchar *pattern,
+											 const gchar *flags, GError **err);
 
 /**
  * Remove regexp from the cache
@@ -221,7 +228,7 @@ rspamd_regexp_t* rspamd_regexp_cache_create (struct rspamd_regexp_cache *cache,
  * @return TRUE if a regexp has been removed
  */
 gboolean rspamd_regexp_cache_remove (struct rspamd_regexp_cache *cache,
-		rspamd_regexp_t *re);
+									 rspamd_regexp_t *re);
 
 /**
  * Destroy regexp cache and unref all elements inside it
@@ -254,11 +261,6 @@ gint rspamd_regexp_cmp (gconstpointer a, gconstpointer b);
  */
 void rspamd_regexp_library_init (struct rspamd_config *cfg);
 
-/**
- * Cleanup internal library structures
- */
-void rspamd_regexp_library_finalize (void);
-
 /**
  * Create regexp from glob
  * @param gl
@@ -267,4 +269,8 @@ void rspamd_regexp_library_finalize (void);
  */
 rspamd_regexp_t *rspamd_regexp_from_glob (const gchar *gl, gsize sz, GError **err);
 
+#ifdef  __cplusplus
+}
+#endif
+
 #endif /* REGEXP_H_ */
