@@ -22,6 +22,9 @@
 
 #define REDIS_DEFAULT_TIMEOUT 1.0
 
+static const gchar *M = "rspamd lua redis";
+static void *redis_null;
+
 /***
  * @module rspamd_redis
  * This module implements redis asynchronous client for rspamd LUA API.
@@ -86,7 +89,6 @@ static const struct luaL_reg redislib_m[] = {
 #define REDIS_RELEASE REF_RELEASE
 #endif
 
-#ifdef WITH_HIREDIS
 struct lua_redis_request_specific_userdata;
 /**
  * Struct for userdata representation
@@ -94,24 +96,33 @@ struct lua_redis_request_specific_userdata;
 struct lua_redis_userdata {
 	redisAsyncContext *ctx;
 	struct rspamd_task *task;
+	struct rspamd_symcache_dynamic_item *item;
 	struct rspamd_async_session *s;
-	struct event_base *ev_base;
+	struct ev_loop *event_loop;
 	struct rspamd_config *cfg;
 	struct rspamd_redis_pool *pool;
 	gchar *server;
-	gchar *reqline;
+	gchar log_tag[RSPAMD_LOG_ID_LEN + 1];
 	struct lua_redis_request_specific_userdata *specific;
 	gdouble timeout;
 	guint16 port;
 	guint16 terminated;
 };
 
+#define msg_debug_lua_redis(...)  rspamd_conditional_debug_fast (NULL, NULL, \
+        rspamd_lua_redis_log_id, "lua_redis", ud->log_tag, \
+        G_STRFUNC, \
+        __VA_ARGS__)
+INIT_LOG_MODULE(lua_redis)
+
 #define LUA_REDIS_SPECIFIC_REPLIED (1 << 0)
 /* session was finished */
 #define LUA_REDIS_SPECIFIC_FINISHED (1 << 1)
 #define LUA_REDIS_ASYNC (1 << 0)
 #define LUA_REDIS_TEXTDATA (1 << 1)
 #define LUA_REDIS_TERMINATED (1 << 2)
+#define LUA_REDIS_NO_POOL (1 << 3)
+#define LUA_REDIS_SUBSCRIBED (1 << 4)
 #define IS_ASYNC(ctx) ((ctx)->flags & LUA_REDIS_ASYNC)
 
 struct lua_redis_request_specific_userdata {
@@ -119,11 +130,10 @@ struct lua_redis_request_specific_userdata {
 	guint nargs;
 	gchar **args;
 	gsize *arglens;
-	struct rspamd_async_watcher *w;
 	struct lua_redis_userdata *c;
 	struct lua_redis_ctx *ctx;
 	struct lua_redis_request_specific_userdata *next;
-	struct event timeout;
+	ev_timer timeout_ev;
 	guint flags;
 };
 
@@ -140,8 +150,9 @@ struct lua_redis_ctx {
 struct lua_redis_result {
 	gboolean is_error;
 	gint result_ref;
-	struct rspamd_async_watcher *w;
+	struct rspamd_symcache_dynamic_item *item;
 	struct rspamd_async_session *s;
+	struct rspamd_task *task;
 	struct lua_redis_request_specific_userdata *sp_ud;
 };
 
@@ -177,14 +188,12 @@ lua_redis_dtor (struct lua_redis_ctx *ctx)
 	struct redisAsyncContext *ac;
 
 	ud = &ctx->async;
-	msg_debug ("desctructing %p", ctx);
+	msg_debug_lua_redis ("destructing %p", ctx);
 
 	if (ud->ctx) {
 
 		LL_FOREACH_SAFE (ud->specific, cur, tmp) {
-			if (rspamd_event_pending (&cur->timeout, EV_TIMEOUT)) {
-				event_del (&cur->timeout);
-			}
+			ev_timer_stop (ud->event_loop, &cur->timeout_ev);
 
 			if (!(cur->flags & LUA_REDIS_SPECIFIC_REPLIED)) {
 				is_successful = FALSE;
@@ -198,7 +207,17 @@ lua_redis_dtor (struct lua_redis_ctx *ctx)
 		ud->terminated = 1;
 		ac = ud->ctx;
 		ud->ctx = NULL;
-		rspamd_redis_pool_release_connection (ud->pool, ac, !is_successful);
+
+		if (!is_successful) {
+			rspamd_redis_pool_release_connection (ud->pool, ac,
+					RSPAMD_REDIS_RELEASE_FATAL);
+		}
+		else {
+			rspamd_redis_pool_release_connection (ud->pool, ac,
+					(ctx->flags & LUA_REDIS_NO_POOL) ?
+					RSPAMD_REDIS_RELEASE_ENFORCE : RSPAMD_REDIS_RELEASE_DEFAULT);
+		}
+
 	}
 
 	LL_FOREACH_SAFE (ud->specific, cur, tmp) {
@@ -239,15 +258,18 @@ static void
 lua_redis_fin (void *arg)
 {
 	struct lua_redis_request_specific_userdata *sp_ud = arg;
+	struct lua_redis_userdata *ud;
 	struct lua_redis_ctx *ctx;
 
 	ctx = sp_ud->ctx;
+	ud = sp_ud->c;
 
-	if (rspamd_event_pending (&sp_ud->timeout, EV_TIMEOUT)) {
-		event_del (&sp_ud->timeout);
+	if (ev_can_stop (&sp_ud->timeout_ev)) {
+		ev_timer_stop (sp_ud->ctx->async.event_loop, &sp_ud->timeout_ev);
 	}
 
-	msg_debug ("finished redis query %p from session %p", sp_ud, ctx);
+	msg_debug_lua_redis ("finished redis query %p from session %p; refcount=%d",
+			sp_ud, ctx, ctx->ref.refcount);
 	sp_ud->flags |= LUA_REDIS_SPECIFIC_FINISHED;
 
 	REDIS_RELEASE (ctx);
@@ -266,12 +288,16 @@ lua_redis_push_error (const gchar *err,
 {
 	struct lua_redis_userdata *ud = sp_ud->c;
 	struct lua_callback_state cbs;
+	lua_State *L;
 
 	if (!(sp_ud->flags & (LUA_REDIS_SPECIFIC_REPLIED|LUA_REDIS_SPECIFIC_FINISHED))) {
 		if (sp_ud->cbref != -1) {
 
 			lua_thread_pool_prepare_callback (ud->cfg->lua_thread_pool, &cbs);
+			L = cbs.L;
 
+			lua_pushcfunction (L, &rspamd_lua_traceback);
+			int err_idx = lua_gettop (L);
 			/* Push error */
 			lua_rawgeti (cbs.L, LUA_REGISTRYINDEX, sp_ud->cbref);
 
@@ -280,18 +306,25 @@ lua_redis_push_error (const gchar *err,
 			/* Data is nil */
 			lua_pushnil (cbs.L);
 
-			if (lua_pcall (cbs.L, 2, 0, 0) != 0) {
-				msg_info ("call to callback failed: %s", lua_tostring (cbs.L, -1));
-				lua_pop (cbs.L, 1);
+			if (ud->item) {
+				rspamd_symcache_set_cur_item (ud->task, ud->item);
 			}
 
+			if (lua_pcall (cbs.L, 2, 0, err_idx) != 0) {
+				msg_info ("call to callback failed: %s", lua_tostring (cbs.L, -1));
+			}
+
+			lua_settop (L, err_idx - 1);
 			lua_thread_pool_restore_callback (&cbs);
 		}
 
 		sp_ud->flags |= LUA_REDIS_SPECIFIC_REPLIED;
 
 		if (connected && ud->s) {
-			rspamd_session_watcher_pop (ud->s, sp_ud->w);
+			if (ud->item) {
+				rspamd_symcache_item_async_dec_check (ud->task, ud->item, M);
+			}
+
 			rspamd_session_remove_event (ud->s, lua_redis_fin, sp_ud);
 		}
 		else {
@@ -311,8 +344,7 @@ lua_redis_push_reply (lua_State *L, const redisReply *r, gboolean text_data)
 		lua_pushinteger (L, r->integer);
 		break;
 	case REDIS_REPLY_NIL:
-		/* XXX: not the best approach */
-		lua_newuserdata (L, sizeof (gpointer));
+		lua_getfield (L, LUA_REGISTRYINDEX, "redis.null");
 		break;
 	case REDIS_REPLY_STRING:
 	case REDIS_REPLY_STATUS:
@@ -351,11 +383,16 @@ lua_redis_push_data (const redisReply *r, struct lua_redis_ctx *ctx,
 {
 	struct lua_redis_userdata *ud = sp_ud->c;
 	struct lua_callback_state cbs;
+	lua_State *L;
 
-	if (!(sp_ud->flags & (LUA_REDIS_SPECIFIC_REPLIED|LUA_REDIS_SPECIFIC_FINISHED))) {
+	if (!(sp_ud->flags & (LUA_REDIS_SPECIFIC_REPLIED|LUA_REDIS_SPECIFIC_FINISHED)) ||
+			(sp_ud->flags & LUA_REDIS_SUBSCRIBED)) {
 		if (sp_ud->cbref != -1) {
 			lua_thread_pool_prepare_callback (ud->cfg->lua_thread_pool, &cbs);
+			L = cbs.L;
 
+			lua_pushcfunction (L, &rspamd_lua_traceback);
+			int err_idx = lua_gettop (L);
 			/* Push error */
 			lua_rawgeti (cbs.L, LUA_REGISTRYINDEX, sp_ud->cbref);
 			/* Error is nil */
@@ -363,22 +400,44 @@ lua_redis_push_data (const redisReply *r, struct lua_redis_ctx *ctx,
 			/* Data */
 			lua_redis_push_reply (cbs.L, r, ctx->flags & LUA_REDIS_TEXTDATA);
 
-			if (lua_pcall (cbs.L, 2, 0, 0) != 0) {
-				msg_info ("call to callback failed: %s", lua_tostring (cbs.L, -1));
-				lua_pop (cbs.L, 1);
+			if (ud->item) {
+				rspamd_symcache_set_cur_item (ud->task, ud->item);
 			}
 
+			gint ret = lua_pcall (cbs.L, 2, 0, err_idx);
+
+			if (ret != 0) {
+				msg_info ("call to lua_redis callback failed (%d): %s",
+						ret, lua_tostring (cbs.L, -1));
+			}
+
+			lua_settop (L, err_idx - 1);
 			lua_thread_pool_restore_callback (&cbs);
 		}
 
+		if (sp_ud->flags & LUA_REDIS_SUBSCRIBED) {
+			if (!(sp_ud->flags & LUA_REDIS_SPECIFIC_REPLIED)) {
+				if (ev_can_stop (&sp_ud->timeout_ev)) {
+					ev_timer_stop (sp_ud->ctx->async.event_loop,
+							&sp_ud->timeout_ev);
+				}
+			}
+		}
+
 		sp_ud->flags |= LUA_REDIS_SPECIFIC_REPLIED;
 
-		if (ud->s) {
-			rspamd_session_watcher_pop (ud->s, sp_ud->w);
-			rspamd_session_remove_event (ud->s, lua_redis_fin, sp_ud);
-		}
-		else {
-			lua_redis_fin (sp_ud);
+		if (!(sp_ud->flags & LUA_REDIS_SUBSCRIBED)) {
+			if (ud->s) {
+				if (ud->item) {
+					rspamd_symcache_item_async_dec_check (ud->task,
+							ud->item, M);
+				}
+
+				rspamd_session_remove_event (ud->s, lua_redis_fin, sp_ud);
+			}
+			else {
+				lua_redis_fin (sp_ud);
+			}
 		}
 	}
 }
@@ -406,12 +465,14 @@ lua_redis_callback (redisAsyncContext *c, gpointer r, gpointer priv)
 		return;
 	}
 
-	msg_debug ("got reply from redis %p for query %p", sp_ud->c->ctx, sp_ud);
+	msg_debug_lua_redis ("got reply from redis %p for query %p", sp_ud->c->ctx,
+			sp_ud);
 
 	REDIS_RETAIN (ctx);
 
 	/* If session is finished, we cannot call lua callbacks */
-	if (!(sp_ud->flags & LUA_REDIS_SPECIFIC_FINISHED)) {
+	if (!(sp_ud->flags & LUA_REDIS_SPECIFIC_FINISHED) ||
+			(sp_ud->flags & LUA_REDIS_SUBSCRIBED)) {
 		if (c->err == 0) {
 			if (r != NULL) {
 				if (reply->type != REDIS_REPLY_ERROR) {
@@ -435,16 +496,22 @@ lua_redis_callback (redisAsyncContext *c, gpointer r, gpointer priv)
 		}
 	}
 
-	ctx->cmds_pending --;
+	if (!(sp_ud->flags & LUA_REDIS_SUBSCRIBED)) {
+		ctx->cmds_pending--;
 
-	if (ctx->cmds_pending == 0 && !ud->terminated) {
-		/* Disconnect redis early as we don't need it anymore */
-		ud->terminated = 1;
-		ac = ud->ctx;
-		ud->ctx = NULL;
+		if (ctx->cmds_pending == 0 && !ud->terminated) {
+			/* Disconnect redis early as we don't need it anymore */
+			ud->terminated = 1;
+			ac = ud->ctx;
+			ud->ctx = NULL;
 
-		if (ac) {
-			rspamd_redis_pool_release_connection (ud->pool, ac, FALSE);
+			if (ac) {
+				msg_debug_lua_redis ("release redis connection ud=%p; ctx=%p; refcount=%d",
+						ud, ctx, ctx->ref.refcount);
+				rspamd_redis_pool_release_connection (ud->pool, ac,
+						(ctx->flags & LUA_REDIS_NO_POOL) ?
+						RSPAMD_REDIS_RELEASE_ENFORCE : RSPAMD_REDIS_RELEASE_DEFAULT);
+			}
 		}
 	}
 
@@ -491,8 +558,16 @@ lua_redis_cleanup_events (struct lua_redis_ctx *ctx)
 	while (!g_queue_is_empty (ctx->events_cleanup)) {
 		struct lua_redis_result *result = g_queue_pop_head (ctx->events_cleanup);
 
-		rspamd_session_watcher_pop (result->s, result->w);
-		rspamd_session_remove_event (result->s, lua_redis_fin, result->sp_ud);
+		if (result->item) {
+			rspamd_symcache_item_async_dec_check (result->task, result->item, M);
+		}
+
+		if (result->s) {
+			rspamd_session_remove_event (result->s, lua_redis_fin, result->sp_ud);
+		}
+		else {
+			lua_redis_fin (result->sp_ud);
+		}
 
 		g_free (result);
 	}
@@ -532,86 +607,113 @@ lua_redis_callback_sync (redisAsyncContext *ac, gpointer r, gpointer priv)
 		return;
 	}
 
-	if (rspamd_event_pending (&sp_ud->timeout, EV_TIMEOUT)) {
-		event_del (&sp_ud->timeout);
+	if (ev_can_stop ( &sp_ud->timeout_ev)) {
+		ev_timer_stop (ud->event_loop, &sp_ud->timeout_ev);
 	}
 
-	msg_debug ("got reply from redis: %p for query %p", ac, sp_ud);
+	if (!(sp_ud->flags & LUA_REDIS_SPECIFIC_FINISHED)) {
+		msg_debug_lua_redis ("got reply from redis: %p for query %p", ac, sp_ud);
 
-	struct lua_redis_result *result = g_malloc0 (sizeof *result);
+		struct lua_redis_result *result = g_malloc0 (sizeof *result);
 
-	/* If session is finished, we cannot call lua callbacks */
-	if (ac->err == 0) {
-		if (r != NULL) {
-			if (reply->type != REDIS_REPLY_ERROR) {
-				result->is_error = FALSE;
-				lua_redis_push_reply (L, reply, ctx->flags & LUA_REDIS_TEXTDATA);
+		if (ac->err == 0) {
+			if (r != NULL) {
+				if (reply->type != REDIS_REPLY_ERROR) {
+					result->is_error = FALSE;
+					lua_redis_push_reply (L, reply, ctx->flags & LUA_REDIS_TEXTDATA);
+				}
+				else {
+					result->is_error = TRUE;
+					lua_pushstring (L, reply->str);
+				}
 			}
 			else {
 				result->is_error = TRUE;
-				lua_pushstring (L, reply->str);
+				lua_pushliteral (L, "received no data from server");
 			}
 		}
 		else {
 			result->is_error = TRUE;
-			lua_pushliteral (L, "received no data from server");
+			if (ac->err == REDIS_ERR_IO) {
+				lua_pushstring (L, strerror (errno));
+			}
+			else {
+				lua_pushstring (L, ac->errstr);
+			}
 		}
-	}
-	else {
-		result->is_error = TRUE;
-		if (ac->err == REDIS_ERR_IO) {
-			lua_pushstring (L, strerror (errno));
-		}
-		else {
-			lua_pushstring (L, ac->errstr);
+
+		/* if error happened, we should terminate the connection,
+		   and release it */
+
+		if (result->is_error && sp_ud->c->ctx) {
+			ac = sp_ud->c->ctx;
+			/* Set to NULL to avoid double free in dtor */
+			sp_ud->c->ctx = NULL;
+			ctx->flags |= LUA_REDIS_TERMINATED;
+
+			/*
+			 * This will call all callbacks pending so the entire context
+			 * will be destructed
+			 */
+			rspamd_redis_pool_release_connection (sp_ud->c->pool, ac,
+					RSPAMD_REDIS_RELEASE_FATAL);
 		}
+
+		result->result_ref = luaL_ref (L, LUA_REGISTRYINDEX);
+		result->s = ud->s;
+		result->item = ud->item;
+		result->task = ud->task;
+		result->sp_ud = sp_ud;
+
+		g_queue_push_tail (ctx->replies, result);
+
 	}
-	/* if error happened, we should terminate the connection,
-	   and release it */
-
-	if (result->is_error && sp_ud->c->ctx) {
-		ac = sp_ud->c->ctx;
-		/* Set to NULL to avoid double free in dtor */
-		sp_ud->c->ctx = NULL;
-		ctx->flags |= LUA_REDIS_TERMINATED;
-
-		/*
-		 * This will call all callbacks pending so the entire context
-		 * will be destructed
-		 */
-		rspamd_redis_pool_release_connection (sp_ud->c->pool, ac, TRUE);
-	}
-
-	result->result_ref = luaL_ref (L, LUA_REGISTRYINDEX);
-	result->s = ud->s;
-	result->w = sp_ud->w;
-	result->sp_ud = sp_ud;
-
-	g_queue_push_tail (ctx->replies, result);
 
 	ctx->cmds_pending --;
 
 	if (ctx->cmds_pending == 0) {
 		if (ctx->thread) {
-			/* somebody yielded and waits for results */
-			thread = ctx->thread;
-			ctx->thread = NULL;
+			if (!(sp_ud->flags & LUA_REDIS_SPECIFIC_FINISHED)) {
+				/* somebody yielded and waits for results */
+				thread = ctx->thread;
+				ctx->thread = NULL;
 
-			results = lua_redis_push_results (ctx, thread->lua_state);
-			lua_thread_resume (thread, results);
-			lua_redis_cleanup_events (ctx);
+				results = lua_redis_push_results(ctx, thread->lua_state);
+
+				if (ud->item) {
+					rspamd_symcache_set_cur_item (ud->task, ud->item);
+				}
+
+				lua_thread_resume (thread, results);
+				lua_redis_cleanup_events(ctx);
+			}
+			else {
+				/* We cannot resume the thread as the associated task has gone */
+				lua_thread_pool_terminate_entry_full (ud->cfg->lua_thread_pool,
+						ctx->thread, G_STRLOC, true);
+				ctx->thread = NULL;
+			}
 		}
 	}
+
 }
 
 static void
-lua_redis_timeout_sync (int fd, short what, gpointer priv)
+lua_redis_timeout_sync (EV_P_ ev_timer *w, int revents)
 {
-	struct lua_redis_request_specific_userdata *sp_ud = priv;
-	struct lua_redis_ctx *ctx = sp_ud->ctx;
+	struct lua_redis_request_specific_userdata *sp_ud =
+			(struct lua_redis_request_specific_userdata *)w->data;
+	struct lua_redis_ctx *ctx;
+	struct lua_redis_userdata *ud;
 	redisAsyncContext *ac;
 
-	msg_debug ("timeout while querying redis server: %p, redis: %p", sp_ud,
+	if (sp_ud->flags & LUA_REDIS_SPECIFIC_FINISHED) {
+		return;
+	}
+
+	ud = sp_ud->c;
+	ctx = sp_ud->ctx;
+	msg_debug_lua_redis ("timeout while querying redis server: %p, redis: %p", sp_ud,
 			sp_ud->c->ctx);
 
 	if (sp_ud->c->ctx) {
@@ -627,14 +729,17 @@ lua_redis_timeout_sync (int fd, short what, gpointer priv)
 		 * This will call all callbacks pending so the entire context
 		 * will be destructed
 		 */
-		rspamd_redis_pool_release_connection (sp_ud->c->pool, ac, TRUE);
+		rspamd_redis_pool_release_connection (sp_ud->c->pool, ac,
+				RSPAMD_REDIS_RELEASE_FATAL);
 	}
 }
 
 static void
-lua_redis_timeout (int fd, short what, gpointer u)
+lua_redis_timeout (EV_P_ ev_timer *w, int revents)
 {
-	struct lua_redis_request_specific_userdata *sp_ud = u;
+	struct lua_redis_request_specific_userdata *sp_ud =
+			(struct lua_redis_request_specific_userdata *)w->data;
+	struct lua_redis_userdata *ud;
 	struct lua_redis_ctx *ctx;
 	redisAsyncContext *ac;
 
@@ -643,9 +748,10 @@ lua_redis_timeout (int fd, short what, gpointer u)
 	}
 
 	ctx = sp_ud->ctx;
+	ud = sp_ud->c;
 
 	REDIS_RETAIN (ctx);
-	msg_debug ("timeout while querying redis server: %p, redis: %p", sp_ud,
+	msg_debug_lua_redis ("timeout while querying redis server: %p, redis: %p", sp_ud,
 			sp_ud->c->ctx);
 	lua_redis_push_error ("timeout while connecting the server", ctx, sp_ud, TRUE);
 
@@ -659,7 +765,8 @@ lua_redis_timeout (int fd, short what, gpointer u)
 		 * This will call all callbacks pending so the entire context
 		 * will be destructed
 		 */
-		rspamd_redis_pool_release_connection (sp_ud->c->pool, ac, TRUE);
+		rspamd_redis_pool_release_connection (sp_ud->c->pool, ac,
+				RSPAMD_REDIS_RELEASE_FATAL);
 	}
 
 	REDIS_RELEASE (ctx);
@@ -765,17 +872,17 @@ lua_redis_parse_args (lua_State *L, gint idx, const gchar *cmd,
 static struct lua_redis_ctx *
 rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_async)
 {
-	struct lua_redis_ctx *ctx;
+	struct lua_redis_ctx *ctx = NULL;
 	rspamd_inet_addr_t *ip = NULL;
-	struct lua_redis_userdata *ud;
+	struct lua_redis_userdata *ud = NULL;
 	struct rspamd_lua_ip *addr = NULL;
 	struct rspamd_task *task = NULL;
-	const gchar *host;
-	const gchar *password = NULL, *dbname = NULL;
+	const gchar *host = NULL;
+	const gchar *password = NULL, *dbname = NULL, *log_tag = NULL;
 	gint cbref = -1;
 	struct rspamd_config *cfg = NULL;
 	struct rspamd_async_session *session = NULL;
-	struct event_base *ev_base = NULL;
+	struct ev_loop *ev_base = NULL;
 	gboolean ret = FALSE;
 	guint flags = 0;
 
@@ -812,13 +919,6 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 			}
 			lua_pop (L, 1);
 
-			lua_pushstring (L, "ev_base");
-			lua_gettable (L, -2);
-			if (lua_type (L, -1) == LUA_TUSERDATA) {
-				ev_base = lua_check_ev_base (L, -1);
-			}
-			lua_pop (L, 1);
-
 			if (cfg && ev_base) {
 				ret = TRUE;
 			}
@@ -832,8 +932,10 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 		else {
 			cfg = task->cfg;
 			session = task->s;
-			ev_base = task->ev_base;
+			ev_base = task->event_loop;
+			log_tag = task->task_pool->tag.uid;
 			ret = TRUE;
+
 		}
 
 		if (pcbref) {
@@ -855,11 +957,13 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 
 		if (lua_type (L, -1) == LUA_TUSERDATA) {
 			addr = lua_check_ip (L, -1);
+			host = rspamd_inet_address_to_string_pretty (addr->addr);
 		}
 		else if (lua_type (L, -1) == LUA_TSTRING) {
 			host = lua_tostring (L, -1);
 
-			if (rspamd_parse_inet_address (&ip, host, strlen (host))) {
+			if (rspamd_parse_inet_address (&ip,
+					host, strlen (host), RSPAMD_INET_ADDRESS_PARSE_DEFAULT)) {
 				addr = g_alloca (sizeof (*addr));
 				addr->addr = ip;
 
@@ -891,6 +995,13 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 		}
 		lua_pop (L, 1);
 
+		lua_pushstring (L, "no_pool");
+		lua_gettable (L, -2);
+		if (!!lua_toboolean (L, -1)) {
+			flags |= LUA_REDIS_NO_POOL;
+		}
+		lua_pop (L, 1);
+
 		lua_pop (L, 1); /* table */
 
 		if (session && rspamd_session_blocked (session)) {
@@ -911,12 +1022,27 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 				ctx->events_cleanup = g_queue_new ();
 
 			}
+
 			ud->s = session;
 			ud->cfg = cfg;
 			ud->pool = cfg->redis_pool;
-			ud->ev_base = ev_base;
+			ud->event_loop = ev_base;
 			ud->task = task;
 
+			if (log_tag) {
+				rspamd_strlcpy (ud->log_tag, log_tag, sizeof (ud->log_tag));
+			}
+			else {
+				/* Use pointer itself as a tag */
+				rspamd_snprintf (ud->log_tag, sizeof (ud->log_tag),
+						"%ud",
+						(int)rspamd_cryptobox_fast_hash (&ud, sizeof (ud), 0));
+			}
+
+			if (task) {
+				ud->item = rspamd_symcache_get_cur_item (task);
+			}
+
 			ret = TRUE;
 		}
 		else {
@@ -944,11 +1070,13 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 			if (ud->ctx) {
 				msg_err_task_check ("cannot connect to redis: %s",
 						ud->ctx->errstr);
-				rspamd_redis_pool_release_connection (ud->pool, ud->ctx, TRUE);
+				rspamd_redis_pool_release_connection (ud->pool, ud->ctx,
+						RSPAMD_REDIS_RELEASE_FATAL);
 				ud->ctx = NULL;
 			}
 			else {
-				msg_err_task_check ("cannot connect to redis: unknown error");
+				msg_err_task_check ("cannot connect to redis (OS error): %s",
+						strerror (errno));
 			}
 
 			REDIS_RELEASE (ctx);
@@ -956,6 +1084,8 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 			return NULL;
 		}
 
+		msg_debug_lua_redis ("opened redis connection host=%s; ctx=%p; ud=%p",
+				host, ctx, ud);
 
 		return ctx;
 	}
@@ -986,7 +1116,6 @@ lua_redis_make_request (lua_State *L)
 	struct lua_redis_userdata *ud;
 	struct lua_redis_ctx *ctx, **pctx;
 	const gchar *cmd = NULL;
-	struct timeval tv;
 	gdouble timeout = REDIS_DEFAULT_TIMEOUT;
 	gint cbref = -1;
 	gboolean ret = FALSE;
@@ -1020,6 +1149,7 @@ lua_redis_make_request (lua_State *L)
 				&sp_ud->nargs);
 		lua_pop (L, 1);
 		LL_PREPEND (ud->specific, sp_ud);
+
 		ret = redisAsyncCommandArgv (ud->ctx,
 				lua_redis_callback,
 				sp_ud,
@@ -1029,25 +1159,34 @@ lua_redis_make_request (lua_State *L)
 
 		if (ret == REDIS_OK) {
 			if (ud->s) {
-				rspamd_session_add_event (ud->s, NULL, lua_redis_fin, sp_ud, g_quark_from_static_string ("lua redis"));
-				sp_ud->w = rspamd_session_get_watcher (ud->s);
-				rspamd_session_watcher_push (ud->s);
-			}
-			else {
-				sp_ud->w = NULL;
+				rspamd_session_add_event (ud->s,
+						lua_redis_fin, sp_ud,
+						M);
+
+				if (ud->item) {
+					rspamd_symcache_item_async_inc (ud->task, ud->item, M);
+				}
 			}
 
 			REDIS_RETAIN (ctx); /* Cleared by fin event */
 			ctx->cmds_pending ++;
-			double_to_tv (timeout, &tv);
-			event_set (&sp_ud->timeout, -1, EV_TIMEOUT, lua_redis_timeout, sp_ud);
-			event_base_set (ud->ev_base, &sp_ud->timeout);
-			event_add (&sp_ud->timeout, &tv);
+
+			if (ud->ctx->c.flags & REDIS_SUBSCRIBED) {
+				msg_debug_lua_redis ("subscribe command, never unref/timeout");
+				sp_ud->flags |= LUA_REDIS_SUBSCRIBED;
+			}
+
+			sp_ud->timeout_ev.data = sp_ud;
+			ev_now_update_if_cheap ((struct ev_loop *)ud->event_loop);
+			ev_timer_init (&sp_ud->timeout_ev, lua_redis_timeout, timeout, 0.0);
+			ev_timer_start (ud->event_loop, &sp_ud->timeout_ev);
+
 			ret = TRUE;
 		}
 		else {
 			msg_info ("call to redis failed: %s", ud->ctx->errstr);
-			rspamd_redis_pool_release_connection (ud->pool, ud->ctx, TRUE);
+			rspamd_redis_pool_release_connection (ud->pool, ud->ctx,
+					RSPAMD_REDIS_RELEASE_FATAL);
 			ud->ctx = NULL;
 			REDIS_RELEASE (ctx);
 			ret = FALSE;
@@ -1114,7 +1253,8 @@ lua_redis_make_request_sync (lua_State *L)
 		}
 		else if (lua_type (L, -1) == LUA_TSTRING) {
 			host = lua_tostring (L, -1);
-			if (rspamd_parse_inet_address (&ip, host, strlen (host))) {
+			if (rspamd_parse_inet_address (&ip,
+					host, strlen (host), RSPAMD_INET_ADDRESS_PARSE_DEFAULT)) {
 				addr = g_alloca (sizeof (*addr));
 				addr->addr = ip;
 
@@ -1212,6 +1352,8 @@ lua_redis_make_request_sync (lua_State *L)
 			rspamd_inet_address_free (ip);
 		}
 		msg_err ("bad arguments for redis request");
+		lua_redis_free_args (args, arglens, nargs);
+
 		lua_pushboolean (L, FALSE);
 	}
 
@@ -1224,7 +1366,7 @@ lua_redis_make_request_sync (lua_State *L)
  * @param {task} task worker task object
  * @param {ip|string} host server address
  * @param {number} timeout timeout in seconds for request (1.0 by default)
- * @return {redis} new connection object or nil if connection failed
+ * @return {boolean,redis} new connection object or nil if connection failed
  */
 static int
 lua_redis_connect (lua_State *L)
@@ -1255,11 +1397,12 @@ lua_redis_connect (lua_State *L)
 		return 2;
 	}
 
+	lua_pushboolean (L, TRUE);
 	pctx = lua_newuserdata (L, sizeof (ctx));
 	*pctx = ctx;
 	rspamd_lua_setclass (L, "rspamd{redis}", -1);
 
-	return 1;
+	return 2;
 }
 
 /***
@@ -1273,7 +1416,6 @@ static int
 lua_redis_connect_sync (lua_State *L)
 {
 	LUA_TRACE_POINT;
-	rspamd_inet_addr_t *ip = NULL;
 	gdouble timeout = REDIS_DEFAULT_TIMEOUT;
 	struct lua_redis_ctx *ctx, **pctx;
 
@@ -1295,13 +1437,8 @@ lua_redis_connect_sync (lua_State *L)
 		pctx = lua_newuserdata (L, sizeof (ctx));
 		*pctx = ctx;
 		rspamd_lua_setclass (L, "rspamd{redis}", -1);
-
 	}
 	else {
-		if (ip) {
-			rspamd_inet_address_free (ip);
-		}
-
 		lua_pushboolean (L, FALSE);
 		lua_pushstring (L, "bad arguments for redis request");
 		return 2;
@@ -1327,7 +1464,6 @@ lua_redis_add_cmd (lua_State *L)
 	const gchar *cmd = NULL;
 	gint args_pos = 2;
 	gint cbref = -1, ret;
-	struct timeval tv;
 
 	if (ctx) {
 		if (ctx->flags & LUA_REDIS_TERMINATED) {
@@ -1396,20 +1532,28 @@ lua_redis_add_cmd (lua_State *L)
 
 		if (ret == REDIS_OK) {
 			if (ud->s) {
-				rspamd_session_add_event (ud->s, NULL, lua_redis_fin, sp_ud, g_quark_from_static_string ("lua redis"));
-				sp_ud->w = rspamd_session_get_watcher (ud->s);
-				rspamd_session_watcher_push (ud->s);
+				rspamd_session_add_event (ud->s,
+						lua_redis_fin,
+						sp_ud,
+						M);
+
+				if (ud->item) {
+					rspamd_symcache_item_async_inc (ud->task, ud->item, M);
+				}
 			}
 
-			double_to_tv (sp_ud->c->timeout, &tv);
+			sp_ud->timeout_ev.data = sp_ud;
+
 			if (IS_ASYNC (ctx)) {
-				event_set (&sp_ud->timeout, -1, EV_TIMEOUT, lua_redis_timeout, sp_ud);
+				ev_timer_init (&sp_ud->timeout_ev, lua_redis_timeout,
+						sp_ud->c->timeout, 0.0);
 			}
 			else {
-				event_set (&sp_ud->timeout, -1, EV_TIMEOUT, lua_redis_timeout_sync, sp_ud);
+				ev_timer_init (&sp_ud->timeout_ev, lua_redis_timeout_sync,
+						sp_ud->c->timeout, 0.0);
 			}
-			event_base_set (ud->ev_base, &sp_ud->timeout);
-			event_add (&sp_ud->timeout, &tv);
+
+			ev_timer_start (ud->event_loop, &sp_ud->timeout_ev);
 			REDIS_RETAIN (ctx);
 			ctx->cmds_pending ++;
 		}
@@ -1451,88 +1595,20 @@ lua_redis_exec (lua_State *L)
 		return 0;
 	}
 	else {
-		if (false /* !ctx->d.sync */) {
-			lua_pushstring (L, "cannot exec commands when not connected");
+		if (ctx->cmds_pending == 0 && g_queue_get_length (ctx->replies) == 0) {
+			lua_pushstring (L, "No pending commands to execute");
 			lua_error (L);
-			return 0;
+		}
+		if (ctx->cmds_pending == 0 && g_queue_get_length (ctx->replies) > 0) {
+			gint results = lua_redis_push_results (ctx, L);
+			return results;
 		}
 		else {
-			if (ctx->cmds_pending == 0 && g_queue_get_length (ctx->replies) == 0) {
-				lua_pushstring (L, "No pending commands to execute");
-				lua_error (L);
-			}
-			if (ctx->cmds_pending == 0 && g_queue_get_length (ctx->replies) > 0) {
-				gint results = lua_redis_push_results (ctx, L);
-				return results;
-			}
-			else {
-				ctx->thread = lua_thread_pool_get_running_entry (ctx->async.cfg->lua_thread_pool);
-				return lua_thread_yield (ctx->thread, 0);
-			}
+			ctx->thread = lua_thread_pool_get_running_entry (ctx->async.cfg->lua_thread_pool);
+			return lua_thread_yield (ctx->thread, 0);
 		}
 	}
 }
-#else
-static int
-lua_redis_make_request (lua_State *L)
-{
-	msg_warn ("rspamd is compiled with no redis support");
-
-	lua_pushboolean (L, FALSE);
-
-	return 1;
-}
-static int
-lua_redis_make_request_sync (lua_State *L)
-{
-	msg_warn ("rspamd is compiled with no redis support");
-
-	lua_pushboolean (L, FALSE);
-
-	return 1;
-}
-static int
-lua_redis_connect (lua_State *L)
-{
-	msg_warn ("rspamd is compiled with no redis support");
-
-	lua_pushboolean (L, FALSE);
-
-	return 1;
-}
-static int
-lua_redis_connect_sync (lua_State *L)
-{
-	msg_warn ("rspamd is compiled with no redis support");
-
-	lua_pushboolean (L, FALSE);
-
-	return 1;
-}
-static int
-lua_redis_add_cmd (lua_State *L)
-{
-	msg_warn ("rspamd is compiled with no redis support");
-
-	lua_pushboolean (L, FALSE);
-
-	return 1;
-}
-static int
-lua_redis_exec (lua_State *L)
-{
-	msg_warn ("rspamd is compiled with no redis support");
-
-	lua_pushboolean (L, FALSE);
-
-	return 1;
-}
-static int
-lua_redis_gc (lua_State *L)
-{
-	return 0;
-}
-#endif
 
 static gint
 lua_load_redis (lua_State * L)
@@ -1542,6 +1618,28 @@ lua_load_redis (lua_State * L)
 
 	return 1;
 }
+
+static gint
+lua_redis_null_idx (lua_State *L)
+{
+	lua_pushnil (L);
+
+	return 1;
+}
+
+static void
+lua_redis_null_mt (lua_State *L)
+{
+	luaL_newmetatable (L, "redis{null}");
+
+	lua_pushcfunction (L, lua_redis_null_idx);
+	lua_setfield (L, -2, "__index");
+	lua_pushcfunction (L, lua_redis_null_idx);
+	lua_setfield (L, -2, "__tostring");
+
+	lua_pop (L, 1);
+}
+
 /**
  * Open redis library
  * @param L lua stack
@@ -1550,17 +1648,14 @@ lua_load_redis (lua_State * L)
 void
 luaopen_redis (lua_State * L)
 {
-	luaL_newmetatable (L, "rspamd{redis}");
-	lua_pushstring (L, "__index");
-	lua_pushvalue (L, -2);
-	lua_settable (L, -3);
-
-	lua_pushstring (L, "class");
-	lua_pushstring (L, "rspamd{redis}");
-	lua_rawset (L, -3);
-
-	luaL_register (L, NULL, redislib_m);
+	rspamd_lua_new_class (L, "rspamd{redis}", redislib_m);
 	lua_pop (L, 1);
-
 	rspamd_lua_add_preload (L, "rspamd_redis", lua_load_redis);
+
+	/* Set null element */
+	lua_redis_null_mt (L);
+	redis_null = lua_newuserdata (L, 0);
+	luaL_getmetatable (L, "redis{null}");
+	lua_setmetatable (L, -2);
+	lua_setfield (L, LUA_REGISTRYINDEX, "redis.null");
 }
