@@ -599,8 +599,8 @@ lua_url_get_protocol (lua_State *L)
 
 /***
  * @method url:get_count()
- * Return number of occurrencies for this particular URL
- * @return {number} number of occurrencies
+ * Return number of occurrences for this particular URL
+ * @return {number} number of occurrences
  */
 static gint
 lua_url_get_count (lua_State *L)
@@ -742,7 +742,6 @@ lua_url_create (lua_State *L)
 	LUA_TRACE_POINT;
 	rspamd_mempool_t *pool;
 	struct rspamd_lua_text *t;
-	gboolean own_pool = FALSE;
 	struct rspamd_lua_url *u;
 
 	if (lua_type (L, 1) == LUA_TUSERDATA) {
@@ -750,7 +749,6 @@ lua_url_create (lua_State *L)
 		t = lua_check_text_or_string (L, 2);
 	}
 	else {
-		own_pool = TRUE;
 		pool = static_lua_url_pool;
 		t = lua_check_text_or_string (L, 2);
 	}
@@ -871,7 +869,7 @@ lua_url_all (lua_State *L)
  * - `host_encoded`: URL host part is encoded
  * - `schema_encoded`: URL schema part is encoded
  * - `query_encoded`: URL query part is encoded
- * - `missing_slahes`: URL has some slashes missing
+ * - `missing_slashes`: URL has some slashes missing
  * - `idn`: URL has international characters
  * - `has_port`: URL has port
  * - `has_user`: URL has user part
@@ -963,7 +961,7 @@ lua_tree_url_callback (gpointer key, gpointer value, gpointer ud)
 		}
 
 		if (cb->skip_prob > 0) {
-			gdouble coin = rspamd_random_double_fast_seed (cb->xoroshiro_state);
+			gdouble coin = rspamd_random_double_fast_seed (&cb->random_seed);
 
 			if (coin < cb->skip_prob) {
 				return;
@@ -1262,7 +1260,7 @@ lua_url_cbdata_fill_exclude_include (lua_State *L,
 	if (pos_arg_type == LUA_TTABLE) {
 		exclude_flags_mask = 0; /* Reset to no flags */
 
-		for (lua_pushnil(L); lua_next(L, pos); lua_pop (L, 1)) {
+		for (lua_pushnil(L); lua_next(L, pos + 1); lua_pop (L, 1)) {
 			int nmask = 0;
 
 			if (lua_type (L, -1) == LUA_TSTRING) {
@@ -1337,7 +1335,7 @@ lua_url_cbdata_dtor (struct lua_tree_cb_data *cbd)
 }
 
 gsize
-lua_url_adjust_skip_prob (gdouble timestamp,
+lua_url_adjust_skip_prob (float timestamp,
 						  guchar digest[16],
 						  struct lua_tree_cb_data *cb,
 						  gsize sz)
@@ -1347,11 +1345,11 @@ lua_url_adjust_skip_prob (gdouble timestamp,
 		/*
 		 * Use task dependent probabilistic seed to ensure that
 		 * consequent task:get_urls return the same list of urls
+		 * We use both digest and timestamp here to avoid attack surface
+		 * based just on digest.
 		 */
-		memset (cb->xoroshiro_state, 0, sizeof (cb->xoroshiro_state));
-		memcpy (&cb->xoroshiro_state[0], &timestamp,
-				MIN (sizeof (cb->xoroshiro_state[0]), sizeof (timestamp)));
-		memcpy (&cb->xoroshiro_state[1], digest, 16);
+		memcpy(&cb->random_seed, digest, 4);
+		memcpy(((unsigned char *)&cb->random_seed) + 4, &timestamp, 4);
 		sz = cb->max_urls;
 	}
 
