@@ -1,5 +1,5 @@
 --[[
-Copyright (c) 2011-2017, Vsevolod Stakhov <vsevolod@highsecure.ru>
+Copyright (c) 2022, Vsevolod Stakhov <vsevolod@rspamd.com>
 Copyright (c) 2016-2017, Andrew Lewis <nerf@judo.za.org>
 
 Licensed under the Apache License, Version 2.0 (the "License");
@@ -26,8 +26,10 @@ local lua_redis = require "lua_redis"
 local fun = require "fun"
 local lua_maps = require "lua_maps"
 local lua_util = require "lua_util"
+local lua_verdict = require "lua_verdict"
 local rspamd_hash = require "rspamd_cryptobox_hash"
 local lua_selectors = require "lua_selectors"
+local ts = require("tableshape").types
 
 -- A plugin that implements ratelimits using redis
 
@@ -49,6 +51,7 @@ local settings = {
   expire = 60 * 60 * 24 * 2, -- 2 days by default
   limits = {},
   allow_local = false,
+  prefilter = true,
 }
 
 -- Checks bucket, updating it if needed
@@ -57,50 +60,59 @@ local settings = {
 -- KEYS[3] - bucket leak rate (messages per millisecond)
 -- KEYS[4] - bucket burst
 -- KEYS[5] - expire for a bucket
+-- KEYS[6] - number of recipients
 -- return 1 if message should be ratelimited and 0 if not
 -- Redis keys used:
 --   l - last hit
 --   b - current burst
+--   p - pending messages (those that are currently processing)
 --   dr - current dynamic rate multiplier (*10000)
 --   db - current dynamic burst multiplier (*10000)
 local bucket_check_script = [[
   local last = redis.call('HGET', KEYS[1], 'l')
   local now = tonumber(KEYS[2])
+  local nrcpt = tonumber(KEYS[6])
   local dynr, dynb, leaked = 0, 0, 0
   if not last then
     -- New bucket
-    redis.call('HSET', KEYS[1], 'l', KEYS[2])
-    redis.call('HSET', KEYS[1], 'b', '0')
-    redis.call('HSET', KEYS[1], 'dr', '10000')
-    redis.call('HSET', KEYS[1], 'db', '10000')
+    redis.call('HMSET', KEYS[1], 'l', KEYS[2], 'b', '0', 'dr', '10000', 'db', '10000', 'p', tostring(nrcpt))
     redis.call('EXPIRE', KEYS[1], KEYS[5])
     return {0, '0', '1', '1', '0'}
   end
 
   last = tonumber(last)
-  local burst = tonumber(redis.call('HGET', KEYS[1], 'b'))
+  local burst,pending = unpack(redis.call('HMGET', KEYS[1], 'b', 'p'))
+  burst,pending = tonumber(burst or '0'),tonumber(pending or '0')
+  -- Sanity to avoid races
+  if burst < 0 then burst = 0 end
+  if pending < 0 then pending = 0 end
+  pending = pending + nrcpt -- this message
   -- Perform leak
-  if burst > 0 then
-   if last < tonumber(KEYS[2]) then
+  if burst + pending > 0 then
+   if burst > 0 and last < tonumber(KEYS[2]) then
     local rate = tonumber(KEYS[3])
     dynr = tonumber(redis.call('HGET', KEYS[1], 'dr')) / 10000.0
     if dynr == 0 then dynr = 0.0001 end
     rate = rate * dynr
     leaked = ((now - last) * rate)
+    if leaked > burst then leaked = burst end
     burst = burst - leaked
     redis.call('HINCRBYFLOAT', KEYS[1], 'b', -(leaked))
     redis.call('HSET', KEYS[1], 'l', KEYS[2])
    end
+
+   dynb = tonumber(redis.call('HGET', KEYS[1], 'db')) / 10000.0
+   if dynb == 0 then dynb = 0.0001 end
+
+   burst = burst + pending
+   if burst > 0 and (burst + tonumber(KEYS[6])) > tonumber(KEYS[4]) * dynb then
+     return {1, tostring(burst - pending), tostring(dynr), tostring(dynb), tostring(leaked)}
+   end
+   -- Increase pending if we allow ratelimit
+   redis.call('HINCRBY', KEYS[1], 'p', nrcpt)
   else
    burst = 0
-   redis.call('HSET', KEYS[1], 'b', '0')
-  end
-
-  dynb = tonumber(redis.call('HGET', KEYS[1], 'db')) / 10000.0
-  if dynb == 0 then dynb = 0.0001 end
-
-  if (burst + 1) > tonumber(KEYS[4]) * dynb then
-   return {1, tostring(burst), tostring(dynr), tostring(dynb), tostring(leaked)}
+   redis.call('HMSET', KEYS[1], 'b', '0', 'p', tostring(nrcpt))
   end
 
   return {0, tostring(burst), tostring(dynr), tostring(dynb), tostring(leaked)}
@@ -116,20 +128,20 @@ local bucket_check_id
 -- KEYS[5] - max dyn rate (min: 1/x)
 -- KEYS[6] - max burst rate (min: 1/x)
 -- KEYS[7] - expire for a bucket
+-- KEYS[8] - number of recipients (or increase rate)
 -- Redis keys used:
 --   l - last hit
 --   b - current burst
+--   p - messages pending (must be decreased by 1)
 --   dr - current dynamic rate multiplier
 --   db - current dynamic burst multiplier
 local bucket_update_script = [[
   local last = redis.call('HGET', KEYS[1], 'l')
   local now = tonumber(KEYS[2])
+  local nrcpt = tonumber(KEYS[8])
   if not last then
-    -- New bucket
-    redis.call('HSET', KEYS[1], 'l', KEYS[2])
-    redis.call('HSET', KEYS[1], 'b', '1')
-    redis.call('HSET', KEYS[1], 'dr', '10000')
-    redis.call('HSET', KEYS[1], 'db', '10000')
+    -- New bucket (why??)
+    redis.call('HMSET', KEYS[1], 'l', KEYS[2], 'b', tostring(nrcpt), 'dr', '10000', 'db', '10000', 'p', '0')
     redis.call('EXPIRE', KEYS[1], KEYS[7])
     return {1, 1, 1}
   end
@@ -180,17 +192,20 @@ local bucket_update_script = [[
     end
   end
 
-  local burst = tonumber(redis.call('HGET', KEYS[1], 'b'))
-  redis.call('HINCRBYFLOAT', KEYS[1], 'b', 1)
-  redis.call('HSET', KEYS[1], 'l', KEYS[2])
+  local burst,pending = unpack(redis.call('HMGET', KEYS[1], 'b', 'p'))
+  burst,pending = tonumber(burst or '0'),tonumber(pending or '0')
+  if burst < 0 then burst = nrcpt else burst = burst + nrcpt end
+  if pending < nrcpt then pending = 0 else pending = pending - nrcpt end
+
+  redis.call('HMSET', KEYS[1], 'b', tostring(burst), 'p', tostring(pending), 'l', KEYS[2])
   redis.call('EXPIRE', KEYS[1], KEYS[7])
 
   return {tostring(burst), tostring(dr), tostring(db)}
 ]]
 local bucket_update_id
 
--- message_func(task, limit_type, prefix, bucket)
-local message_func = function(_, limit_type, _, _)
+-- message_func(task, limit_type, prefix, bucket, limit_key)
+local message_func = function(_, limit_type, _, _, _)
   return string.format('Ratelimit "%s" exceeded', limit_type)
 end
 
@@ -264,48 +279,71 @@ local function parse_string_limit(lim, no_error)
   return nil
 end
 
+local function str_to_rate(str)
+  local divider,divisor = parse_string_limit(str, false)
+
+  if not divisor then
+    rspamd_logger.errx(rspamd_config, 'bad rate string: %s', str)
+
+    return nil
+  end
+
+  return divisor / divider
+end
+
+local bucket_schema = ts.shape{
+  burst = ts.number + ts.string / lua_util.dehumanize_number,
+  rate = ts.number + ts.string / str_to_rate,
+  skip_recipients = ts.boolean:is_optional(),
+  symbol = ts.string:is_optional(),
+  message = ts.string:is_optional(),
+  skip_soft_reject = ts.boolean:is_optional(),
+}
+
 local function parse_limit(name, data)
-  local buckets = {}
   if type(data) == 'table' then
-    -- 3 cases here:
+    -- 2 cases here:
     --  * old limit in format [burst, rate]
-    --  * vector of strings in Andrew's string format
+    --  * vector of strings in Andrew's string format (removed from 1.8.2)
     --  * proper bucket table
     if #data == 2 and tonumber(data[1]) and tonumber(data[2]) then
       -- Old style ratelimit
       rspamd_logger.warnx(rspamd_config, 'old style ratelimit for %s', name)
       if tonumber(data[1]) > 0 and tonumber(data[2]) > 0 then
-        table.insert(buckets, {
+        return {
           burst = data[1],
           rate = data[2]
-        })
+        }
       elseif data[1] ~= 0 then
         rspamd_logger.warnx(rspamd_config, 'invalid numbers for %s', name)
       else
         rspamd_logger.infox(rspamd_config, 'disable limit %s, burst is zero', name)
       end
+
+      return nil
     else
-      -- Recursively map parse_limit and flatten the list
-      fun.each(function(l)
-        -- Flatten list
-        for _,b in ipairs(l) do table.insert(buckets, b) end
-      end, fun.map(function(d) return parse_limit(d, name) end, data))
+      local parsed_bucket,err = bucket_schema:transform(data)
+
+      if not parsed_bucket or err then
+        rspamd_logger.errx(rspamd_config, 'cannot parse bucket for %s: %s; original value: %s',
+            name, err, data)
+      else
+        return parsed_bucket
+      end
     end
   elseif type(data) == 'string' then
     local rep_rate, burst = parse_string_limit(data)
-
+    rspamd_logger.warnx(rspamd_config, 'old style rate bucket config detected for %s: %s',
+        name, data)
     if rep_rate and burst then
-      table.insert(buckets, {
+      return {
         burst = burst,
-        rate = 1.0 / rep_rate -- reciprocal
-      })
+        rate = burst / rep_rate -- reciprocal
+      }
     end
   end
 
-  -- Filter valid
-  return fun.totable(fun.filter(function(val)
-    return type(val.burst) == 'number' and type(val.rate) == 'number'
-  end, buckets))
+  return nil
 end
 
 --- Check whether this addr is bounce
@@ -496,8 +534,6 @@ local function limit_to_prefixes(task, k, v, prefixes)
         end
       end
     end
-
-
   end
 
   return n
@@ -550,8 +586,8 @@ local function ratelimit_cb(task)
 
     if ret then
       local bucket = parse_limit(k, bd)
-      if bucket[1] then
-        prefixes[redis_key] = make_prefix(redis_key, k, bucket[1])
+      if bucket then
+        prefixes[redis_key] = make_prefix(redis_key, k, bucket)
       end
       nprefixes = nprefixes + 1
     else
@@ -560,7 +596,7 @@ local function ratelimit_cb(task)
     end
   end
 
-  local function gen_check_cb(prefix, bucket, lim_name)
+  local function gen_check_cb(prefix, bucket, lim_name, lim_key)
     return function(err, data)
       if err then
         rspamd_logger.errx('cannot check limit %s: %s %s', prefix, err, data)
@@ -572,26 +608,33 @@ local function ratelimit_cb(task)
 
         if data[1] == 1 then
           -- set symbol only and do NOT soft reject
-          if settings.symbol then
-            task:insert_result(settings.symbol, 0.0, lim_name .. "(" .. prefix .. ")")
-            rspamd_logger.infox(task,
-                'set_symbol_only: ratelimit "%s(%s)" exceeded, (%s / %s): %s (%s:%s dyn)',
-                lim_name, prefix,
-                bucket.burst, bucket.rate,
-                data[2], data[3], data[4])
-            return
-            -- set INFO symbol and soft reject
-          elseif settings.info_symbol then
-            task:insert_result(settings.info_symbol, 1.0,
-                lim_name .. "(" .. prefix .. ")")
+          if bucket.symbol then
+            -- Per bucket symbol
+            task:insert_result(bucket.symbol, 1.0,
+                string.format('%s(%s)', lim_name, lim_key))
+          else
+            if settings.symbol then
+              task:insert_result(settings.symbol, 1.0,
+                  string.format('%s(%s)', lim_name, lim_key))
+            elseif settings.info_symbol then
+              task:insert_result(settings.info_symbol, 1.0,
+                  string.format('%s(%s)', lim_name, lim_key))
+            end
           end
           rspamd_logger.infox(task,
-              'ratelimit "%s(%s)" exceeded, (%s / %s): %s (%s:%s dyn)',
+              'ratelimit "%s(%s)" exceeded, (%s / %s): %s (%s:%s dyn); redis key: %s',
               lim_name, prefix,
               bucket.burst, bucket.rate,
-              data[2], data[3], data[4])
-          task:set_pre_result('soft reject',
-              message_func(task, lim_name, prefix, bucket), N)
+              data[2], data[3], data[4], lim_key)
+
+          if not settings.symbol and not bucket.skip_soft_reject then
+            if not bucket.message then
+              task:set_pre_result('soft reject',
+                  message_func(task, lim_name, prefix, bucket, lim_key), N)
+            else
+              task:set_pre_result('soft reject', bucket.message)
+            end
+          end
         end
       end
     end
@@ -600,6 +643,11 @@ local function ratelimit_cb(task)
   -- Don't do anything if pre-result has been already set
   if task:has_pre_result() then return end
 
+  local _,nrcpt = task:has_recipients('smtp')
+  if not nrcpt or nrcpt <= 0 then
+    nrcpt = 1
+  end
+
   if nprefixes > 0 then
     -- Save prefixes to the cache to allow update
     task:cache_set('ratelimit_prefixes', prefixes)
@@ -610,18 +658,23 @@ local function ratelimit_cb(task)
     for pr,value in pairs(prefixes) do
       local bucket = value.bucket
       local rate = (bucket.rate) / 1000.0 -- Leak rate in messages/ms
+      local bincr = nrcpt
+      if bucket.skip_recipients then bincr = 1 end
+
       lua_util.debugm(N, task, "check limit %s:%s -> %s (%s/%s)",
           value.name, pr, value.hash, bucket.burst, bucket.rate)
       lua_redis.exec_redis_script(bucket_check_id,
-              {key = value.hash, task = task, is_write = true},
-              gen_check_cb(pr, bucket, value.name),
-              {value.hash, tostring(now), tostring(rate), tostring(bucket.burst),
-                  tostring(settings.expire)})
+          {key = value.hash, task = task, is_write = true},
+          gen_check_cb(pr, bucket, value.name, value.hash),
+          {value.hash, tostring(now), tostring(rate), tostring(bucket.burst),
+           tostring(settings.expire), tostring(bincr)})
     end
   end
 end
 
 local function ratelimit_update_cb(task)
+  if task:has_flag('skip') then return end
+  if not settings.allow_local and lua_util.is_rspamc_or_controller(task) then return end
   local prefixes = task:cache_get('ratelimit_prefixes')
 
   if prefixes then
@@ -631,7 +684,11 @@ local function ratelimit_update_cb(task)
       return
     end
 
-    local verdict = lua_util.get_task_verdict(task)
+    local verdict = lua_verdict.get_specific_verdict(N, task)
+    local _,nrcpt = task:has_recipients('smtp')
+    if not nrcpt or nrcpt <= 0 then
+      nrcpt = 1
+    end
 
     -- Update each bucket
     for k, v in pairs(prefixes) do
@@ -639,7 +696,7 @@ local function ratelimit_update_cb(task)
       local function update_bucket_cb(err, data)
         if err then
           rspamd_logger.errx(task, 'cannot update rate bucket %s: %s',
-                  k, err)
+              k, err)
         else
           lua_util.debugm(N, task,
               "updated limit %s:%s -> %s (%s/%s), burst: %s, dyn_rate: %s, dyn_burst: %s",
@@ -661,12 +718,15 @@ local function ratelimit_update_cb(task)
         mult_rate = bucket.ham_factor_rate or 1.0
       end
 
+      local bincr = nrcpt
+      if bucket.skip_recipients then bincr = 1 end
+
       lua_redis.exec_redis_script(bucket_update_id,
-              {key = v.hash, task = task, is_write = true},
-              update_bucket_cb,
-              {v.hash, tostring(now), tostring(mult_rate), tostring(mult_burst),
-               tostring(settings.max_rate_mult), tostring(settings.max_bucket_mult),
-               tostring(settings.expire)})
+          {key = v.hash, task = task, is_write = true},
+          update_bucket_cb,
+          {v.hash, tostring(now), tostring(mult_rate), tostring(mult_burst),
+           tostring(settings.max_rate_mult), tostring(settings.max_bucket_mult),
+           tostring(settings.expire), tostring(bincr)})
     end
   end
 end
@@ -683,42 +743,72 @@ if opts then
   if opts['rates'] and type(opts['rates']) == 'table' then
     -- new way of setting limits
     fun.each(function(t, lim)
-      local buckets
-      if type(lim) == 'table' and lim.selector and lim.bucket then
-        local selector = lua_selectors.parse_selector(rspamd_config, lim.selector)
-        if not selector then
-          rspamd_logger.errx(rspamd_config, 'bad ratelimit selector for %s: "%s"',
-              t, lim.selector)
-          return
-        end
-        local bucket = parse_limit(t, lim.bucket)
+      local buckets = {}
 
-        if not bucket then
-          rspamd_logger.errx(rspamd_config, 'bad ratelimit bucket for %s: "%s"',
-              t, lim.bucket)
-          return
+      if type(lim) == 'table' and lim.bucket then
+
+        if lim.bucket[1] then
+          for _,bucket in ipairs(lim.bucket) do
+            local b = parse_limit(t, bucket)
+
+            if not b then
+              rspamd_logger.errx(rspamd_config, 'bad ratelimit bucket for %s: "%s"',
+                  t, b)
+              return
+            end
+
+            table.insert(buckets, b)
+          end
+        else
+          local bucket = parse_limit(t, lim.bucket)
+
+          if not bucket then
+            rspamd_logger.errx(rspamd_config, 'bad ratelimit bucket for %s: "%s"',
+                t, lim.bucket)
+            return
+          end
+
+          buckets = {bucket}
         end
+
         settings.limits[t] = {
-          selector = selector,
-          buckets = bucket
+          buckets = buckets
         }
 
+        if lim.selector then
+          local selector = lua_selectors.parse_selector(rspamd_config, lim.selector)
+          if not selector then
+            rspamd_logger.errx(rspamd_config, 'bad ratelimit selector for %s: "%s"',
+                t, lim.selector)
+            settings.limits[t] = nil
+            return
+          end
+
+          settings.limits[t].selector = selector
+        end
       else
+        rspamd_logger.warnx(rspamd_config, 'old syntax for ratelimits: %s', lim)
         buckets = parse_limit(t, lim)
-        if buckets and #buckets > 0 then
+        if buckets then
           settings.limits[t] = {
-            buckets = buckets
+            buckets = {buckets}
           }
         end
       end
     end, opts['rates'])
   end
 
-  local enabled_limits = fun.totable(fun.map(function(t)
-    return t
+  -- Display what's enabled
+  fun.each(function(s)
+    rspamd_logger.infox(rspamd_config, 'enabled ratelimit: %s', s)
+  end, fun.map(function(n,d)
+    return string.format('%s [%s]', n,
+        table.concat(fun.totable(fun.map(function(v)
+          return string.format('symbol: %s, %s msgs burst, %s msgs/sec rate',
+              v.symbol, v.burst, v.rate)
+        end, d.buckets)), '; ')
+    )
   end, settings.limits))
-  rspamd_logger.infox(rspamd_config,
-          'enabled rate buckets: [%1]', table.concat(enabled_limits, ','))
 
   -- Ret, ret, ret: stupid legacy stuff:
   -- If we have a string with commas then load it as as static map
@@ -783,24 +873,56 @@ if opts then
     lua_util.disable_module(N, "redis")
   else
     local s = {
-      type = 'prefilter,nostat',
+      type = settings.prefilter and 'prefilter' or 'callback',
       name = 'RATELIMIT_CHECK',
-      priority = 7,
+      priority = lua_util.symbols_priorities.medium,
       callback = ratelimit_cb,
-      flags = 'empty',
+      flags = 'empty,nostat',
+      augmentations = {string.format("timeout=%f", redis_params.timeout or 0.0)},
     }
 
+    local id = rspamd_config:register_symbol(s)
+
+    -- Register per bucket symbols
+    -- Display what's enabled
+    fun.each(function(set, lim)
+      if type(lim.buckets) == 'table' then
+        for _,b in ipairs(lim.buckets) do
+          if b.symbol then
+            rspamd_config:register_symbol{
+              type = 'virtual',
+              name = b.symbol,
+              score = 0.0,
+              parent = id
+            }
+          end
+        end
+      end
+    end, settings.limits)
+
+    if settings.info_symbol then
+      rspamd_config:register_symbol{
+        type = 'virtual',
+        name = settings.info_symbol,
+        score = 0.0,
+        parent = id
+      }
+    end
     if settings.symbol then
-      s.name = settings.symbol
-    elseif settings.info_symbol then
-      s.name = settings.info_symbol
+      rspamd_config:register_symbol{
+        type = 'virtual',
+        name = settings.symbol,
+        score = 0.0, -- Might be overridden if needed
+        parent = id
+      }
     end
 
-    rspamd_config:register_symbol(s)
     rspamd_config:register_symbol {
       type = 'idempotent',
       name = 'RATELIMIT_UPDATE',
+      flags = 'explicit_disable,ignore_passthrough',
       callback = ratelimit_update_cb,
+      augmentations = {string.format("timeout=%f", redis_params.timeout or 0.0)},
     }
   end
 end
