@@ -55,20 +55,20 @@
 
 #define msg_err_session(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         session->pool->tag.tagname, session->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_session(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         session->pool->tag.tagname, session->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_session(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         session->pool->tag.tagname, session->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 #define msg_debug_session(...)  rspamd_conditional_debug_fast (NULL, session->client_addr, \
         rspamd_proxy_log_id, "proxy", session->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 INIT_LOG_MODULE(proxy)
@@ -157,6 +157,7 @@ struct rspamd_proxy_ctx {
 	struct rspamd_milter_context milter_ctx;
 	/* Language detector */
 	struct rspamd_lang_detector *lang_det;
+	gdouble task_timeout;
 };
 
 enum rspamd_backend_flags {
@@ -328,12 +329,27 @@ rspamd_proxy_parse_upstream (rspamd_mempool_t *pool,
 	L = ctx->lua_state;
 
 	if (ucl_object_type (obj) != UCL_OBJECT) {
+		if (ucl_object_type (obj) != UCL_NULL) {
+			msg_info_pool_check("upstream %s is disabled by setting it to NULL",
+					ucl_object_key (obj));
+			return TRUE;
+		}
+
 		g_set_error (err, rspamd_proxy_quark (), 100,
 				"upstream option must be an object");
 
 		return FALSE;
 	}
 
+	if (!rspamd_config_is_enabled_from_ucl (pool, obj)) {
+		/* Upstream is valid but disabled */
+		msg_info_pool_check("upstream %s is disabled",
+				ucl_object_lookup (obj, "name") ?
+					ucl_object_tostring(ucl_object_lookup (obj, "name")) :
+					ucl_object_key (obj));
+		return TRUE;
+	}
+
 	up = rspamd_mempool_alloc0 (pool, sizeof (*up));
 
 	elt = ucl_object_lookup (obj, "name");
@@ -491,12 +507,27 @@ rspamd_proxy_parse_mirror (rspamd_mempool_t *pool,
 	L = ctx->lua_state;
 
 	if (ucl_object_type (obj) != UCL_OBJECT) {
+		if (ucl_object_type (obj) != UCL_NULL) {
+			msg_info_pool_check("mirror %s is disabled by setting it to NULL",
+					ucl_object_key (obj));
+			return TRUE;
+		}
+
 		g_set_error (err, rspamd_proxy_quark (), 100,
 				"mirror option must be an object");
 
 		return FALSE;
 	}
 
+	if (!rspamd_config_is_enabled_from_ucl (pool, obj)) {
+		/* Upstream is valid but disabled */
+		msg_info_pool_check("mirror %s is disabled",
+				ucl_object_lookup (obj, "name") ?
+				ucl_object_tostring(ucl_object_lookup (obj, "name")) :
+				ucl_object_key (obj));
+		return TRUE;
+	}
+
 	up = rspamd_mempool_alloc0 (pool, sizeof (*up));
 
 	elt = ucl_object_lookup (obj, "name");
@@ -981,13 +1012,17 @@ proxy_call_cmp_script (struct rspamd_proxy_session *session, gint cbref)
 		}
 	}
 
-	if (lua_pcall (L, 1, 0, err_idx) != 0) {
+	gchar log_tag[RSPAMD_LOG_ID_LEN + 1];
+	rspamd_strlcpy(log_tag, session->pool->tag.uid, sizeof(log_tag));
+	lua_pushstring(L, log_tag);
+
+	if (lua_pcall (L, 2, 0, err_idx) != 0) {
 		msg_err_session (
 				"cannot run lua compare script: %s",
 				lua_tostring (L, -1));
 	}
 
-	lua_settop (L, 0);
+	lua_settop (L, err_idx - 1);
 }
 
 static void
@@ -1436,7 +1471,7 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 
 			msg->method = HTTP_GET;
 			rspamd_http_connection_write_message_shared (bk_conn->backend_conn,
-					msg, NULL, NULL, bk_conn,
+					msg, rspamd_upstream_name(bk_conn->up), NULL, bk_conn,
 					bk_conn->timeout);
 		}
 		else {
@@ -1463,7 +1498,7 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 			}
 
 			rspamd_http_connection_write_message (bk_conn->backend_conn,
-					msg, NULL, NULL, bk_conn,
+					msg, rspamd_upstream_name(bk_conn->up), NULL, bk_conn,
 					bk_conn->timeout);
 		}
 
@@ -1804,11 +1839,11 @@ rspamd_proxy_self_scan (struct rspamd_proxy_session *session)
 	task->sock = -1;
 
 	if (session->client_milter_conn) {
-		task->client_addr = rspamd_inet_address_copy (
-				session->client_milter_conn->addr);
+		task->client_addr = rspamd_inet_address_copy(
+				session->client_milter_conn->addr, NULL);
 	}
 	else {
-		task->client_addr = rspamd_inet_address_copy (session->client_addr);
+		task->client_addr = rspamd_inet_address_copy(session->client_addr, NULL);
 	}
 
 	task->fin_arg = session;
@@ -1852,7 +1887,7 @@ rspamd_proxy_self_scan (struct rspamd_proxy_session *session)
 
 	}
 	else if (session->ctx->has_self_scan) {
-		if (session->ctx->cfg->task_timeout > 0) {
+		if (!isnan(session->ctx->task_timeout) && session->ctx->task_timeout > 0) {
 			task->timeout_ev.data = task;
 			ev_timer_init (&task->timeout_ev, rspamd_task_timeout,
 					session->ctx->cfg->task_timeout,
@@ -2000,7 +2035,8 @@ retry:
 
 			rspamd_http_connection_write_message_shared (
 					session->master_conn->backend_conn,
-					msg, NULL, NULL, session->master_conn,
+					msg, rspamd_upstream_name(session->master_conn->up),
+					NULL, session->master_conn,
 					session->master_conn->timeout);
 		}
 		else {
@@ -2028,7 +2064,8 @@ retry:
 
 			rspamd_http_connection_write_message (
 					session->master_conn->backend_conn,
-					msg, NULL, NULL, session->master_conn,
+					msg, rspamd_upstream_name(session->master_conn->up),
+					NULL, session->master_conn,
 					session->master_conn->timeout);
 		}
 	}
@@ -2338,40 +2375,9 @@ start_rspamd_proxy (struct rspamd_worker *worker)
 		/* Additional initialisation needed */
 		rspamd_worker_init_scanner (worker, ctx->event_loop, ctx->resolver,
 				&ctx->lang_det);
+		ctx->task_timeout = rspamd_worker_check_and_adjust_timeout(ctx->cfg, NAN);
 
-		if (worker->index == 0) {
-			/*
-			 * If there are no controllers and no normal workers,
-			 * then pretend that we are a controller
-			 */
-			gboolean controller_seen = FALSE;
-			GList *cur;
-
-			cur = worker->srv->cfg->workers;
-
-			while (cur) {
-				struct rspamd_worker_conf *cf;
-
-				cf = (struct rspamd_worker_conf *)cur->data;
-				if ((cf->type == g_quark_from_static_string ("controller")) ||
-						(cf->type == g_quark_from_static_string ("normal"))) {
-
-					if (cf->enabled && cf->count >= 0) {
-						controller_seen = TRUE;
-						break;
-					}
-				}
-
-				cur = g_list_next (cur);
-			}
-
-			if (!controller_seen) {
-				msg_info ("no controller or normal workers defined, execute "
-							  "controller periodics in this worker");
-				worker->flags |= RSPAMD_WORKER_CONTROLLER;
-				is_controller = TRUE;
-			}
-		}
+		is_controller = rspamd_worker_check_controller_presence (worker);
 	}
 	else {
 		worker->flags &= ~RSPAMD_WORKER_SCANNER;
@@ -2422,6 +2428,7 @@ start_rspamd_proxy (struct rspamd_worker *worker)
 
 	REF_RELEASE (ctx->cfg);
 	rspamd_log_close (worker->srv->logger);
+	rspamd_unset_crash_handler (worker->srv);
 
 	exit (EXIT_SUCCESS);
 }
