@@ -23,9 +23,10 @@
 #include <glob.h>
 #endif
 
-static const char *lua_src = BUILDROOT "/test/lua/tests.lua";
+static const char *lua_src_name = "lua/tests.lua";
 extern gchar *lua_test;
 extern gchar *lua_test_case;
+extern gchar *argv0_dirname;
 extern struct rspamd_main *rspamd_main;
 
 static int
@@ -55,16 +56,18 @@ traceback (lua_State *L)
 	return 1;
 }
 
-void
+_Noreturn void
 rspamd_lua_test_func (void)
 {
-	lua_State *L = rspamd_lua_init ();
-	gchar *rp, rp_buf[PATH_MAX], path_buf[PATH_MAX], *tmp, *dir, *pattern;
+	lua_State *L = (lua_State *)rspamd_main->cfg->lua_state;
+	gchar *lua_src, *rp, rp_buf[PATH_MAX], path_buf[PATH_MAX], *tmp, *dir, *pattern;
 	const gchar *old_path;
 	glob_t globbuf;
 	gint i, len;
 
-	rspamd_lua_set_globals (rspamd_main->cfg, L, NULL);
+	rspamd_lua_set_env (L, NULL, NULL, NULL);
+	rspamd_lua_set_globals (rspamd_main->cfg, L);
+	rspamd_lua_start_gc (rspamd_main->cfg);
 
 	if (lua_test_case) {
 		lua_pushstring (L, lua_test_case);
@@ -73,11 +76,13 @@ rspamd_lua_test_func (void)
 
 	rspamd_printf ("Starting lua tests\n");
 
+	lua_src = g_build_filename (argv0_dirname, lua_src_name, NULL);
 	if ((rp = realpath (lua_src, rp_buf)) == NULL) {
 		msg_err ("cannot find path %s: %s",
 				lua_src, strerror (errno));
 		g_assert (0);
 	}
+	g_free (lua_src);
 
 	tmp = g_strdup (rp);
 	dir = dirname (tmp);
